/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#ifndef _labels_h_
#define _labels_h_

#include "Defs.h"
#include "Objects.h"
#include "Sets.h"

class StmtList ;  // defined later in Statements.h

class FtnLabel : public Enumerable {

friend class FortranParser ;
friend class Universe<FtnLabel> ;

private:
  PTR_STMT 	scope;		// level at which ident is declared
  PTR_CNTXT	pc ;            // parse context containing this label
  PTR_BLOCK	ud_chain;	// use-definition chain
  unsigned	labused :1;	// if it's been referenced
  unsigned	labinacc:1;	// illegal use of this label
  unsigned	labdefined:1;	// if this label been defined
  unsigned	labtype:2;	// UNKNOWN, EXEC, FORMAT, and OTHER
  int		stateno;	// statement label
  PTR_FTN_LABEL	next;		// point to next label in current scope
  PTR_STMT	statbody;	// point to body of statement

public:

  FtnLabel (int i) : Enumerable(i) { stateno = i ; }

  int variant()     { return (int)labtype ; }
  int value()       { return stateno ; }
  PTR_STMT statement() { return statbody ; }
		
  void set_fields (PTR_CNTXT cc, PTR_STMT sc, long num, unsigned used,
		   unsigned inacc, unsigned defined, unsigned ltype,
		   PTR_STMT body) {
    pc = cc ;
    scope = sc ;
    stateno = num ;
    labused = used ;
    labinacc = inacc ;
    labdefined = defined ;
    labtype = ltype ;
    statbody = body ;
  }
};


class CcLabel {
friend class LabelTable ;
private:
  string ident ;       // name of the label
  PTR_STMT stat ;      // if null, the label is undefined
  PTR_STMT func ;      // enclosing function header statement
  StmtList *uses ;     // list of 'goto' statements using this label

public:
  CcLabel (string* s, PTR_STMT f = 0, PTR_STMT st = 0) : ident (s)
    {
      func = f ;
      uses = 0 ;
      stat = st ;
    }

  // Obligatory functions for creating Maps of CcLabel objects

  char* index () { return ident(); }
  int variant () { return 0 ; }

  void define (PTR_STMT st) ;
  void set_func  (PTR_STMT f) { func = f ; }
  void add_use (PTR_STMT st) ;
  bool isDefined () { return (bool)(stat != 0); }
  PTR_STMT labelStmt () { return stat ; }
};


class LabelTable : public NameMap<CcLabel> {
public:

  LabelTable (PTR_STMT f, PTR_CNTXT p) { func_stmt = f ; pc = p ; }
  LabelTable (NameMap<CcLabel>& ct) : NameMap<CcLabel> (ct) { func_stmt = 0 ; }
  ~LabelTable () { }

  CcLabel* operator[] (PTR_HASH d) ;
  void add_goto (PTR_HASH d, PTR_STMT st) ;
  bool define_label (PTR_HASH d, PTR_STMT st) ;
  void check_undefined_labels () ;

private:
  PTR_STMT func_stmt ;
  PTR_CNTXT pc ;
};

class LabelTable_Iterator : public NameMap_Iterator<CcLabel> {
public:
  LabelTable_Iterator (NameMap<CcLabel>& ltab) : NameMap_Iterator<CcLabel> (ltab) {}
};

class CaseLabel {
friend class CaseTable ;
public:
  int index () { return caseValue ; }
  int variant () { return 0 ; }
  int value () { return caseValue ; }
  PTR_STMT statement () { return caseStmt ; }

  CaseLabel (int cval, PTR_STMT cstmt)
  { 
    caseValue = cval ;
    caseStmt = cstmt ;
    next = 0 ;
  }

private:
  int caseValue ;      // actual integer constant value of case label
  PTR_STMT caseStmt ;  // ptr to originating `case' statement
  CaseLabel* next ;    // to make a list of duplicate labels
};

class CaseTable : public Map<CaseLabel> {
public:
  
  CaseTable (PTR_STMT sw, PTR_CNTXT p) { switch_stmt = sw; pc = p; }
  CaseTable (Map<CaseLabel>& ct) : Map<CaseLabel> (ct)
  {
    switch_stmt = 0 ;
    pc = 0 ;
    low = INT_MAX ;
    high = INT_MIN ;
  }

  void add_case (int, PTR_STMT) ;
  int lwb () { return low ; }
  int upb () { return high ; }
  int range () { return high - low + 1 ; }
  float density () { return (float)ncases/(high - low + 1) ; }

private:
  PTR_STMT switch_stmt ;
  PTR_CNTXT pc ;
  int low, high, ncases ;
};

#endif // _labels_h_

