/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _ftnParser_h_
#define _ftnParser_h_

#include <string.h>
#include <stdio.h>
#include "Octave/Base/Strings.h"
#include "Octave/Base/ParseContext.h"
#include "Octave/Fortran/FortranLexer.h"
#include "Octave/Fortran/Fortran.h"

/* F90 permits us to declare variable either by ENTITY or by ATTRIBUTE
   So the following flags are used by the parser to indicate to the
   'deal_with_options' routine which of the two we are dealing with
   when an array declaration is being processed. This permits it to
   create either a VAR_REF or an ARRAY_REF node as appropriate for the
   variable(s) in the entity declaration list.
*/

#define ENTITY_DECL 1
#define ATTRIB_DECL 2

// CLASS GROUP: fortran parser classes


/*----------------------------------------------------------
CLASS
     FortranContext

     A keeper of variables for the fortran parser.

OVERVIEW

 Each object of this class holds all contextual info for
 parsing a Fortran source file. It is publicly derived
 from the generic ParseContext class, and in addition it
 provides state variables, language-specific flags and
 semantic routines used by the fortran parser.

 A set of access functions for the resulting parse-tree
 are also provided. This class contains a PTR_STMT pointer
 to the root of the parse tree.

 This is a base class for the FortranParser class.

----------------------------------------------------------*/

class FortranContext : public ParseContext {

public:
  string   infname ;    // source file name
  int      undeftype ;
  int      saveall ;
  int      maxdim ;
  int      parstate ;
  int      procclass ;
  PTR_TYPE impltype[26] ;
  int      nioctl ;
  int      errcnt ;
  int      nwarn ;
  int      do_name_err ;
  int      ndim ;
  int      varleng ;
  int      lastwasbranch ;
  int      thiswasbranch ;
  int      subscript_status ;
  int      type_options ;
  int      position ;
  int      attr_ndim ;
  int      end_group ;
  int      extend_flag ;
  PTR_SYMB type_var ;
  PTR_EXPR first_unresolved_call ;
  PTR_EXPR last_unresolved_call ;
  PTR_EXPR attr_dims ;
  PTR_FTN_LABEL thislabel ;
  PTR_TYPE vartype ;
  PTR_CMNT stmt_comments, cur_comment ;
  PTR_SCOPE root_scope ;
  PTR_STMT pred_stmt ;
  PTR_STMT last_stmt ;
  PTR_STMT module_header ;
  PTR_SCOPE module_scope ;
  int      cur_group ;      // indicates Statement list being built
  PTR_BLOCK *blocks [MAX_NEST_LEVEL] ;
  int      cur_level ;

  void setimpl (PTR_TYPE, int, int, PTR_SCOPE) ;
  void procinit () ;
  void create_procedure (PTR_STMT) ;

  PTR_EXPR intrinsic_op_node (char *, expression_tag, PTR_EXPR, PTR_EXPR) ;
  PTR_EXPR defined_op_node (PTR_HASH, PTR_EXPR, PTR_EXPR) ;

  PTR_SCOPE fileScope () { return root_scope ; }

  int scope_is_global  (PTR_SCOPE s) { return s == root_scope ; }
  int scope_is_current (PTR_SCOPE s) { return s == cur_scope() ; }

  void start_new_group1 (PTR_STMT s) {
    pred_stmt = s ;
    cur_group = NEW_GROUP1 ;
    blocks[++cur_level] = &(s->node.fields.bl_ptr1) ; 
  }

  void start_new_group2 (PTR_STMT s) { 
    pred_stmt = s ;
    cur_group = NEW_GROUP2 ;
    blocks[++cur_level] = &(s->node.fields.bl_ptr2) ;
  }

  PTR_BLOCK* current_block () { return blocks[cur_level] ; }

  PTR_STMT last_in_block () {
    PTR_BLOCK *b = current_block() ;
    if (*b)
      return (*b)->tail() ;
    else
      return 0 ;
  }

  void pop_level () {
    if (pred_stmt) {
      pred_stmt = pred_stmt->control_parent ;
      cur_level-- ;
      if (cur_level < 0)
	error ("internal: inconsistency in pop_level") ;
    }
    else
      error ("unexpected end of nest level") ;
  }

  void terminate_block (PTR_STMT) ;
  void terminate_block () ;
  
  void exit_group (PTR_STMT) ;
  void close_groups () ;
  void append_statement (PTR_STMT) ;
  void attach_statement (PTR_STMT) ;

  PTR_EXPR check_and_install (PTR_HASH, PTR_EXPR, int) ;
  PTR_SYMB install_entry (PTR_HASH, int) ;
  PTR_TYPE install_array (PTR_EXPR, PTR_TYPE, int) ;
  PTR_TYPE lookup_type (PTR_HASH) ;

  void chase_qual_index (PTR_EXPR, PTR_EXPR) ;
  void chase_qual_field (PTR_EXPR, PTR_SYMB) ;

  PTR_SYMB make_sym_entry (PTR_HASH, symbol_tag, PTR_TYPE, PTR_SCOPE, int) ;

  PTR_SYMB make_local_symbol (const char *, symbol_tag, PTR_TYPE, int) ;

  PTR_SYMB make_global_symbol (const char *, symbol_tag, PTR_TYPE, int) ;

  PTR_SYMB get_proc_symbol (PTR_HASH) ;

  PTR_SYMB make_constant (PTR_HASH entry, PTR_TYPE type) ;

  PTR_SYMB make_scalar (PTR_HASH entry, PTR_TYPE type, int kind) ;

  PTR_SYMB make_array (PTR_HASH entry, PTR_TYPE base_type, 
		       PTR_EXPR ranges, int ndim, int kind) ;

  PTR_SYMB make_pointer (PTR_HASH entry, PTR_TYPE type, int kind) ;

  PTR_SYMB make_function (PTR_HASH entry, PTR_TYPE type, int kind) ;

  PTR_SYMB make_procedure (PTR_HASH entry, int kind) ;

  PTR_SYMB make_external (PTR_HASH entry, PTR_TYPE type) ;

  PTR_SYMB make_program (PTR_HASH entry) ;

  PTR_SYMB make_common (PTR_HASH entry) ;

  PTR_SYMB make_derived_type (PTR_HASH entry, PTR_TYPE type, int kind) ;

  PTR_SYMB make_local_entity (PTR_HASH entry, symbol_tag, 
			      PTR_TYPE type, int kind) ;

  PTR_SYMB make_global_entity (PTR_HASH entry, symbol_tag, 
			       PTR_TYPE type, int kind) ;

  PTR_EXPR deal_with_func_call (PTR_EXPR fcall) ;

  PTR_EXPR deal_with_options (PTR_HASH name, PTR_TYPE type, 
			      int attributes, PTR_EXPR dims, 
			      int ndim, PTR_EXPR value, int decl_flag) ;
  void initialise () ;

  FortranContext (string& fn) 
                : ParseContext (fn, F77)
		, infname (fn)  {
    first_unresolved_call = last_unresolved_call = 0 ;
    thislabel = 0 ;
    maxdim = 8 ;
    vartype = 0 ;
    stmt_comments = cur_comment = 0 ;
    end_group = 0 ;
    cur_level = 0 ;
    undeftype = 0 ;
    lastwasbranch = thiswasbranch = NO ;
    type_var = 0 ;
    subscript_status = 0 ;
    attr_ndim = 0 ;
    position = 0 ;
    procclass = 0 ;
    do_name_err = 0 ;
    initialise () ;
  }
};


/*---------------------------------------------------------
CLASS
     FortranParser

     A parser for Fortran

OVERVIEW
    To be written.

DESCRIPTION
    To be written.
----------------------------------------------------------*/

class FortranParser : public FortranContext, FortranLexer {

private:

  int parser () ;           // produced by Bison or Yacc
  void stack_overflow (char*, short**, int, YYSTYPE**, int,
		       int*) ;
public:

  // GROUP: user functions

      //////////////////////////
      //
      // Calls parser and returns the number of parse errors.
      // If there are no errors, the root statement within
      // the ParseContext will be the root of the parse tree
      // created by this call.
      //
  int parse () { (void)parser () ; return nerrors() ; }


      //////////////////////////
      //
      // Constructor.
      //
      //  [in] string&  filename string for source file
      //  [in] int      optional value for lexeme text buffer
      //
  FortranParser (string& fn, int sz = YY_TEXT_SIZE) 
               : FortranLexer (fn, &yylineno), FortranContext (fn) {
    nowarnflag = YES ;
    inioctl = 0 ;
  }


  // GROUP: internal parser functions

      //////////////////////////
      //
  char *convic (long) ;

      //////////////////////////
      //
  long convci (int, const char *) ;

      //////////////////////////
      //
  void startioctl() { inioctl = 1 ; nioctl = 0 ; }

      //////////////////////////
      //
  void endioctl()   { inioctl = 0 ; }

      //////////////////////////
      //
  void newprog() ;

      //////////////////////////
      //
  void set_root_statement (PTR_STMT r) { 
    root_stmt = r ;
    pred_stmt = root_stmt ;
    start_new_group1 (root_stmt) ;
  }

      //////////////////////////
      //
  PTR_STMT make_stmt (statement_tag t, PTR_SYMB s, PTR_EXPR n1, PTR_EXPR n2, PTR_EXPR n3)
  {
    return ParseContext::make_stmt (t, s, n1, n2, n3, yylineno, yylineno-mod_offset) ;
  }
  
  
      //////////////////////////
      //
  PTR_STMT make_if (PTR_EXPR) ;

      //////////////////////////
      //
  PTR_STMT make_logif (PTR_STMT, PTR_STMT) ;

      //////////////////////////
      //
  PTR_STMT set_stat_list (PTR_STMT, PTR_STMT) ;

      //////////////////////////
      //
  PTR_STMT make_enddoall () ;

      //////////////////////////
      //
  PTR_STMT make_do (statement_tag, PTR_FTN_LABEL, PTR_SYMB, PTR_EXPR, PTR_EXPR, PTR_EXPR) ;

      //////////////////////////
      //
  PTR_STMT make_pardo (statement_tag, PTR_FTN_LABEL, PTR_SYMB, PTR_EXPR, PTR_EXPR, 
		       PTR_EXPR, PTR_EXPR) ;


      //////////////////////////
      //
  PTR_STMT make_elseif (PTR_EXPR) ;

      //////////////////////////
      //
  void make_else () ;

      //////////////////////////
      //
  void make_elsewhere () ;

      //////////////////////////
      //
  void make_section (PTR_EXPR, PTR_EXPR) ;

      //////////////////////////
      //
  void make_extend () ;

      //////////////////////////
      //
  void make_section_extend () ;

      //////////////////////////
      //
  void make_endblock () ;

      //////////////////////////
      //
  void make_endif () ;

      //////////////////////////
      //
  void make_endwhere () ;

      //////////////////////////
      //
  void make_endextend () ;

      //////////////////////////
      //
  void make_loop () ;
  

      //////////////////////////
      //
  PTR_SYMB proc_decl_init (char *, int) ;
  

      //////////////////////////
      //
  int chk_params (PTR_SYMB formal, PTR_EXPR actual) { return 1 ; }

      //////////////////////////
      //
  void bind () ;

      //////////////////////////
      //
  void late_bind_if_needed (PTR_EXPR) ;

      //////////////////////////
      //
  void make_prog_header () ;

      //////////////////////////
      //
  void enddcl () ;

      //////////////////////////
      //
  void process_interface (PTR_STMT) ;

      //////////////////////////
      //
  void set_blobs (PTR_STMT, PTR_STMT, int) ;


      //////////////////////////
      //
  ~FortranParser () { }

};

#endif // _ftnParser_h_

