/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// $Id: Scope.cc,v 1.5 1998/11/03 15:03:11 npt Exp $

#include <ctype.h>
#include "Octave/Base/Defs.h"
#include "Octave/Base/Scope.h"
#include "Octave/Base/ParseContext.h"
#include "Octave/Fortran/Tokdefs.h"  // need this for the operator tokens

// #define DEBUG_SCOPES
// #define TRACE_SYMTAB
// #define TRACE_ON

/* The 'struct operand' array is specific to Fortran */

struct operand
{
  int opval ;
  char *opname ;
} oplist [] = {
        {PLUS, "+"},
	{MINUS, "-"},
	{ASTER, "*"},
	{DASTER, "**"},
	{SLASH, "/"},
	{DSLASH, "//"},
	{AND, ".and."},
	{OR, ".or."},
	{XOR, ".xor."},
	{NOT, ".not."},
	{EQ, ".eq."},
	{NE, ".ne."},
	{GT, ".gt."},
	{GE, ".ge."},
	{NEQV, ".neqv."},
	{EQV, ".eqv."},
	{0, 0}
} ;

/*--------------------------------------------------------------------+
 |                                                                    |
 |                    ScopeNode::look_up_op                           |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_HASH ScopeNode::look_up_op (int op)
{
  struct operand *p;
  
  for (p = oplist; p->opname; p++)
    if (p->opval == op)
      return (look_up_ftn_sym (p->opname)) ;

  header->context->erri ("unknown operator (token = %d)", op) ;
  return 0 ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                    ScopeStack::cur_parent                          | 
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_STMT ScopeStack::cur_parent () {
  PTR_SCOPE cs = cur_scope () ;

  if (cs)
    return cs->head () ;
  else
    return 0 ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                  ScopeStack::convert_to_global                     | 
 |                                                                    |
 | The given symbol 's', which is declared at a non-global scoping    |
 | level, is moved so that its declaration is at the global level.    |
 |                                                                    |
 | The existing local declaration, if present, is removed.            |
 |                                                                    |
 +--------------------------------------------------------------------*/

void ScopeStack::convert_to_global (PTR_SYMB s) {
  PTR_SCOPE cur = s->scope ;
  PTR_HASH  global_entry ;
  char *sym_name = s->ident() ;

#ifdef TRACE_ON
  printf ("convert_to_global: %s\n", sym_name) ;
#endif

  // Check for a global SymbolTable entry under this symbol's name

  global_entry = global_scope()->just_look_up_ftn_sym_in_scope (sym_name) ;

  if (!global_entry)
  {
    // no existing global entry, so create one...

    global_entry = global_scope()->look_up_ftn (sym_name, HARD) ;
    global_entry->id_attr = s ;
    s->parent = global_entry ;
    s->scope = global_scope () ;
    global_scope()->install_entry (global_entry) ;
  }
  else
  {
    // check declaration of local symbol is consistent with
    // existing global declaration...eventually.
  }

  // Finally, delete the old local symbol table entry:
  // - this removes the entry at the original scope for 's'
  //   but leaves the original symbol and intact.

#ifdef TRACE_ON
  printf (" Map of scoped symbols BEFORE removal of symbol:\n") ;
  cur->declared.print () ;
#endif
  cur->remove_symbol (s) ;
#ifdef TRACE_ON
  printf (" Map of scoped symbols AFTER removal of symbol:\n") ;
  cur->declared.print () ;
#endif
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                 ScopeStack::push_scope_level                       |
 |                                                                    |
 | The given statement 's', create a new scope for 's and make it the |
 | current scope.                                                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

void ScopeStack::push_scope_level (PTR_STMT s, int use_scope) {
  PTR_SCOPE cscope = scope_stack [top_scope_level] ;
  char *n ;
  char buff [100] ;
  char no_name[1] = "" ;
  int args = 0 ;

  if (s->node.fields.symbol)
    n = s->node.fields.symbol->ident() ;
  else {
    n = no_name ;
  }
  
  switch (s->stmt_variant()) {
  case PROG_HEDR:  args = 1 ; sprintf (buff, "PROG(%s)", n)   ; break ;
  case PROC_HEDR:  args = 1 ; sprintf (buff, "PROC(%s)", n)   ; break ;
  case FUNC_HEDR:  args = 1 ; sprintf (buff, "FUNC(%s)", n)   ; break ;
  case GLOBAL:     args = 0 ; sprintf (buff, "GLOBAL(%s)", n) ; break ;
  case STRUCT_DECL:args = 0 ; sprintf (buff, "STRUCT(%s)", n) ; break ;
  case BASIC_BLOCK:args = 0 ; sprintf (buff, "BLOCK(%s)", n) ; break ;
  default:         args = 0 ; sprintf (buff, "scope(%s)", n)  ; break ;
  }
  
  if (s->scope == 0) {
    if (args && scope_stack[0]) {

      // create a new scope, based on existing argument scope

#ifdef DEBUG_SCOPES
      printf ("Creating scope %s from %s\n", buff, scope_stack[0]->name()) ;
#endif // DEBUG_SCOPES

      s->scope = new ScopeNode (*(scope_stack[0]), s, cscope, buff) ;

      // change scope of copied symbol table entries to new scope

      SymbolTable_Iterator next_entry (s->scope->declared) ;
      PTR_HASH h ;
      while (h = next_entry() )
      {
	h->id_attr->scope = s->scope ;
#ifdef DEBUG_SCOPES
	char *nm = h->identifier() ;
	PTR_SYMB s = h->symbol() ;
	
	if (!s)
	  printf ("\t%-15s%10s%6d%10d\n", nm, "<undef>", 0, 0) ;
	else
	  {
	    if (s->type)
	      printf ("\t%-15s%10d%6d%10d\n", nm, s->variant(), s->decl, s->type->variant()) ;
	    else
	      printf ("\t%-15s%10d%6d%10d\n", nm, s->variant(), s->decl, 0) ;
	  }
#endif // DEBUG_SCOPES
      }

      // delete the argument symbol table

      delete scope_stack [0] ;
      scope_stack [0] = 0 ;
    }
    else
      {
#ifdef DEBUG_SCOPES
	printf ("Creating scope %s from scratch\n", buff) ;
#endif // DEBUG_SCOPES
	s->scope = new ScopeNode (s, cscope, buff) ;
      }
  }

  if (use_scope) {
    scope_stack [top_scope_level+1] = scope_stack[top_scope_level] ;
    scope_stack [top_scope_level++] = s->scope ;
  }
  else
    scope_stack [++top_scope_level] = s->scope ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                  ScopeStack::pop_scope_level                       |
 |                                                                    |
 | The given scoping statement, remove all scopes that are beyond     |
 | the level defined by that statement, i.e. no longer visible.       |
 | Then, we pop the scope level defined by the given scoping statement|
 |                                                                    |
 +--------------------------------------------------------------------*/

void ScopeStack::pop_scope_level (PTR_STMT s) {

  // first pop scope levels until we reach the scope defined by 's'

  while ( top_scope_level > 0 && s != scope_stack[top_scope_level]->header)
    -- top_scope_level ;

  // now pop the given scope level

  if (scope_stack[top_scope_level]->header != s)
    s->context->error ("pop_scope_level: internal scoping problems") ;
  else
    -- top_scope_level ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                     ScopeStack::push_scope                         |
 |                                                                    |
 | The given ScopeNode 's', push that scope onto the scope stack and  |
 | make it the current scope.                                         |
 |                                                                    |
 +--------------------------------------------------------------------*/

void ScopeStack::push_scope (PTR_SCOPE s)
{
  scope_stack [++top_scope_level] = s ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                     ScopeStack::push_scope                         |
 |                                                                    |
 | Pop the current scope off the top of the scope stack.              |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SCOPE ScopeStack::pop_scope ()
{
  return scope_stack [top_scope_level--] ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  ScopeStack::push_arg_scope ()                     |
 |                                                                    |
 | Whilst parsing the argument list for a function declaration, the   |
 | argument symbols may need to be installed in a temporary scope.    |
 | This is certainly the case in the Fortran parser.                  |
 |                                                                    |
 | The push_arg_scope() routine saves the current scope, installs a   |
 | new scope as the current scope and sets this up as the argument    |
 | scope.                                                             |
 |                                                                    |
 |                                                                    |
 |                  ScopeStack::pop_arg_scope ()                      |
 |                                                                    |
 | The pop_arg_scope() routine reverses this process and reinstates   |
 | the preceding scope.                                               |
 |                                                                    |
 +--------------------------------------------------------------------*/

void ScopeStack::push_arg_scope () {
  old_level = top_scope_level ; 
  top_scope_level = 0 ;
  scope_stack [0] = new ScopeNode (0, 0, "arguments") ;
}

void ScopeStack::pop_normal_scope () {
  top_scope_level = old_level ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                       ScopeNode::dump                              |
 |                                                                    |
 | Print out the contents of this scope node. The contents is a symbol|
 | table, and if the recursive flag is set we print all enclosing     |
 | scopes as well.                                                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

void ScopeNode::dump (int recursive) {
  printf ("---------- Symbol Table for Scope '%s' ------------\n\n", name() ) ;
  printf ("%-15s%15s%6s%6s %s\n", "Name", "Category", "Decl", "Attr", "Type") ;
  SymbolTable_Iterator next_entry (declared) ;
  PTR_HASH h ;
  int items = declared.contents() ;
  int seen = 0 ;

  while (h = next_entry() )
  {
    ++ seen ;
    const char *nm = h->identifier() ;
    PTR_SYMB s = h->symbol() ;

    if (!s)
      printf ("%-15s%15s%6d%6d\n", nm, "<undef>", 0, 0) ;
    else
    {
      while (s)
	{
	  printf ("%-15s%15s%6d%6d", nm, enumToText(s->symb_variant()), s->decl, s->attr) ;
	  if (s->type)
	    {
	      char *ttext = s->type->sprintf () ;
	      printf (" %s\n", ttext) ;
	      if (ttext)
		delete [] ttext ;
	    }
	  else
	    printf (" <untyped>\n") ;
	  s = s->id_list ;
	}
    }
  }

  if (seen != items)
    printf ("\nCONSISTENCY ERROR: Map claims to hold %d items, but %d seen\n",
	    items, seen) ;

  if (recursive && enclosing)
  {
    printf ("\n ... next outer scope is '%s' ...\n\n", enclosing->name() ) ;
    enclosing->dump (recursive) ;
  }
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                       ScopeNode::sprint                            |
 |                                                                    |
 | Print the full name of this scope node in the given char buffer.   |
 | Note that the nesting is from inner to outer, rather than the      |
 | outer to inner normally expected in C++                            |
 |                                                                    |
 +--------------------------------------------------------------------*/

/* ERROR: each sprintf places a NULL byte at the end - strcpy instead */

void ScopeNode::sprint (char** buf, char* limit)
{
  char* sname = name() ;
  char lbuf[100] ;
  sname = sname ? sname : (char*) "<undefined>" ;
  int l = strlen (sname) ;
  int recurse ;

  if ((*buf) < limit) {
    sprintf (lbuf, "...::%s", name()) ;
    l += 5 ;
    *buf = (*buf) - l ;
    recurse = 0 ;
  }
  else {
    sprintf (lbuf, "::%s", name() ) ;
    l += 2 ;
    *buf = (*buf) - l ;
    recurse = enclosing != 0 ;
  }

  char* bp = *buf, *lp = lbuf ;
  for ( ; l ; l--)
    *bp++ = *lp++ ;

  if (recurse)
    enclosing->sprint (buf, limit) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                      ScopeNode::category                           |
 |                                                                    |
 | Returns a const char* string which describes the category of the   |
 | given scope.                                                       |
 |                                                                    |
 +--------------------------------------------------------------------*/

const char* ScopeNode::category ()
{
  if (!this)
    return "unknown" ;
  if (scope_symb)
    switch (scope_symb->symb_variant())
      {
      case NAMESPACE_NAME:
	return "namespace" ;
      case CLASS_NAME:
	return "class" ;
      case UNION_NAME:
	return "union" ;
      case STRUCT_NAME:
	return "struct" ;
      case TEMPLATE_NAME:
	return "template" ;
      case FUNCTION_NAME:
      case MEMBER_FUNC:
	return "function" ;
      default:
	return "unspecified" ;
      }
  if (header)
    return "local" ;
  else
    return "unknown" ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                       ScopeNode::find                              |
 |                                                                    |
 | Searches this scope node to find a given symbol. It may search     |
 | recursively through outer scope levels if the "recursive" flag     |
 | is set. It may try to match a particular symbol variant if the     |
 | "varnt" parameter is non-zero, and if the "create" flag is set,    |
 | it will create a new symbol table entry if the given symbol is     |
 | not found.                                                         |
 |                                                                    |
 +--------------------------------------------------------------------*/


PTR_HASH ScopeNode::find (const char *sym, int recursive, int varnt,
			  int create, int csens)
{
  PTR_HASH h ;
  char sbuf [256] ;

  if (csens == NOT_CASE_SENS)
    {
      char *to = sbuf ;
      const char *from = sym ;
      while (*from)
	*to++ = tolower (*from++) ;
      *to = '\0' ;
      sym = sbuf ;
    }

  h = declared.lookup ((char*)sym, varnt) ;

#ifdef TRACE_SYMTAB
  printf ("\nLooking for symbol '%s' in following scope nest...\n", sym) ;
  dump (1) ;
#endif

  if (h) {
#ifdef TRACE_SYMTAB
    printf ("..found %s at scope %s\n", sym, name()) ;
#endif
    return h ;
  }
  else if (recursive && enclosing) {
#ifdef TRACE_SYMTAB
    printf ("..symbol '%s', variant %d, not at scope %s, searching outer level\n", 
	    sym, varnt, name()) ;
#endif
    h = enclosing->find (sym, recursive, varnt, SEARCH_ONLY) ;
  }
  if (!h && create) {
    h = new hash_entry (sym) ;
#ifdef TRACE_SYMTAB
    printf ("..Created new hash entry 0x%08x (%s, %d) not found in scope %s\n", 
	    (unsigned long)h, sym, varnt, name()) ;
#endif
    // Note: newly declared entries are not installed in the
    // hash table for the current scope, as this may be the
    // wrong place for them. This is left to the symbol creation
    // routine, which must call install_entry (h).
  }
  return h ;
}

char* Scope_Kind::kstrings[] = { "class", "file", "local", "function" } ;
