/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// $Id: Sets.cc,v 1.4 1998/11/04 16:15:56 npt Exp $

// Source for library of set operations and set classes

#include <stdlib.h>   // for qsort()
#include <signal.h>   // for SIGINT and raise()
#include <string.h>   // for strcat() and strlen()

#define _LIBOBASE_

#include "Octave/Base.h"   // for all basic objects and methods


/*--------------------------------------------------------------------
 * 
 * Routine: void SetNode::free_thread (SetNode *this)
 *
 * Inputs:  `this\' is a pointer to a SetNode class.
 *
 * Outputs: none, but afterwards the parameter class is deleted.
 *
 * Actions: deletes the thread headed by node `this`.
 *
 *-------------------------------------------------------------------*/

void SetNode::free_thread ()
{
  if (this) {
    SetNode* t = this->thread ; 
    
    while (t)
    {
      SetNode* del = t ;
      t = t->thread ;
      delete del ;
    }
    delete this ;
  }
}


/*--------------------------------------------------------------------
 * 
 * Routine: void print_set_stats ()
 *
 * Inputs:  none
 * 
 * Outputs: textual output to file concerning usage of sets
 *
 * Actions: prints lots of information.
 *
 *-------------------------------------------------------------------*/
void print_set_stats()
{
  Set* s = 0 ;
   int last[20] = {0} ;
  int mem ;

  printf ("\nSETS & SET OPERATION STATISTICS\n\n") ;
  printf ("\tCreated\t(total)\tDeleted\t(total)\n") ;
  printf ("----------------------------------\n") ;
  printf ("Sets:\t%d\t(%d)\t%d\t(%d)\n", s->new_sets()-last[0], s->new_sets(), 
	  s->sets_freed()-last[1], s->sets_freed()) ;
  printf ("Nodes:\t%d\t(%d)\t%d\t(%d)\n\n", s->new_nodes()-last[2], s->new_nodes(), 
	  s->nodes_freed()-last[3], s->nodes_freed()) ;

  printf ("Numbers of set operations (totals)\n") ;
  printf ("----------------------------------\n") ;
  printf ("empty_set\t%d\t(%d)\n", _oct_nb_emptys-last[4],_oct_nb_emptys) ;
  printf ("universal set\t%d\t(%d)\n", _oct_nb_universals-last[5],_oct_nb_universals) ;
  printf ("union    \t%d\t(%d)\n", _oct_nb_unions-last[6],_oct_nb_unions) ;
  printf ("intersection\t%d\t(%d)\n", _oct_nb_intersections-last[7],_oct_nb_intersections) ;
  printf ("difference\t%d\t(%d)\n", _oct_nb_differences-last[8],_oct_nb_differences) ;
  printf ("equivalence\t%d\t(%d)\n", _oct_nb_equivalences-last[9],_oct_nb_equivalences) ;
  printf ("!equivalence\t%d\t(%d)\n", _oct_nb_non_equivalences-last[10],
	  _oct_nb_non_equivalences) ;
  printf ("copy set\t%d\t(%d)\n", _oct_nb_copys-last[11],_oct_nb_copys) ;
  printf ("is_empty_set\t%d\t(%d)\n", _oct_nb_empty_preds-last[12],_oct_nb_empty_preds) ;
  printf ("!is_empty_set\t%d\t(%d)\n", _oct_nb_non_empty_preds-last[13],
	  _oct_nb_non_empty_preds) ;
  printf ("is_universal\t%d\t(%d)\n", _oct_nb_universal_preds-last[14],
	  _oct_nb_universal_preds) ;
  printf ("is_member\t%d\t(%d)\n", _oct_nb_is_members-last[15], _oct_nb_is_members) ;
  printf ("add_member\t%d\t(%d)\n", _oct_nb_add_members-last[16], _oct_nb_add_members) ;
  printf ("remove_member\t%d\t(%d)\n", _oct_nb_remove_members-last[17], 
	  _oct_nb_remove_members) ;
  mem = (s->new_sets() - s->sets_freed()) * sizeof(class Set) +
        (s->new_nodes() - s->nodes_freed()) * sizeof(class SetNode) ;
  printf ("\n") ;
  printf ("Current memory allocated for sets = %dK\n", (mem+1023)/1024) ;

  last[0] = s->new_sets() ;
  last[1] = s->sets_freed() ;
  last[2] = s->new_nodes() ;
  last[3] = s->nodes_freed() ;
  last[4] = _oct_nb_emptys ;
  last[5] = _oct_nb_universals ;
  last[6] = _oct_nb_unions ;
  last[7] = _oct_nb_intersections ;
  last[8] = _oct_nb_differences ;
  last[9] = _oct_nb_equivalences;
  last[10] = _oct_nb_non_equivalences;
  last[11] = _oct_nb_copys ;
  last[12] = _oct_nb_empty_preds ;
  last[13] = _oct_nb_non_empty_preds ;
  last[14] = _oct_nb_universal_preds ;
  last[15] = _oct_nb_is_members ;
  last[16] = _oct_nb_add_members ;
  last[17] = _oct_nb_remove_members ;
}

/* -------------------- Non-template Set support functions ------------------- */



/*--------------------------------------------------------------------
 * 
 * Routine:  void copy_subset (subSet S1, subSet S2)
 *
 * Inputs:  subSet S1
 *
 * Outputs: none.
 *
 * Actions: copies the elements of subSet S1 to subSet S2.
 *
 *-------------------------------------------------------------------*/


void copy_subset (subSet S1, subSet S2)
{
   int i;
   
   for (i = 0; i < WORDS_PER_SET; i++)
      S2[i] = S1[i];
}


/*--------------------------------------------------------------------
 * 
 * Routine:  void subset_union (subSet S, subSet S1, 
 *                                     subSet S2 )
 *
 * Inputs:  subSets S1 and S2
 *
 * Outputs: subSet S
 *
 * Actions: S becomes S1 Union S2
 *
 *-------------------------------------------------------------------*/


void subset_union (subSet S, subSet S1, subSet S2 )
{
   int i;
   
   for (i = 0; i < WORDS_PER_SET; i++)
      S[i] = S1[i] | S2[i];
}


/*--------------------------------------------------------------------
 * 
 * Routine:  int subset_intersection (subSet S, subSet S1, 
 *                                                 subSet S2)
 *
 * Inputs:  subSets S1 and S2
 *
 * Outputs: subSet S
 *
 * Actions: S becomes S1 Intersection S2
 *
 *-------------------------------------------------------------------*/


int subset_intersection (subSet S, subSet S1, subSet S2)
{
   int i, t = false;
   
   for (i = 0; i < WORDS_PER_SET; i++) {
      S[i] = S1[i] & S2[i];
      t = t || (S[i] != 0) ;
    }
   return t ;  // return S != empty set
}


/*--------------------------------------------------------------------
 * 
 * Routine:  int subset_difference (subSet S, subSet S1, 
 *                                               subSet S2)
 *
 * Inputs:  subSets S1 and S2
 *
 * Outputs: subSet S
 *
 * Actions: S becomes S1 with all elements in S2 removed
 *
 *-------------------------------------------------------------------*/


int subset_difference (subSet S, subSet S1, subSet S2)
{
  int i, t = false;
  
  for (i = 0; i < WORDS_PER_SET; i++) {
    S[i] = S1[i] & (S2[i] ^ ~0);
    t = t || (S[i] != 0) ;
  }
  return t ;  // return S != empty set
}


/*--------------------------------------------------------------------
 * 
 * Routine: int subset_equivalence (subSet S1, subSet S2)
 *
 * Inputs:  two subsets S1 and S2
 *
 * Outputs: TRUE or FALSE
 *
 * Actions: return TRUE if S1 and S2 contain identical elements.
 *
 *-------------------------------------------------------------------*/


int subset_equivalence (subSet S1, subSet S2)
{
   int i ;
   
   for (i = 0; i < WORDS_PER_SET; i++)
     if (S1[i] != S2[i])
       return false ;
   return true;
}


/*--------------------------------------------------------------------
 * 
 * Routine:  int intcompare (const void *i, const void *j)
 *
 * Inputs:  pointers to two integers.
 *
 * Outputs: neg, pos, or zero depending on order of *i and *j
 *
 * Actions: This is used by the stdlib.h routine 'qsort'.
 *
 *-------------------------------------------------------------------*/


int intcompare (const void *i, const void *j)
{
  return (*(int *)i - *(int *)j) ;
}

/*--------------------------------------------------------------------
 * 
 * Routine:  void make_subset (subSet s, int lb, int ub)
 *
 * Inputs:  a subset 's' and two integers, lb and ub.
 *
 * Entry Conditions:
 *
 *          (lb >= 0) and (ub < CARDINALITY)
 *
 * Outputs: void, but modifies s.
 *
 * Actions: assigns the subset {lb,...,ub} to subset s.
 *
 *-------------------------------------------------------------------*/

#define ALL_BITS_FROM(_l_) (((unsigned)(-1) << (_l_)))


void make_subset (subSet s, int lb, int ub)
{
  int w ;
  
  if ((lb < 0) || (ub >= CARDINALITY) || (ub < lb))
    FATAL_ERROR ("invalid range for 'make_subset'") ;
  
  for (w = 0; w < WORDS_PER_SET; w++)
    {
      if (w == lb/BITS_PER_WORD)
	s[w] = ALL_BITS_FROM((lb % BITS_PER_WORD)) ;
      else {
	if (w < lb/BITS_PER_WORD)
	  s[w] = 0 ;
	else
	  s[w] = (unsigned)(-1) ;
      }
      
      if (w == ub/BITS_PER_WORD) {
	int um = ub % BITS_PER_WORD ;
	if (um != BITS_PER_WORD-1)
	  s[w] = s[w] ^ ALL_BITS_FROM(um);
      }
      else {
	if (w > ub/BITS_PER_WORD)
	  s[w] = 0 ;
      }
    }
}

/*-- disable this temporarily ---

void SetOf<T>::analyze() 
{
  int histo[HASH_SET_SIZE] = {0} ;
  int nodes = 0, b ;

  for (b = 0; b < HASH_SET_SIZE; b++) {
    SetNode* p = hash[b] ;
    while (p) {
      ++histo[b] ;
      ++nodes ;
      p = p->link ;
    }
  }
  printf ("Total %d nodes\n", nodes) ;
  for (b = 0; b < HASH_SET_SIZE; b++)
    printf ("  length(bin %d) = %d\n", b, histo[b]) ;
}

*/
