/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// $Id: Statements.cc,v 1.6 1999/09/14 14:01:12 npt Exp $

#include "Octave/Base/ParseContext.h"

// most Statement-related routines are currently declared inline.

Statement::~Statement ()
{
  // must delete the structures which are only reachable from this
  // node, i.e. the control lists, possibly the universe of labels,
  // and the scope if defined.
  
  if ( (stmt_variant() == PROC_HEDR) ||
       (stmt_variant() == FUNC_HEDR) ||
       (stmt_variant() == PROG_HEDR) ||
       (stmt_variant() == MODULE_STMT) )
    delete node.procedure.labels ;
  
  delete node.fields.bl_ptr1 ;
  delete node.fields.bl_ptr2 ;

  if (scope != 0)
  {
    scope->zap () ;
    delete scope ;
  }
}

void Statement::set_fields (Statement& stat)
{
  stmt_variant (stat.stmt_variant()) ;
  g_line = stat.g_line ;
  l_line = stat.l_line ;
  if (context->OctLanguage() == F77)
    label.ftn = new FtnLabel (*(stat.label.ftn)) ;
  else
    label.cc = new CcLabel (*(stat.label.cc)) ;

  node.fields.symbol = stat.node.fields.symbol ;
  
  if (stat.node.fields.e1)
    node.fields.e1 = stat.context->make_expr (*(stat.node.fields.e1)) ;
  if (stat.node.fields.e2)
    node.fields.e1 = stat.context->make_expr (*(stat.node.fields.e2)) ;
  if (stat.node.fields.e3)
    node.fields.e1 = stat.context->make_expr (*(stat.node.fields.e3)) ;

  if (stat.node.fields.bl_ptr1)
    node.fields.bl_ptr1 = &stat.node.fields.bl_ptr1->copy(this) ;

  if (stat.node.fields.bl_ptr2)
    node.fields.bl_ptr2 = &stat.node.fields.bl_ptr2->copy(this) ;
}

int Statement::gotoLabel ()
{
  if (!this)
    return 0 ;
  else
  {
    PTR_EXPR label_ref = expr3() ;
    PTR_FTN_LABEL label_node = label_ref ? label_ref->label() : 0 ;
    if (label_node)
      return label_node->value() ;
    else
      return 0 ;
  }
}

void StmtListIterator::goto_first_exec ()
{
  rewind () ;

  PTR_STMT s = head() ;

  while (s && !(s->is_executable()))
    {
      (*this)() ; // advances the iterator by one list item
      s = head() ;
    }
}

void StmtListIterator::goto_declarations ()
{
  rewind () ;

  PTR_STMT s = head() ;

  while (s && !(s->post_declaration()))
    {
      (*this)() ; // advances the iterator by one list item
      s = head() ;
    }
}

int StmtListIterator::split_loop_nest (PTR_STMT S)
{
  if (!this || !S) return 0 ;
  PTR_STMT outer = head() ;
  if (!outer || outer->stmt_variant() != FOR_NODE)
    return 0 ;
  
  PTR_STMT inner = S->control_parent ;

  if (inner != outer && !outer->is_ancestor_of(inner))
    return 0 ;
  
  if (!outer->is_perfectly_nested_upto(inner))
    return 0 ;

  PTR_CNTXT context = S->context ;

  outer->enforce_enddo () ;

  StmtListIterator inner_body (inner->loop_body()) ;
  inner_body.goto_node (S) ;
  inner_body.set_mark () ;
  inner_body.goto_end () ;
  StmtList* L2 = inner_body.cut () ;
  
  PTR_STMT newnest = context->make_stmt (*S) ;
  PTR_STMT newinner = newnest ;

  for (PTR_STMT lp = outer; lp != inner; )
    {
      lp = lp->loop_body()->hd() ;
      newinner = newinner->loop_body()->hd() ;
    }

  StmtListIterator newbody (newinner->loop_body()) ;
  newbody.goto_end () ;
  StmtList* junk = newbody.cut () ;
  delete junk ;
  newbody.paste (*L2) ;
  return 1 ;
}

int Statement::is_perfectly_nested_upto (PTR_STMT S)
{
  if (!this || !S)
    return 1 ;

  if (this->stmt_variant() == FOR_NODE)
    if (this == S)
      return 1 ;
    else
      return (this->body_size() == 1 && 
	      this->loop_body()->hd()->is_perfectly_nested_upto (S)) ;
  else
    return (this == S) ;
}

int Statement::is_perfectly_nested ()
{
  if (!this)
    return 0 ;

  if (this->stmt_variant() == FOR_NODE)
    {
      if (this->body_size() == 1)
	return 1 + this->loop_body()->hd()->is_perfectly_nested() ;
      else
	return 1 ;
    }
  else
    return 0 ;
}

int Statement::is_do_end_loop ()
{
  if (!this || this->stmt_variant() != FOR_NODE)
    return 0 ;
  else
    {
      PTR_STMT last = node.fields.bl_ptr1->tail() ;
      if (last && last->stmt_variant() == CONTROL_END)
	return 1 ;
      else
	return 0 ;
    }
}

int Statement::body_size ()
{
  if (!this) return 0 ;

  StmtListIterator next_stmt (this->loop_body()) ;
  PTR_STMT stmt ;
  int count = 0 ;

  while ( (stmt = next_stmt()) && stmt->stmt_variant() != CONTROL_END)
    if (stmt->is_executable())
      ++count ;
  return count ;
}

int Statement::is_ancestor_of (PTR_STMT S)
{
  if (!this || !S) return 0 ;

  for (PTR_STMT p = S; p; p = p->control_parent)
    if (p == this)
      return 1 ;

  return 0 ;
}

void Statement::enforce_enddo ()
{
  if (!this) return ;

  // First remove any reference to the label of the ending CONTINUE

  if (this->stmt_variant() == FOR_NODE)
    this->node.fields.lbl_ptr = 0 ;

  PTR_BLOCK s1 = this->then_part() ;
  PTR_BLOCK s2 = this->else_part() ;

  if (s1)
    {
      StmtListIterator next_child (s1) ;
      PTR_STMT stmt, laststmt = 0 ;
      
      while (stmt = next_child())
	{
	  stmt->enforce_enddo () ;
	  laststmt = stmt ;
	}
      if (this->stmt_variant() == FOR_NODE) {
	statement_tag stag = 
	  laststmt ? laststmt->stmt_variant() : (statement_tag) 0 ;
	if (stag != CONTROL_END) {
	  Statement *ce = context->make_stmt(CONTROL_END, 0, 0, 0, 0) ;
	  ce->control_parent = this ;
	  s1->append (ce) ;
	}
      }
    }
  
  if (s2)
    {
      StmtListIterator next_child (s2) ;
      PTR_STMT stmt ;
      
      while (stmt = next_child())
	stmt->enforce_enddo () ;
    }
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |                      Statement::CcUnparse                           |
 |                                                                     |
 +---------------------------------------------------------------------*/

void Statement::CcUnparse (SourceListing& src)
{
  if (stmt_variant() == GLOBAL)
    node.fields.bl_ptr1->CcUnparseStmtList (src) ;
  else {
    CcUnparseStatement (src) ;
  }
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                void CcUnparseStatement                              |
 |                                                                     |
 +---------------------------------------------------------------------*/

void Statement::CcUnparseStatement (SourceListing& src) 
{ 
  if (!this)
    return ;

  int tabs = CcIndentation () ;
  PTR_CC_LABEL lab = label.cc ;

  switch (stmt_variant ())
    {
    case BASIC_BLOCK:
      node.fields.bl_ptr1->CcUnparseCompoundStmt (src, tabs, this) ;
      break ;

    case FUNC_HEDR:
      src.CcStartLine (tabs, this) ;
      node.fields.symbol->CcUnparseFunDeclarator (src) ;
      src << '\n'  ;
      src.appendBuffer () ;
      if (node.function.format_group)
	node.function.format_group->CcUnparseStmtList (src) ;
      node.function.control->CcUnparseCompoundStmt (src, 0, this) ;
      break ;

    case VAR_DECL:
      src.CcStartLine (tabs, this) ;
      node.fields.e1->CcUnparseExpression (src, 0) ; /* decl_specs */
      if (node.fields.e2)
	{
	  src << " " ;
	  node.fields.e2->CcUnparseExpression (src, 0) ; /* init_declarators */
	}
      src << " ;" ;
      src << '\n' ;
      src.appendBuffer () ;
      break ;

    case EXPR_STMT_NODE:
      src.CcStartLine (tabs, this, lab) ;
      node.fields.e1->CcUnparseExpression (src, 0) ;
      src << " ;" ;
      src << '\n' ;
      src.appendBuffer () ;
      break ;

    case FOR_NODE:
      src.CcStartLine (tabs, this, lab) ;
      src << "for (" ;
      node.fields.e1->CcUnparseExpression (src, 0) ;
      src << "; " ;
      node.fields.e2->CcUnparseExpression (src, 0) ;
      src << "; " ;
      node.fields.e3->CcUnparseExpression (src, 0) ;
      src << ')' ;
      src << '\n' ;
      src.appendBuffer () ;
      node.fields.bf_ptr1->CcUnparseStatement (src) ;
      break ;

    case WHILE_NODE:
      src.CcStartLine (tabs, this, lab) ;
      src << "while (" ;
      node.fields.e1->CcUnparseExpression (src, 0) ;
      src << ')' ;
      src << '\n' ;
      src.appendBuffer () ;
      node.fields.bf_ptr1->CcUnparseStatement (src) ;
      break ;

    case DO_WHILE_NODE:
      src.CcStartLine (tabs, this, lab) ;
      src << "do" ;
      src << '\n' ;
      src.appendBuffer () ;
      node.fields.bf_ptr1->CcUnparseStatement (src) ;
      src.CcStartLine (tabs, this) ;
      src << "while (" ;
      node.fields.e1->CcUnparseExpression (src, 0) ;
      src << ") ;" ;
      src << '\n' ;
      src.appendBuffer () ;
      break ;

    case IF_NODE: 
      src.CcStartLine (tabs, this, lab) ;
      src << "if (" ;
      node.fields.e1->CcUnparseExpression (src, 0) ;
      src << ')' ;
      src << '\n' ;
      src.appendBuffer () ;
      node.if_node.then_block->CcUnparseStatement (src) ;
      if (node.if_node.else_block)
	{
	  src.CcStartLine (tabs, this) ;
	  src << "else" ;
	  src << '\n' ;
	  src.appendBuffer () ;
	  node.if_node.else_block->CcUnparseStatement (src) ;
	}
      break ;

    case SWITCH_NODE:
      src.CcStartLine (tabs, this, lab) ;
      src << "switch (" ;
      node.switch_stat.switch_expr->CcUnparseExpression (src, 0) ;
      src << ')' ;
      src << '\n' ;
      src.appendBuffer () ;
      node.switch_stat.body->CcUnparseStatement (src) ;
      break ;

    case CASE_NODE:
      src.CcStartLine (tabs, this, lab) ;
      src << "case " ;
      node.fields.e1->CcUnparseExpression (src, 0) ;
      src << " :" ; 
      src << '\n' ;
      src.appendBuffer () ;
      node.fields.bf_ptr1->CcUnparseStatement (src) ;
      break ;

    case DEFAULT_NODE:
      src.CcStartLine (tabs, this, lab) ;
      src << "default :" ;
      src << '\n' ;
      src.appendBuffer () ;
      node.fields.bf_ptr1->CcUnparseStatement (src) ;
      break ;

    case BREAK_NODE:
      src.CcStartLine (tabs, this, lab) ;
      src << "break ;" ;
      src << '\n' ;
      src.appendBuffer () ;
      break ;

    case CONTINUE_NODE:
      src.CcStartLine (tabs, this, lab) ;
      src << "continue ;" ;
      src << '\n' ;
      src.appendBuffer () ;
      break ;

    case RETURN_NODE:
      src.CcStartLine (tabs, this, lab) ;
      if (node.fields.e1)
	{
	  src << "return " ;
	  node.fields.e1->CcUnparseExpression (src, 0) ;
	  src << " ;" ;
	  src << '\n' ;
	}
      else
	{
	  src << "return ;" ;
	  src << '\n' ;
	}
      src.appendBuffer () ;
      break ;

    case GOTO_NODE:
      {
	src.CcStartLine (tabs, this, lab) ;
	src << "goto " ;
	PTR_STMT target = node.fields.bf_ptr1 ;
	if (target)
	  src << target->label.cc->index() ;
	else
	  src << "/* label undefined */ " ;
	src << " ;" ;
	src << '\n' ;
	src.appendBuffer () ;
	break ;
      }

    default:
      src.CcStartLine (tabs, this, lab) ;
      src << " /* unknown statement encountered */" ;
      src << '\n' ;
      src.appendBuffer () ;
      break ;
    }
}


StmtList& StmtList::copy (PTR_STMT S)
{
  StmtList* nsl = new StmtList () ;

  if (!S || !this) return *nsl ;

  StmtListIterator next_stmt (*this) ;
  StmtListIterator theCopy (*nsl) ;
  PTR_STMT s ;

  while (s = next_stmt())
    {
      PTR_STMT ns = S->context->make_stmt (*s) ; 
      theCopy.insert (ns, S) ;
    }
  return *nsl ;
}

void StmtList::set_block_parent (PTR_STMT parent)
{
  if (!this) return ;

  StmtListIterator next_child (*this) ;
  PTR_STMT child ;

  while (child = next_child())
    child->control_parent = parent ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                 void CcUnparseStmtList                              |
 |                                                                     |
 +---------------------------------------------------------------------*/

void StmtList::CcUnparseStmtList (SourceListing& src) 
{
  if (!this)
    return ;

  StmtListIterator next_stmt (this) ;
  PTR_STMT stmt ;

  while ( stmt = next_stmt() )
    stmt->CcUnparseStatement (src) ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |              void CcUnparseCompoundStmt                             |
 |                                                                     |
 +---------------------------------------------------------------------*/

void StmtList::CcUnparseCompoundStmt (SourceListing& src, int tabs,
				      PTR_STMT parent) 
{ 
  src.CcStartLine (tabs, parent) ;
  src << '{' ;
  src << '\n' ;
  src.appendBuffer() ;
  CcUnparseStmtList (src) ;
  src.CcStartLine (tabs, parent) ;
  src << '}' ;
  src << '\n' ;
  src.appendBuffer () ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                   void  CcIndentation                               |
 |                                                                     |
 +---------------------------------------------------------------------*/

int Statement::CcIndentation ()
{
  int count = -1 ;
  PTR_STMT bf = this ;
  for ( ; bf && bf->stmt_variant () != GLOBAL; bf = bf->control_parent)
    count++;

  return count ;
}

PTR_STMT Statement::nextInnerLoop ()
{
  PTR_BLOCK b = node.fields.bl_ptr1 ;
  if (!b)
    return 0 ;
  
  StmtListIterator next_stmt (b) ;
  PTR_STMT s ;

  while (s = next_stmt())
    if (s->stmt_variant() == FOR_NODE)
      return s ;
  return 0 ;
}

PTR_STMT Statement::nextOuterLoop ()
{
  if (control_parent && control_parent->stmt_variant() != FOR_NODE)
    return control_parent->nextOuterLoop() ;
  else
    return control_parent ;
}
