/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) Nigel Topham and University of Edinburgh, 1996.      |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               | 
+--------------------------------------------------------------------*/
// $Id: FtnBuilder.cc,v 1.12 2000/01/07 12:44:45 fgg Exp $

#include <ctype.h>
#include "Octave/Fortran.h"

// Fortran intrinsic functions are defined by the following structure
// which provides a mechanism for performing type consistency checks
// on the arguments, as well as permitting up to four argument type
// signatures per overloaded generic function name.

struct Intrinsic_signature {
  int returnType ;            // type variant of function result
  int arg1, arg2 ;            // list of argument type tags for this signature (max 2)
  // arg1 = type tag of all args if numArgs == 0
  // (arg2 == 0) => no second argument for this particular signature (e.g. 'cmplx')
};

struct Intrinsic_func {
  char *funcName ;              // name of this intrinsic function
  int numArgs ;                 //  0 => arg count from minArgs to infinity
                                //  n => arg count from minArgs to n
  int minArgs ;                 // minimum number of arguments
  int nforms ;                  // number of different legal argument signatures (1..4)
  Intrinsic_signature sig [4] ; // list of legal type signatures for this function
};

// Declarations of static function prototypes.

static void initIntrinsics (FortranContext *) ;
static bool matchIntrinsic (const char*, Intrinsic_func **) ;

// Declaration of static initalised data.

static Intrinsic_func *funstart [27], *funend [27] ;

static Intrinsic_func intrinsics [] = {

  {// funcName, numArgs, minArgs, nforms,
    "abs",      1,        1,      4,
    {
      {T_INT,     T_INT,     0},
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {T_COMPLEX, T_COMPLEX, 0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "acos",     1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "aimag",    1,       1,       1,
    {
      {T_FLOAT,   T_COMPLEX,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "aint",     1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "alog10",   1,       1,       1,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "alog",     1,       1,       1,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "amax0",    0,       2,       1,
    {
      {T_FLOAT,   T_INT,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "amax1",    0,       2,       1,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "amin0",    0,       2,       1,
    {
      {T_FLOAT,   T_INT,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "amin1",    0,       2,       1,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "amod",     2,       2,       1,
    {
      {T_FLOAT,   T_FLOAT,   T_FLOAT},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "anint",    1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "asin",     1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "atan2",    2,       2,       2,
    {
      {T_FLOAT,   T_FLOAT,   T_FLOAT},
      {T_DOUBLE,  T_DOUBLE,  T_DOUBLE},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "atan",     1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "cabs",     1,       1,       1,
    {
      {T_FLOAT,   T_COMPLEX, 0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "ccos",     1,       1,       1,
    {
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "cexp",     1,       1,       1,
    {
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "char",     1,       1,       1,
    {
      {T_STRING,  T_INT,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "clog",     1,       1,       1,
    {
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "cmplx",    2,       1,       4,
    {
      {T_COMPLEX, T_INT,     T_INT},
      {T_COMPLEX, T_FLOAT,   T_FLOAT},
      {T_COMPLEX, T_DOUBLE,  T_DOUBLE},
      {T_COMPLEX, T_COMPLEX, 0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "conjg",    1,       1,       1,
    {
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "cosh",     1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "cos",      1,       1,       3,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "csin",     1,       1,       1,
    {
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "csqrt",     1,       1,       1,
    {
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dabs",     1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dacos",    1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dasin",    1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "datan2",   2,       2,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "datan",    1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dble",     1,       1,       4,
    {
      {T_DOUBLE,  T_INT,     0},
      {T_DOUBLE,  T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {T_DOUBLE,  T_COMPLEX, 0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dconjg",     1,       1,       1,
    {
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dcosh",    1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dcos",     1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "ddim",     2,       2,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  T_DOUBLE},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dexp",     1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dim",      2,       2,       3,
    {
      {T_INT,     T_INT,     T_INT},
      {T_FLOAT,   T_FLOAT,   T_FLOAT},
      {T_DOUBLE,  T_DOUBLE,  T_DOUBLE},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dimag",    1,       1,       1,
    {
      {T_DOUBLE,   T_COMPLEX,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dint",     1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dlog10",   1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dlog",     1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dmax1",    0,       2,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dmin1",    0,       2,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dmod",     2,       2,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  T_DOUBLE},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dnint",    1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dprod",    2,       2,       1,
    {
      {T_DOUBLE,  T_FLOAT,   T_FLOAT},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dsign",    2,       2,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  T_DOUBLE},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dsinh",    1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dsin",     1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dsqrt",    1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dtanh",    1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "dtan",     1,       1,       1,
    {
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "exp",      1,       1,       3,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "float",    1,       1,       1,
    {
      {T_FLOAT,   T_INT,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "iabs",     1,       1,       1,
    {
      {T_INT,     T_INT,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "ichar",    1,       1,       1,
    {
      {T_INT,     T_STRING,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "idim",     2,       2,       1,
    {
      {T_INT,     T_INT,     T_INT},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "idint",    1,       1,       1,
    {
      {T_INT,     T_DOUBLE,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "idnint",   1,       1,       1,
    {
      {T_INT,     T_DOUBLE,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "index",    2,       2,       1,
    {
      {T_INT,     T_STRING,  T_STRING},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "int",      1,       1,       4,
    {
      {T_INT, T_INT,         0},
      {T_INT, T_FLOAT,       0},
      {T_INT, T_DOUBLE,      0},
      {T_INT, T_COMPLEX,     0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "isign",    2,       2,       1,
    {
      {T_INT,     T_INT,     T_INT},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "len",      1,       1,       1,
    {
      {T_INT,     T_STRING,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "lge",      2,       2,       1,
    {
      {T_BOOL,    T_STRING,  T_STRING},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "lgt",      2,       2,       1,
    {
      {T_BOOL,    T_STRING,  T_STRING},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "lle",      2,       2,       1,
    {
      {T_BOOL,    T_STRING,  T_STRING},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "llt",      2,       2,       1,
    {
      {T_BOOL,    T_STRING,  T_STRING},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "log10",    1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "log",      1,       1,       3,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "max0",     0,       2,       1,
    {
      {T_INT,     T_INT,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "max1",     0,       2,       1,
    {
      {T_INT,     T_FLOAT,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "max",      0,       2,       3,
    {
      {T_INT,     T_INT,     0},
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "min0",     0,       2,       1,
    {
      {T_INT,     T_INT,     0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "min1",     0,       2,       1,
    {
      {T_INT,     T_FLOAT,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "min",      0,       2,       3,
    {
      {T_INT,     T_INT,     0},
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "mod",      2,       2,       3,
    {
      {T_INT,     T_INT,     T_INT},
      {T_FLOAT,   T_FLOAT,   T_FLOAT},
      {T_DOUBLE,  T_DOUBLE,  T_DOUBLE},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "nint",     1,       1,       1,
    {
      {T_INT,     T_FLOAT,   0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "real",     1,       1,       4,
    {
      {T_FLOAT,   T_INT,     0},
      {T_FLOAT,   T_FLOAT,   0},
      {T_FLOAT,   T_DOUBLE,  0},
      {T_FLOAT,   T_COMPLEX, 0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "sign",     2,       2,       3,
    {
      {T_INT,     T_INT,     T_INT},
      {T_FLOAT,   T_FLOAT,   T_FLOAT},
      {T_DOUBLE,  T_DOUBLE,  T_DOUBLE},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "sinh",     1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "sin",      1,       1,       3,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "sngl",     1,       1,       1,
    {
      {T_FLOAT,   T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "sqrt",     1,       1,       3,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {T_COMPLEX, T_COMPLEX, 0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "tanh",     1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// funcName, numArgs, minArgs, nforms,
    "tan",      1,       1,       2,
    {
      {T_FLOAT,   T_FLOAT,   0},
      {T_DOUBLE,  T_DOUBLE,  0},
      {0,         0,         0},
      {0,         0,         0}
    }
  },

  {// Termination record - not used.
    0,       0,       0,       0,
    {
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0},
      {0,         0,         0}
    }
  } 
};

/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::terminate_block                     |
 |                                                                    |
 | The given (optional) statement is appended to the current group    |
 | and, if necessary, the current group will be terminated with a     |
 | CONTROL_END node.                                                  |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::terminate_block (PTR_STMT p) {
  PTR_STMT t ;
  if (p) {
    t = p ;
    append_statement (p) ;
  }
  else
    t = last_in_block () ;
  
  if (!t || t->variant() != CONTROL_END )
    append_statement (make_stmt(CONTROL_END, 0, 0, 0, 0)) ;
}

void FortranContext::terminate_block () {
  PTR_STMT t = last_in_block () ;
  
  if (!t || t->variant() != CONTROL_END )
    append_statement (make_stmt(CONTROL_END, 0, 0, 0, 0)) ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::append_statement                    |
 |                                                                    |
 | The given statement is appended to the current block.              |
 |                                                                    |
 | If necessary, a new statement list will be created for the         |
 | current block if it is uninitialised.                              |
 |                                                                    |
 | Any comment list pointed to by 'stmt_comments' is attached to the  |
 | statement before it is included in the current statement list.     |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::append_statement (PTR_STMT s)
{
  PTR_BLOCK *bp = current_block () ;
  
  if ( ! (*bp) )
    *bp = new StmtList ;
  
  (*bp)->append (s) ;
  s->node.fields.cmnt_ptr = stmt_comments ;
  stmt_comments = cur_comment = 0 ;
  s->control_parent = pred_stmt ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranContext::exit_group                       |
 |                                                                    |
 | Causes the parse tree insertion point to be re-positioned under    |
 | the assumption that the current group is complete after insertion  |
 | of statement 'p' (which may be null)                               |
 |                                                                    |
 | If necessary, the current group will be terminated with a          |
 | CONTROL_END node.                                                  |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::exit_group (PTR_STMT p) {
  if ((pred_stmt == 0) || (pred_stmt->control_parent == 0))
    error ("misplaced nest-level ending") ;
  else {
    terminate_block (p) ;

    if ((pred_stmt->variant() == PDO_NODE) && 
	(pred_stmt->idx == EXTEND_NODE))
      blocks[cur_level] = &(pred_stmt->node.fields.bl_ptr2) ;
    else {
      pop_level () ;
      if (pred_stmt->variant() == GLOBAL)
	parstate = OUTSIDE ;
    }
  }
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                      FortranContext::setimpl                       |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::setimpl (PTR_TYPE type , int c1, int c2, 
			      PTR_SCOPE scope)
{
  int i;
  char buff[100];
  PTR_SYMB ss, scope_sym ;
  
  if (c1 == 0 || c2 == 0)
    return;
  
  if (c1 > c2) {
    (void)sprintf (buff, "illegal implicit range: %c-%c", c1, c2);
    error (buff);
  } 
  else {
    for (i = tolower(c1); i <= tolower(c2); ++i) {
      impltype[i - 'a'] = type;
    }
  }
  
  // alter implicitly-typed symbols that are already in scope
  
  if (scope)
    scope->set_implicit (impltype) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |               FortranContext::create_procedure                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::create_procedure (PTR_STMT s) {
  procinit () ;
  
  if (s && !s->node.procedure.labels) {
    s->node.procedure.labels = new Universe<FtnLabel> ;
    set_current_proc (s) ;
  }

  push_scope_level (s, NO) ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                      FortranContext::procinit                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::procinit ()
{
  nwarn = 0 ;
  thislabel = 0 ;
  
  if (undeftype)
    setimpl (global_default, 'a', 'z', 0) ;
  else {
    setimpl (global_float, 'a', 'z', 0) ;
    setimpl (global_int, 'i', 'n', 0) ;
  }
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                       FortranParser::make_if                       |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_STMT FortranParser::make_if (PTR_EXPR expr)
{
  PTR_STMT p;
  
  if (pred_stmt->variant() == GLOBAL)
    make_prog_header () ;
  
  if (expr->base_type() != T_BOOL) {
    err ("IF condition was not boolean");
    expr = 0;
  }
  p = make_stmt (IF_NODE, 0, expr, 0, 0);
  return p ;
}


/*--------------------------------------------------------------------+
  |                                                                    |
  |                     FortranParser::make_elseif                     |
  |                                                                    |
  +--------------------------------------------------------------------*/

PTR_STMT FortranParser::make_elseif (PTR_EXPR expr)
{
  PTR_STMT p = pred_stmt ;
  
  if (expr->base_type() != T_BOOL) {
    error ("ELSE IF condition was not boolean") ;
    expr = 0 ;
  }
  if (p->stmt_variant() == IF_NODE || p->stmt_variant() == ELSEIF_NODE) {
    terminate_block () ;
    pop_level () ;
    start_new_group2 (p) ;
    p = make_stmt (ELSEIF_NODE, 0, expr, 0, 0);
  }
  else
    error ("misplaced ELSE IF") ;
  
  return p ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                      FortranParser::make_else                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_else ()
{
  PTR_STMT p = pred_stmt ;
  
  if (p->variant() != IF_NODE && 
      p->variant() != ELSEIF_NODE)
    error ("misplaced ELSE") ;
  else {
    terminate_block () ;
    pop_level () ;
    start_new_group2 (p) ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranParser::make_elsewhere                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_elsewhere ()
{
  PTR_STMT p = pred_stmt ;

  if (p->variant() != WHERE_BLOCK_STMT)
    error ("misplaced ELSEWHERE") ;
  else {
    terminate_block () ;
    pop_level () ;
    start_new_group2 (p) ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                     FortranParser::make_endif                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_endif ()
{
  if ((pred_stmt == 0)
      || (pred_stmt->control_parent == 0)
      || (pred_stmt->stmt_variant() != IF_NODE)
      && (pred_stmt->stmt_variant() != ELSEIF_NODE) ) {
    error ("misplaced ENDIF") ;
  }
  else {
    terminate_block () ;
    while (pred_stmt->stmt_variant() == ELSEIF_NODE) 
      pop_level () ;
    
    if (pred_stmt->stmt_variant() == IF_NODE)
      pop_level () ;
    else
      error ("ENDIF has no matching IF statement") ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                    FortranParser::make_endwhere                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_endwhere ()
{
  if ( (pred_stmt == 0)
        || (pred_stmt->control_parent == 0)
        || (pred_stmt->variant() != WHERE_BLOCK_STMT) ) {
    error ("unexpected ENDWHERE") ;
  }
  else {
    terminate_block () ;
    pop_level () ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranParser::make_endextend                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_endextend ()
{
  if ( (pred_stmt  ==0)
        || (pred_stmt->control_parent == 0)
        || (pred_stmt->variant() != PDO_NODE) 
            && (pred_stmt->variant() != PSECTIONS_NODE)
      )
    error ("illegal end extend[ed] statement") ;
  else {
    terminate_block () ;
    pop_level () ;
    if (last_in_block()->variant() == CONTROL_END)
      blocks[cur_level] = &(pred_stmt->node.fields.bl_ptr2) ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                    FortranParser::make_enddoall                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_STMT FortranParser::make_enddoall ()
{
  PTR_STMT p = 0 ;
  if ((!pred_stmt) ||
      ((pred_stmt->variant() != CDOALL_NODE) &&
       (pred_stmt->variant() != SDOALL_NODE) &&
       (pred_stmt->variant() != DOACROSS_NODE) &&
       (pred_stmt->variant() != CDOACROSS_NODE) &&
       (pred_stmt->variant() != PARDO_NODE) &&
       (pred_stmt->variant() != PDO_NODE)))
    error ("misplaced END DOALL statement") ;
  else {
    p = make_stmt (CONTROL_END, 0, 0, 0, 0) ;
    if (pred_stmt->node.fields.bl_ptr1 != *(blocks[cur_level])) {
      pred_stmt->node.fields.bl_ptr1 = pred_stmt->node.fields.bl_ptr2 ;
      pred_stmt->node.fields.bl_ptr2 = 0 ;
      *(blocks[cur_level]) = pred_stmt->node.fields.bl_ptr1 ;
    }
  }  
  return p ;
}
  
/*--------------------------------------------------------------------+
 |                                                                    |
 |                      FortranParser::make_loop                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_loop ()
{
  if ((pred_stmt->variant() != CDOALL_NODE) &&
	    (pred_stmt->variant() != SDOALL_NODE) &&
      (pred_stmt->variant() != DOACROSS_NODE) &&
      (pred_stmt->variant() != CDOACROSS_NODE))
    error ("misplaced LOOP statement");
  else {
    terminate_block () ;
    blocks[cur_level] = &(pred_stmt->node.fields.bl_ptr1) ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                     FortranParser::make_logif                      |
 |                                                                    |
 | The IF statement given, is converted to a logical if statement and |
 | the body statement is attached in the group 1 position. The IF     |
 | statement must have its control parent set before the call to      |
 | start_new_group, otherwise the control parent thread will not be   |
 | continuous.                                                        |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_STMT FortranParser::make_logif (PTR_STMT if_stmt, PTR_STMT body)
{
  if_stmt->stmt_variant (LOGIF_NODE) ;
  if_stmt->control_parent = pred_stmt ;
  if_stmt->node.if_node.true_branch = body ;
  body->control_parent = if_stmt ;
  return if_stmt ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                     FortranParser::make_extend                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_extend ()
{
  terminate_block () ;
  
  if ((pred_stmt->variant() == PDO_NODE) &&
      (pred_stmt->idx == EXTEND_NODE))
    blocks[cur_level] = &(pred_stmt->node.fields.bl_ptr2) ;
  else 
  {
    pop_level () ;
    if (pred_stmt->node.fields.bl_ptr1 != *(blocks[cur_level]))
      blocks[cur_level] = &(pred_stmt->node.fields.bl_ptr2) ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                 FortranParser::make_section_extend                 |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_section_extend ()
{
  // mark the end of the section

  exit_group (0) ;

  // mark end of psections's sections

  terminate_block () ;

  // prepare for extend statements of psections

  if ((pred_stmt->variant() == PSECTIONS_NODE) && 
      (pred_stmt->idx == EXTEND_NODE))
  {
    PTR_STMT p = pred_stmt ;
    pop_level () ;
    start_new_group2 (p) ; 
  }
  else 
  {
    if (pred_stmt->node.fields.bl_ptr1 != *(blocks[cur_level]))
      blocks[cur_level] = &(pred_stmt->node.fields.bl_ptr2) ;
  }
}
  

/*--------------------------------------------------------------------+
 |                                                                    |
 |                    FortranParser::make_section                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_section (PTR_EXPR section_name, PTR_EXPR wait_list)
{
  PTR_STMT p = 0 ;

  switch (pred_stmt->variant()) {
  case SECTION_NODE:
    exit_group (0) ;
    break ;
  case PSECTIONS_NODE:
  case PARSECTIONS_NODE:
    p = make_stmt (SECTION_NODE, 0, section_name, wait_list, 0);
    attach_statement (p) ;
    break ;
  default:
    error ("SECTION does not follow a PSECTION") ;
  }
}



/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranParser::attach_statement                   |
 |                                                                    |
 | Attaches the given statement to the parse tree by appending it to  |
 | the list of statements pointed to by blocks[cur_level].            |
 |                                                                    |
 | If blocks[cur_level] is NULL, then a new list of statements is     |
 | created prior to the statement being appended.                     |
 |                                                                    |
 | Certain statements require that a new group be created; basically  |
 | all those statements which are potential control parents.          |
 |                                                                    |
 | The basic sequence of parse tree generation is :-                  |
 |                                                                    |
 |   1. recognise the statement and its component sub-expressions     |
 |                                                                    |
 |   2. generate a Statement node, and set fields appropriately       |
 |                                                                    |
 |   3. call attach_statement, to install the statement in the tree   |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::attach_statement (PTR_STMT stat)
{
  if (!stat)  // comment lines return null statements
    return ;  // and return immediately.

  int start_new_group = SAME_GROUP ;
    
  switch (stat->variant()) {
    // these nodes start a new group 1 and a new procedure
  case (PROG_HEDR):
  case (PROC_HEDR):
  case (FUNC_HEDR):
  case (MODULE_STMT):
  case (BLOCK_DATA):
    create_procedure (stat) ;
    // these nodes start a new group 1
  case (IF_NODE):
  case (ELSEIF_NODE) :
  case (INTERFACE_STMT):
  case (STRUCT_DECL):
  case (FOR_NODE):
  case (WHILE_NODE):
  case (FORALL_NODE):
  case (PARDO_NODE):
  case (PDO_NODE):
  case (PARREGION_NODE):
  case (PSECTIONS_NODE):
  case (PARSECTIONS_NODE):
  case (SECTION_NODE):
  case (CRITSECTION_NODE):
  case (SINGLEPROCESS_NODE):
  case (CASE_NODE):
  case (WHERE_BLOCK_STMT):
    start_new_group = NEW_GROUP1 ;
    break ;
    
    // these nodes start a new group 2
  case (CDOALL_NODE) :
  case (SDOALL_NODE) :
  case (DOACROSS_NODE) :
  case (CDOACROSS_NODE) :
    start_new_group = NEW_GROUP2 ;
    break ;
    
    // no new group for these nodes
  case (ENTRY_STAT):
  case (LOGIF_NODE):
  case (COMMENT_STAT):
  case (VAR_DECL):
  case (PARAM_DECL):
  case (COMM_STAT):
  case (NAMELIST_STAT):
  case (TASK_COM):
  case (DIM_STAT):
  case (DATA_DECL):
  case (EXTERN_STAT):
  case (INTRIN_STAT):
  case (EQUI_STAT):
  case (IMPL_DECL):
  case (SAVE_DECL):
  case (INCLUDE_STAT):
  case (ATTR_DECL):
  case (GOTO_NODE):
  case (ASSGOTO_NODE):
  case (COMGOTO_NODE):
  case (ARITHIF_NODE):
  case (LOOP_NODE):
  case (EXIT_NODE):
  case (CONT_STAT):
  case (RETURN_STAT):
  case (STOP_STAT):
  case (PAUSE_NODE):
  case (ASSIGN_STAT):
  case (STMTFN_STAT):
  case (POINTER_STMT):
  case (SUM_ACC):
  case (MULT_ACC):
  case (MAX_ACC):
  case (MIN_ACC):
  case (CAT_ACC):
  case (OR_ACC):
  case (AND_ACC):
  case (READ_STAT):
  case (WRITE_STAT):
  case (PRINT_STAT):
  case (BACKSPACE_STAT):
  case (REWIND_STAT):
  case (ENDFILE_STAT):
  case (INQUIRE_STAT):
  case (OPEN_STAT):
  case (CLOSE_STAT):
  case (OTHERIO_STAT):
  case (FORMAT_STAT):
  case (PROC_STAT):
  case (ASSLAB_STAT):
  case (LOCK_NODE):
  case (UNLOCK_NODE):
  case (POST_NODE):
  case (WAIT_NODE):
  case (CLEAR_NODE):
  case (POSTSEQ_NODE):
  case (WAITSEQ_NODE):
  case (SETSEQ_NODE):
  case (PRIVATE_NODE):
  case (GUARDS_NODE):
  case (CYCLE_STMT):
  case (EXIT_STMT):
  case (CONTAINS_STMT):
  case (SWITCH_NODE):
  case (WHERE_NODE):
  case (USE_STMT):
  case (MODULE_PROC_STMT):
  case (OVERLOADED_ASSIGN_STAT): 
  case (POINTER_ASSIGN_STAT):
  case (OVERLOADED_PROC_STAT):
    break ;
  case (CONTROL_END) :
    start_new_group = END_GROUP ;
    break ;
  }
  
  
  switch (start_new_group) {
  case SAME_GROUP:
    append_statement (stat) ;  // add statement to current list
    break ;
  case NEW_GROUP1:
    append_statement (stat) ;  // add statement to current list
    start_new_group1 (stat) ;  // start a new "then" sub-tree
    break ;
  case NEW_GROUP2:
    append_statement (stat) ;  // add statement to current list
    start_new_group2 (stat) ;  // start a new "else" sub-tree
    break ;
  case END_GROUP:
    exit_group (stat) ;        // end a block and re-position
    break ;                    // the tree insertion point
  default:
    err ("internal: code 2 in 'attach_statement'");
    break ;
  }
  close_groups () ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                    FortranContext::close_groups                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::close_groups ()
{
  while (end_group > 0) {
    pred_stmt = pred_stmt->control_parent ;
    --cur_level ;
    --end_group ;
  }
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranParser::make_prog_header                  |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::make_prog_header ()
{
  if (pred_stmt->variant() != GLOBAL)
    error ("internal: not at global level between sub-programs") ;
  
  parstate = INSIDE;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                       FortranParser::enddcl                        |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::enddcl ()
{
  parstate = INEXEC;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                     FortranParser::newprog                         |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::newprog ()
{
  mod_offset = yylineno - 1 ;
  parstate = INSIDE ;
  setimpl (global_float, 'a', 'z', 0) ;
  setimpl (global_int, 'i', 'n', 0) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                       FortranParser::make_do                       |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_STMT FortranParser::make_do (statement_tag type , PTR_FTN_LABEL label, 
				 PTR_SYMB sym,  PTR_EXPR start, PTR_EXPR end , 
				 PTR_EXPR stride)
{
  PTR_STMT stmt ;
  PTR_EXPR ddot ;

  sym->dovar = 1 ;
  ddot = make_expr (DDOT, start, end, 0) ;
  stmt = make_stmt (type, sym, ddot, stride, 0) ;
  stmt->node.for_node.doend = label ;
  return stmt ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                     FortranParser::make_pardo                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_STMT FortranParser::make_pardo (
				    statement_tag type, 
				    PTR_FTN_LABEL label, 
				    PTR_SYMB sym, 
				    PTR_EXPR start, 
				    PTR_EXPR end, 
				    PTR_EXPR stride, 
				    PTR_EXPR qualities
				    )
{
  PTR_STMT stmt ;
  PTR_EXPR ddot ;

  ddot = make_expr (DDOT, start, end, 0) ;
  stmt = make_stmt (type, sym, ddot, stride, 0) ;
  stmt->node.for_node.doend = label ;
  stmt->node.for_node.where_cond = qualities ;
  return stmt ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranParser::process_interface                  |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::process_interface (PTR_STMT end_of_interface)
{
  PTR_SYMB sym_temp, last_symbol, interface_symbol;
  PTR_BLOCK b ;
  
  b = pred_stmt->node.fields.bl_ptr1 ;
  interface_symbol = pred_stmt->node.fields.symbol ;

  if (interface_symbol == 0) 
    return ;
  
  last_symbol = b->hd()->node.fields.symbol ;
  interface_symbol->node.fields.declared_name = last_symbol ;
  
  dlist_iterator<int,Statement> next(*b) ;
  PTR_STMT s ;
  while ( (s = next()) && (s != end_of_interface) )
  {
    sym_temp = s->node.fields.symbol ;
    last_symbol->node.fields.declared_name = sym_temp ;
    last_symbol = sym_temp ;
  }
  last_symbol->node.fields.declared_name = 0 ;
} 

    
/*--------------------------------------------------------------------+
 |                                                                    |
 |               FortranContext::deal_with_func_call                  |
 |                                                                    |
 | This routine is called when a function call is reduced. It must    |
 | check to see whether the call is to an intrinsic function, and     |
 | must check arguments and set the result type appropriately.        |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_EXPR FortranContext::deal_with_func_call (PTR_EXPR fcall)
{
  PTR_SYMB s = fcall->node.fields.symbol ;
  char *fname = s->ident() ;
  Intrinsic_func *fentry ;
  
  // Firstly set the default expression type from the function symbol type.
  // This is necessary in cases where this is not an intrinsic function.

  fcall->type = s->type ;

  // Now check if this IS an intrinsic function, which may yield an
  // altogether different type.

  if ( s->decl != YES && matchIntrinsic (fname, &fentry))
    {
      // 1. Check the number of actual arguments matches the definition
    
      PTR_EXPR actuals = fcall->node.fields.e1 ;
      int numActuals = actuals ? actuals->exprListLength() : 0 ;
      bool minOK = numActuals >= fentry->minArgs ;
      bool maxOK = numActuals <= fentry->numArgs ;
      bool var_args = fentry->numArgs == 0 ;

      if ( maxOK && minOK || var_args && minOK )
	{
	  // 2. Try to find a matching argument type signature
	  int sigNum ;
	  bool matchFound = false ;
	  Intrinsic_signature *sptr ;
	  for (sigNum = 0; !matchFound && sigNum < fentry->nforms; sigNum++)
	    {
	      sptr = &(fentry->sig[sigNum]) ;
	      PTR_EXPR args = fcall->node.fields.e1 ;
	      bool typeMismatch = false ;
	      int argNum = 1 ;
	      while (args && !typeMismatch)
		{
		  PTR_EXPR exp = args->node.fields.e1 ;
		  PTR_TYPE etyp = exp ? exp->type : 0 ;
		  type_tag atag = etyp ? etyp->type_variant() : T_UNKNOWN ;
		  if (fentry->numArgs == 0)
		    typeMismatch = atag != sptr->arg1 ;
		  else
		    switch (argNum) {
		    case 1: typeMismatch = atag != sptr->arg1 ; break ;
		    case 2: typeMismatch = atag != sptr->arg2 ; break ;
		    default: typeMismatch = true ;
		    }
		  args = args->node.fields.e2 ;
		  ++argNum ;
		}
	      matchFound = ! typeMismatch ;
	    }
            if (!matchFound) 
	    {
	      // Error: argument type mismatch for intrinsic function
	      // FGG changed "Error" to "Message",
	      // because we have some problems with intrinsic functions 
	      char ebuf[4096] ;
	      char *bp = ebuf ;
	      bp += sprintf (bp, "(possible error) argument type mismatch in call to intrinsic '%s'\n", fname) ;
	      PTR_EXPR args = fcall->node.fields.e1 ;
	      int anum = 1 ;
	      while (args) {
		PTR_EXPR exp = args->node.fields.e1 ;
		char *tstr = exp->type->sprintf() ;
		bp += sprintf (bp, "\targ %d was %s (%d)\n", anum++, tstr,
			       exp ? exp->type->type_variant() : 0) ;
		delete [] tstr ;
		args = args->node.fields.e2 ;
	      }
	      //errstr ("%s", ebuf) ;
	      warnstr ("%s", ebuf) ; //may be turned off in octave
	                             //in that case you will not see any warnings
	      if (!s->type)
		s->type = global_default ;
	    }
            //Next code runs even if match is not found 
	    s->attr = s->attr | INTRINSIC_BIT ;
	    switch (sptr->returnType) {
	    case T_INT:     s->type = global_int ; break ;
	    case T_FLOAT:   s->type = global_float ; break ;
	    case T_DOUBLE:  s->type = global_double ; break ;
	    case T_STRING:  s->type = global_string ; break ;
	    case T_COMPLEX: s->type = global_complex ; break ;
	    case T_BOOL:    s->type = global_bool ; break ;
	    }
/*
	  if (matchFound)
	    {
	      s->attr = s->attr | INTRINSIC_BIT ;
	      switch (sptr->returnType) {
	      case T_INT:     s->type = global_int ; break ;
	      case T_FLOAT:   s->type = global_float ; break ;
	      case T_DOUBLE:  s->type = global_double ; break ;
	      case T_STRING:  s->type = global_string ; break ;
	      case T_COMPLEX: s->type = global_complex ; break ;
	      case T_BOOL:    s->type = global_bool ; break ;
	      }
	    }
	  else
	    {
	      // Error: argument type mismatch for intrinsic function
	      // FGG changed "Error" to "Message",
	      // because we have some problems with intrinsic functions 
	      char ebuf[4096] ;
	      char *bp = ebuf ;
	      bp += sprintf (bp, "(possible error) argument type mismatch in call to intrinsic '%s'\n", fname) ;
	      PTR_EXPR args = fcall->node.fields.e1 ;
	      int anum = 1 ;
	      while (args) {
		PTR_EXPR exp = args->node.fields.e1 ;
		char *tstr = exp->type->sprintf() ;
		bp += sprintf (bp, "\targ %d was %s (%d)\n", anum++, tstr,
			       exp ? exp->type->type_variant() : 0) ;
		delete [] tstr ;
		args = args->node.fields.e2 ;
	      }
	      //errstr ("%s", ebuf) ;
	      warnstr ("%s", ebuf) ;
	      if (!s->type)
		s->type = global_default ;
	    }
*/
	}
      else 
	{
	  // Error: incorrect number of arguments to intrinsic function
	  errstr ("wrong number of arguments in call to '%s'", fname) ;
	  if (!s->type)
	    s->type = global_default ;
	}
    }
  fcall->type = s->type ;
  return fcall ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                        FortranParser::bind                         |
 |                                                                    |
 | This routine is called at the end of each sub-program unit, and    |
 | the purpose is to scan all the call statements which reference an  |
 | unknown function. The aim is to look their symbols up in an outer  |
 | scope level (possibly global) and if not found, then those symbols |
 | are automatically given global (i.e. external) scope.              |
 +--------------------------------------------------------------------*/

void FortranParser::bind ()
{
  PTR_SYMB s, sym;
  PTR_EXPR tmp;
  PTR_HASH hash_node;
  PTR_SCOPE sc ;
  
  while (first_unresolved_call != 0) 
  {
    s = first_unresolved_call->node.fields.symbol;
    sc = s->scope->enclosing_scope() ;
    if (s->decl == 0)
    {
      hash_node = sc->just_look_up_ftn_sym_in_scope (s->ident()) ;
      if (!hash_node) 
	s->scope = root_stmt->scope ;
      else if (hash_node->symbol()->variant() == s->variant()) 
      {
	sym = hash_node->symbol();
	//  remove temporary entry for s from current scope.
	s->scope->remove_symbol (s) ;

	first_unresolved_call->node.fields.symbol = sym;
      }
      else  
	errstr ("binding inconsistent call to %s", s->ident());
    }
    
    tmp = first_unresolved_call;
    first_unresolved_call = tmp->node.fields.e2;
    tmp->node.fields.e2 = 0;
  }
} 

    
/*--------------------------------------------------------------------+
 |                                                                    |
 |                 FortranParser::late_bind_if_needed                 |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranParser::late_bind_if_needed (PTR_EXPR ll_node)
{
  PTR_SYMB s;
  
  s = ll_node->node.fields.symbol;
  if ((s->node.var_decl.local == IO) || 
      (s->decl == YES))
    return ;
  else 
  {
    if (first_unresolved_call == 0) 
      last_unresolved_call = first_unresolved_call = ll_node ;
    else 
    {
      last_unresolved_call->node.fields.e2 = ll_node ;
      last_unresolved_call = ll_node ;
    }
  }	  
} 

      
/*--------------------------------------------------------------------+
 |                                                                    |
 |                 FortranContext::intrinsic_op_node                  |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_EXPR FortranContext::intrinsic_op_node (char *opname, 
					    expression_tag op, 
					    PTR_EXPR rand1,
					    PTR_EXPR rand2
					   )
{
  PTR_HASH hash_node ;
  PTR_SYMB s, sym ;
  PTR_EXPR temp1, temp2, l, result ;
  
  hash_node = this->cur_scope()->just_look_up_ftn_sym (opname) ;
  if (!hash_node) 
  {
    result = make_expr (op, rand1, rand2, 0) ;
    result->set_expr_type () ; 
    return  result ;
  }
  else 
  {
    s = hash_node->symbol();
    if (s->variant() != INTERFACE_NAME)
    {
      errstr ("can't resolve intrinsic operator %s", opname) ;
      return 0 ;
    }
    
    temp1 = make_expr (EXPR_LIST, rand1, 0, 0) ;
    temp2 = make_expr (EXPR_LIST, rand2, 0, 0) ;
    temp1->node.fields.e2 = temp2 ;
    sym = s->resolve_overloading (temp1) ;
    if (sym)
    {
      l = make_expr (FUNC_CALL, temp1, 0, sym) ;
      l->type = sym->type ;
      result = make_expr (OVERLOADED_CALL, l, 0, s) ;
      result->type = sym->type ;
    }
    else {
      result = make_expr (op, rand1, rand2, 0) ;
      result->set_expr_type () ;
      return result ;
    }
  }
  return 0 ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranContext::defined_op_node                   |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_EXPR FortranContext::defined_op_node (PTR_HASH hash_node, 
					  PTR_EXPR rand1, 
					  PTR_EXPR rand2
					 )
{
  PTR_SYMB s, sym;
  PTR_EXPR temp1, temp2, l, result;
  
  if (!hash_node) 
  {
    errstr ("unknown operator %s", hash_node->identifier()) ;
    return 0 ;
  }
  else 
  {
    s = hash_node->symbol() ;
    if (s->variant() != INTERFACE_NAME)
    {
      errstr ("cannot resolve call %s", hash_node->identifier()) ;
      return 0 ;
    }
    
    temp1 = make_expr (EXPR_LIST, rand1, 0, 0) ;
    temp2 = make_expr (EXPR_LIST, rand2, 0, 0) ;
    temp1->node.fields.e2 = temp2 ;
    sym = s->resolve_overloading (temp1) ;
    if (sym != 0)
    {
      l = make_expr (FUNC_CALL, temp1, 0, sym) ;
      l->type = sym->type ;
      result = make_expr (OVERLOADED_CALL, l, 0, s) ;
      result->type = sym->type ;
    }
    else {
      errstr ("can't resolve call %s", hash_node->identifier());
      return 0 ;
    }
  }
  return 0 ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::check_and_install                   |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_EXPR FortranContext::check_and_install (PTR_HASH h, PTR_EXPR d, int nd)
{
  PTR_TYPE p ;
  PTR_SYMB s ;
  PTR_EXPR r ;

  if (!h || !d)
    return 0 ;
  else
    s = h->symbol() ;

  // Check if the variable is already declared

  if ( (s->scope != 0) && scope_is_current (s->scope) && s->type) {
    if (d && s->type->variant() != T_ARRAY) {
      p = install_array (d, s->type, nd);
      s->type = p;
    }
  }
  else {
    if (h->symbol() && h->symbol()->type)
      p = h->symbol()->type;
    else if (!undeftype)
      p = impltype[tolower(*(h->identifier())) - 'a'];
    else
      err ("variable type unknown");
    if (d)
      p = install_array (d, p, nd);
    s = install_entry (h, SOFT);
    s->set_type (p, LOCAL);
  }
  if (d) {
    r = p->node.ar_decl.ranges
      = make_expr (ARRAY_REF, d, 0, s);
  }
  else
    r = make_expr (VAR_REF, 0, 0, s);
  return  make_expr (EXPR_LIST, r, 0, 0);
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                 FortranContext::install_entry                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::install_entry (PTR_HASH entry, int decl_type)
{
  register PTR_SYMB symb_ptr ;
  PTR_STMT p ;
  PTR_SCOPE sc ;
  
  if (decl_type == HARD && entry->symbol() &&
      entry->symbol()->scope != cur_scope ())
    entry = cur_scope()->look_up_ftn (entry->identifier(), HARD);
  if ((entry->symbol()) && 
      ( scope_is_current ((entry->symbol()->scope) ) ||
       ((entry->symbol()->variant() == FUNCTION_NAME) &&
	(entry->symbol()->scope->head()->variant() == FUNC_HEDR))))
  {
    if (entry->symbol()->decl == SOFT) {
      entry->symbol()->decl = decl_type;
      return (entry->symbol());
    }
    if (decl_type == SOFT)
      return (entry->symbol());
    /* else */
    errstr ("identifier redeclared: %s", 
	    entry->symbol()->ident()) ;
    return 0 ;
  }

  // Otherwise, make a new default symbol within the current scope
 
  symb_ptr = make_symb (DEFAULT_NAME, entry, decl_type) ;
  return symb_ptr ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::chase_qual_index                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::chase_qual_index (PTR_EXPR ref, PTR_EXPR index_list)
{
  if (ref->expr_variant() == ARRAY_REF)
    chase_qual_index (ref->node.array_ref.array_elt, index_list) ;
  else if (ref->expr_variant() == RECORD_REF)
    chase_qual_index (ref->node.record_ref.rec_field, index_list) ;
  else if (ref->expr_variant() == VAR_REF) {
    /*
     * check that this is an array ref symbol and that dimensions are ok
     * etc. 
     */
    ref->expr_variant (ARRAY_REF);
    ref->node.array_ref.index = index_list;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::chase_qual_field                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::chase_qual_field (PTR_EXPR ref, PTR_SYMB field_id)
{
  if (ref->expr_variant() == ARRAY_REF) {
    if (! ref->node.array_ref.array_elt)
      ref->node.array_ref.array_elt =
	make_expr (VAR_REF, 0, 0, field_id);
    else
      chase_qual_field (ref->node.array_ref.array_elt, field_id) ;
  } 
  else if (ref->expr_variant() == RECORD_REF) {
    if (! ref->node.record_ref.rec_field)
      ref->node.record_ref.rec_field =
	make_expr (VAR_REF, 0, 0, field_id);
    else
      chase_qual_field (ref->node.record_ref.rec_field, field_id) ;
  } 
  else if (ref->variant() == VAR_REF) {
    ref->expr_variant (RECORD_REF) ;
    ref->node.record_ref.rec_field =
      make_expr (VAR_REF, 0, 0, field_id);
  } 
  else
    error ("internal cqf01");
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranContext::install_array                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_TYPE FortranContext::install_array (PTR_EXPR range, PTR_TYPE base, 
					int nd)
{
  PTR_TYPE ret;
  
  ret = make_type (T_ARRAY) ;
  ret->node.ar_decl.num_dimensions = nd ;
  ret->node.ar_decl.base_type = base ;
  return ret ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::make_local_symbol                   |
 |                                                                    |
 | Given a symbol identifier, a symbol variant tag, and a pointer to  |
 | a type node, create a new symbol at the current scope and return   |
 | a pointer to that new symbol.                                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_local_symbol (const char *nm, symbol_tag varnt,
					    PTR_TYPE type, int kind)
{
#ifdef TRACE_ON
  printf ("make_local_symbol: %s (%d)\n", nm, (int)varnt) ;
#endif

  PTR_HASH entry = cur_scope()->look_up_ftn (nm, HARD) ;
  return ( make_sym_entry (entry, varnt, type, cur_scope(), kind) ) ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::make_global_symbol                  |
 |                                                                    |
 | Given a symbol identifier, a symbol variant tag, and a pointer to  |
 | a type node, create a new symbol at the global scope and return    |
 | a pointer to that new symbol.                                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_global_symbol (const char *nm, symbol_tag varnt,
					     PTR_TYPE type, int kind)
{
#ifdef TRACE_ON
  printf ("make_global_symbol: %s (%d)\n", nm, (int)varnt) ;
#endif

  PTR_HASH entry = global_scope()->look_up_ftn (nm, HARD) ;
  return ( make_sym_entry (entry, varnt, type, global_scope(), kind) ) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranContext::make_sym_entry                    |
 |                                                                    |
 | Given a symbol table entry h, a symbol tag varnt, a type, a scope  |
 | and a kind, we create a new symbol, initialise its fields, link it |
 | to its symbol table entry, and install that entry at the given     |
 | scope.                                                             |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_sym_entry (PTR_HASH h, symbol_tag varnt,
					 PTR_TYPE type, PTR_SCOPE scope, 
					 int kind)
{
  PTR_SYMB sym ;
  int typing = IMPL ;
  
  // If type is undefined, then obtain type from implicit type table.

#ifdef TRACE_ON
  printf ("make_sym_entry '%s' (%d)\n", h->identifier(), (int)varnt) ;
#endif

  if ((varnt != PROGRAM_NAME) 
      && (varnt != DEFAULT_NAME) &&
      (varnt != PROCEDURE_NAME)) 
  {
    if (type == 0) {
      typing = IMPL ;
      type =  impltype[tolower(*(h->identifier())) - 'a'];
    }
    else if ((type->variant() == T_ARRAY) &&
	     (type->node.ar_decl.base_type == 0)) {
      typing = IMPL ;
      type->node.ar_decl.base_type = 
	impltype[tolower(*(h->identifier())) - 'a'];
    }
    else
      typing = TYPED ;
  }

  // Note: may want to avoid duplication and copying of identifiers
  //       from table entries into symbol nodes....

  sym = make_symb (varnt, h->identifier() ) ;
  
  // Set up the fields of the new symbol.
  
  sym->symb_variant (varnt) ;
  sym->type = type ;
  sym->scope = scope ;
  sym->id_list = 0 ;
  sym->impl_type = typing ;
  
  switch (varnt) {
  case VARIABLE_NAME: 
    // If symbol is not a formal parameter, 
    // then mark it as either local or global.
    if (sym->node.var_decl.local != IO) 
      sym->node.var_decl.local = kind ;
    break ;
  case FUNCTION_NAME:
  case PROCEDURE_NAME:
    sym->node.proc_decl.seen = kind ;
    break ;
  default :
    break ;
  }
  
  // Initialize administrative stuff for the new symbol.

  sym->parent = h ;               // symbol points to its table entry
  
  // Point the hash entry to the symbol table entry.

  h->set_attr (sym) ;

  // Install the hash entry in the given scope.

  scope->install_entry (h) ;

  return sym ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranContext::make_constant                     |
 |                                                                    |
 | Given a symbol table entry and a type, this function either finds  |
 | and returns the symbol associated with that entry, or creates a    |
 | new symbol and associates it with that entry.                      |
 |                                                                    |
 | If (no symbol)                                                     |
 |   return new CONST_NAME symbol, in current scope, with given type  |
 | Else                                                               |
 |   If (symbol is defined at the global level)                       |
 |      make new symbol table entry                                   |
 |      make new CONST_NAME symbol, scoped locally, with given type   |
 |      new entry points to new symbol                                |
 |      return new symbol                                             |
 |                                                                    |
 | If (located symbol is scoped locally)                              |
 |   set symbol variant to CONST_NAME                                 |
 |   set type appropriately                                           |
 |   return the located symbol                                        |
 |                                                                    |
 | Otherwise: make a new symbol table entry, make a new symbol,       |
 | scoped locally with variant CONST_NAME and given type, and         |
 | return that new symbol.                                            |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_constant (PTR_HASH entry, PTR_TYPE type)
{
  PTR_SYMB sym ;

  sym = entry->symbol() ;

#ifdef TRACE_ON
  printf ("make_constant: %s\n", entry->identifier()) ;
#endif

  if (!sym) {
    sym = make_sym_entry (entry, CONST_NAME, type, cur_scope (), 0) ;
    return sym ;
  }
  
  if (!scope_is_current (sym->scope) ) {
    sym = make_local_symbol (entry->identifier(), CONST_NAME, type, 0) ;
    return sym ;
  }
  else {
    symbol_tag sym_variant = sym->symb_variant() ;
    
    // TBD: perhaps we should check that sym->type is consistent with
    // TBD: that symbol becoming a constant

    if (!(sym->type)) 
      sym->type = type;

    if (sym_variant != CONST_NAME)
      sym->symb_variant (CONST_NAME) ;
    else
      errstr ("re-declaration of constant '%s'", 
	      entry->identifier()) ;
  }
  return sym ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranContext::make_scalar                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_scalar (PTR_HASH entry, PTR_TYPE type, int kind)
{
  PTR_SYMB sym = entry->symbol();

#ifdef TRACE_ON
  printf ("make_scalar: %s \n", entry->identifier()) ;
#endif

  if (!sym)
    return make_sym_entry (entry, VARIABLE_NAME, type, cur_scope (), kind) ;
  
  symbol_tag sym_variant = sym->symb_variant();

  PTR_STMT header = cur_scope()->head() ;

  PTR_SYMB head_symbol = header->symbol() ;

  if ( scope_is_global (sym->scope) )
    if ((header->stmt_variant() == FUNC_HEDR) &&
	(head_symbol->ident == *(entry->name()) ))
      if (type == 0)
	return sym ;
      else
      {
	PTR_SYMB result_symbol = head_symbol->node.fields.declared_name ;
	head_symbol->type = type ;
	if (result_symbol) 
	  result_symbol->type = type ;
	return sym ;
      }
    else 
      return make_local_symbol (entry->identifier(), VARIABLE_NAME, type, kind) ;
  
  if ( scope_is_current (sym->scope) )
    switch (sym_variant) 
    {
    case VARIABLE_NAME:
      if (!(sym->type)) 
	sym->type = type ;
      else {
	switch (sym->type->type_variant () )
	  {
	  case T_STRING:
	  case T_ARRAY:
	    if (type != 0)
	      sym->type->node.ar_decl.base_type = type ;
	    return sym ;
	  case T_POINTER:
	    if (type != 0)
	      sym->type->node.fields.base_type = type ;
	    return sym ;
	  case T_INT:
	  case T_FLOAT:
	  case T_DOUBLE:
	  case T_COMPLEX:
	  case T_BOOL:
	  case T_STRUCT:
	    if (type != 0)
	      sym->type = type ;
	    return sym ;
	  }
      break;
      }
    case CONST_NAME:
      if (type != 0)
	sym->type = type ;
      return sym ;
    case ROUTINE_NAME:
      if (type != 0)
	sym->type = type;
      return sym ;
    case LABEL_VAR:
      return sym ;
    default:
      errstr ("make_scalar: inconsistent declaration of identifier %s", 
	      entry->identifier()) ;
      return 0 ;
    }
  
  return make_local_symbol (entry->identifier(), VARIABLE_NAME, type, kind) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                    FortranContext::make_array                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_array (PTR_HASH entry, PTR_TYPE base_type,
				     PTR_EXPR ranges, int nd, int kind)
{
  PTR_SYMB sym = entry->symbol() ;
  PTR_TYPE array_type = 0 ;

#ifdef TRACE_ON
  printf ("make_array: %s\n", entry->identifier()) ;
#endif

  if (!sym)
  {
    array_type = make_type (T_ARRAY) ;
    array_type->node.ar_decl.base_type = base_type ;
    array_type->node.ar_decl.num_dimensions = nd ;
    array_type->node.ar_decl.ranges = ranges ;
    return (make_sym_entry (entry, VARIABLE_NAME, array_type,
			    cur_scope (), kind)) ;
  }
  
  symbol_tag sym_variant = sym->symb_variant() ;

  if ( scope_is_global (sym->scope) )
  {
    array_type = make_type (T_ARRAY) ;
    array_type->node.ar_decl.base_type = base_type ;
    array_type->node.ar_decl.num_dimensions = nd ;
    array_type->node.ar_decl.ranges = ranges ;
    return make_local_symbol (entry->identifier(), VARIABLE_NAME,
			      array_type, kind) ;
  }
  
  if ( scope_is_current (sym->scope) )
    switch (sym_variant) 
    {
    case VARIABLE_NAME:

      if (!(sym->type)) {
	array_type = make_type (T_ARRAY) ;
	array_type->node.ar_decl.base_type = base_type ;
	array_type->node.ar_decl.num_dimensions = nd ;
	array_type->node.ar_decl.ranges = ranges ;
	sym->type = array_type ;
	return sym ;
      }

      switch ( sym->type->type_variant() ) 
      {
      case T_ARRAY:
	if (base_type != 0)
	  sym->type->node.ar_decl.base_type = base_type ;
	if (nd)
	  {
	    sym->type->node.ar_decl.num_dimensions = nd ;
	    if (ranges)
	      sym->type->node.ar_decl.ranges = ranges ;
	  }
	return sym ;
      case T_STRING:
      case T_POINTER:
      case T_INT:
      case T_FLOAT:
      case T_DOUBLE:
      case T_COMPLEX:
      case T_BOOL:
      case T_STRUCT:
      {
	array_type = make_type (T_ARRAY) ;
	if (base_type == 0)
	  array_type->node.ar_decl.base_type = 
	    sym->type ;
	else
	  array_type->node.ar_decl.base_type = base_type ;
	array_type->node.ar_decl.num_dimensions = nd ;
	array_type->node.ar_decl.ranges = ranges ;
	sym->type = array_type ;
	return sym ;
      }
      default:
	return sym ;
      }
//      break ;
    default:
      errstr ("inconsistent array declaration of identifier %s", 
	      entry->identifier()) ;
      return 0 ;
    }
  
  array_type = make_type (T_ARRAY) ;
  array_type->node.ar_decl.base_type = base_type ;
  array_type->node.ar_decl.num_dimensions = nd ;
  array_type->node.ar_decl.ranges = ranges ;
  return make_local_symbol (entry->identifier(), VARIABLE_NAME, array_type,
			    kind) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranContext::make_pointer                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_pointer (PTR_HASH entry, PTR_TYPE base_type, int kind)
{
  PTR_SYMB sym = entry->symbol() ;
  PTR_TYPE pointer_type = 0 ;
  
  if (!sym)
  {
    pointer_type = make_type (T_POINTER) ;
    pointer_type->node.fields.base_type = base_type ;
    return (make_sym_entry (entry, VARIABLE_NAME, pointer_type,
			    cur_scope (), kind)) ;
  }
  
  symbol_tag sym_variant = sym->symb_variant() ;

  if ( scope_is_global (sym->scope) )
  {
    pointer_type = make_type (T_POINTER) ;
    pointer_type->node.fields.base_type = base_type ;
    return make_local_symbol (entry->identifier(), VARIABLE_NAME, pointer_type, 
			      kind) ;
  }
  
  if ( scope_is_current (sym->scope) )
    switch (sym_variant) 
    {
    case VARIABLE_NAME:
      if (!(sym->type)) {
	pointer_type = make_type (T_POINTER) ;
	pointer_type->node.fields.base_type = base_type ;
	sym->type = pointer_type ;
	return sym ;
      }
      switch ( sym->type->type_variant() )
	{
	case T_POINTER:
	  if (base_type)
	    sym->type->node.fields.base_type = 
	      base_type ;
	  return sym ;
	case T_STRING:
	case T_ARRAY:
	case T_INT:
	case T_FLOAT:
	case T_DOUBLE:
	case T_COMPLEX:
	case T_BOOL:
	case T_STRUCT:
	  {
	    pointer_type = make_type (T_POINTER) ;
	    if (!base_type)
	      pointer_type->node.fields.base_type = 
		sym->type ;
	    else
	      pointer_type->node.fields.base_type = base_type ;
	    sym->type = pointer_type ;
	    return sym ;
	  }
	default:
	  return sym ;
	}
//      break ;
    default:
      errstr ("inconsistent pointer declaration of identifier %s", 
	      entry->identifier()) ;
      return 0 ;
    }
  
  pointer_type = make_type (T_POINTER) ;
  pointer_type->node.fields.base_type = base_type ;
  return make_local_symbol (entry->identifier(), VARIABLE_NAME, pointer_type,
			    kind) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranContext::make_function                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_function (PTR_HASH entry, PTR_TYPE type, int kind)
{
  PTR_SYMB sym = entry->symbol() ;

  if (!sym)
    if (kind == LOCAL)
      return (make_sym_entry (entry, FUNCTION_NAME, type,
			      cur_scope (), kind)) ;
    else 
      return (make_sym_entry (entry, FUNCTION_NAME, type,
			      global_scope (), kind)) ;
  
  symbol_tag sym_variant = sym->symb_variant () ;

  if ( !scope_is_current (sym->scope) ) 
    if (sym_variant == FUNCTION_NAME)
      if (kind == LOCAL)
	return make_local_symbol (entry->identifier(), FUNCTION_NAME, type,
				  kind) ;
      else
	return sym ;
    else if (sym_variant == ROUTINE_NAME)
    {
      sym->symb_variant (FUNCTION_NAME) ;
      return sym ;
    }
    else if (sym_variant == INTERFACE_NAME)
      return sym ;
    else if (sym_variant == VARIABLE_NAME)
    {
      sym->symb_variant (FUNCTION_NAME) ;
      return sym ;
    }
    else
    {
      errstr ("inconsistent function %s", entry->identifier()) ;
      return 0 ;
    }

  if ( scope_is_current (sym->scope) )
    switch (sym_variant) 
    {
    case VARIABLE_NAME:

      if (!(sym->type)) 
	sym->type = type;

      switch ( sym->type->type_variant() ) 
      {
      case T_STRING:
      case T_INT:
      case T_FLOAT:
      case T_DOUBLE:
      case T_COMPLEX:
      case T_BOOL:
      case T_POINTER:
      case T_STRUCT:
	// if not a formal parameter, convert it into a global function.
	if (sym->node.var_decl.local != IO)
	{
	  sym->symb_variant (FUNCTION_NAME) ;
/*
   was before fgg changed it on 2000/12/05
   since declaration of function like    a(i,j)=i+j   is local
   
   convert_to_global (sym) ;
*/
          if (kind!=LOCAL)
	  {
	    convert_to_global (sym) ;
          }
	}
	if (type)
	  sym->type = type ;
	return sym ;
      case T_ARRAY:
	if (kind == LOCAL)
	{
	  errstr ("inconsistent function %s", entry->identifier()) ;
	  return 0 ;
	} 
	else
	  return sym ;
      }
    case FUNCTION_NAME:
      sym->symb_variant (FUNCTION_NAME) ;
      return sym ;
    case ROUTINE_NAME:
      sym->symb_variant (FUNCTION_NAME) ;
      return sym ;
    case INTERFACE_NAME:
      return sym ;
    default:
      errstr ("inconsistent function %s", entry->identifier()) ;
      return 0 ;
    }
  
  return make_local_symbol (entry->identifier(), FUNCTION_NAME, type, kind) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranContext::make_external                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_external (PTR_HASH entry, PTR_TYPE type)
{
  PTR_SYMB sym = entry->symbol() ;
  
  if (!sym)
    return (make_sym_entry (entry, ROUTINE_NAME, type,
			    global_scope (), NO)) ;
  
  symbol_tag sym_variant = sym->symb_variant () ;
  
  if ( scope_is_global (sym->scope) ) 
    if ((sym_variant == FUNCTION_NAME) ||
	(sym_variant == PROCEDURE_NAME) ||
	(sym_variant == ROUTINE_NAME) )
      return sym ;
    else
    {
      errstr ("illegal category of external %s", entry->identifier()) ;
      return 0 ;
    }
       
  if ( scope_is_current (sym->scope) )
    switch (sym_variant)
      {
      case VARIABLE_NAME:
	if (!(sym->type)) 
	  sym->type = type;
	
	// if not a formal parameter, convert it into a global function.
	if (sym->node.var_decl.local != IO)
	  {
	    sym->symb_variant (ROUTINE_NAME) ;
	    convert_to_global (sym) ;
	  }
	if (type)
	  sym->type = type ;
	return sym ;

      case ROUTINE_NAME:
	return sym ;
      default:
	errstr ("cannot make '%s' EXTERNAL", entry->identifier()) ;
	return 0 ;
    }

  // If scope of sym is not global and not current, then we must
  // assume that a local routine-name is being defined.

  return make_local_symbol (entry->identifier(), ROUTINE_NAME, type, 0) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranContext::make_procedure                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_procedure (PTR_HASH entry, int kind)
{
  PTR_SYMB sym = entry->symbol() ;

  if (!sym)
    if (kind == LOCAL)
      return (make_sym_entry (entry, PROCEDURE_NAME, 0,
			      cur_scope (), kind)) ;
    else
      return (make_sym_entry (entry, PROCEDURE_NAME, 0,
			      global_scope(), kind)) ;

  symbol_tag sym_variant = sym->symb_variant () ;

  if ( scope_is_global (sym->scope) )
    if (sym_variant == PROCEDURE_NAME)
      return sym ;
    else if (sym_variant == ROUTINE_NAME)
    {
      sym->symb_variant (PROCEDURE_NAME) ;
      return sym ;
    }
    else if (sym_variant == INTERFACE_NAME)
      return sym ;
    else
    {
      errstr ("inconsistent subroutine %s", entry->identifier()) ;
      return 0 ;
    }
  
  if ( scope_is_current (sym->scope) )
    switch (sym_variant)
    {
    case PROCEDURE_NAME:
      return sym ;
    case VARIABLE_NAME:
      // if not a formal parameter, convert it into a global procedure

      if (sym->node.var_decl.local != IO)
      {
	sym->symb_variant (PROCEDURE_NAME) ;
	convert_to_global (sym) ;
      }
      return sym ;
    case ROUTINE_NAME:
      sym->symb_variant (PROCEDURE_NAME) ;
      return sym ;
    case INTERFACE_NAME:
      return sym ;
    default:
      errstr ("inconsistent subroutine %s", entry->identifier()) ;
      return 0 ;
    }
  
  // If scope of sym is neither local nor global, then assume it is
  // not visible locally, and make a fresh local symbol.

  return make_local_symbol (entry->identifier(), PROCEDURE_NAME, 0, kind) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranContext::make_program                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_program (PTR_HASH entry)
{
  PTR_SYMB sym = entry->symbol () ;

  if (!sym)
    return (make_sym_entry (entry, PROGRAM_NAME, 0, 
			    global_scope (), 0)) ;
  
  if ( scope_is_global (sym->scope) )
  {
    errstr ("duplicate program declaration %s", entry->identifier()) ;
    return 0 ;
  }
  
  if ( scope_is_current (sym->scope) )
  {
    errstr ("misplaced declaration of program %s", entry->identifier()) ;
    return 0 ;
  }
  
  return make_global_symbol (entry->identifier(), PROGRAM_NAME, 0, 0) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranContext::make_common                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_common (PTR_HASH entry)
{
  PTR_SYMB sym = entry->symbol();

  if (!sym)
  {
    return (make_sym_entry(entry, DEFAULT_NAME, 0, global_scope (), 0));
  }
  
  char* fname=sym->ident();
  Intrinsic_func* fentry;
  
  if (scope_is_global(sym->scope))
  {
    //fgg added next if (only one) on 2000/11/29 to cope
    //with intrinsic functions correctly
    //[if common block name is the same as one of intrinsic names
    // - it is allowed]
    if ((sym->symb_variant() == FUNCTION_NAME) &&
        matchIntrinsic(fname, &fentry))
    {
      return sym;
    }
    else if (sym->symb_variant() == DEFAULT_NAME)
    {
      return sym;
    }
    else 
    {
      errstr ("inconsistent common declaration %s", entry->identifier()) ;
      return 0;
    }
  }
  
  // A local entity can have same name as a common block name.
  // We therefore ignore the local entity, and create a global
  // symbol and symbol-table entry for this common block name.
  
  return make_global_symbol(entry->identifier(), DEFAULT_NAME, 0, 0) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::make_derived_type                   |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_derived_type (PTR_HASH entry, 
					    PTR_TYPE type, 
					    int kind)
{
  PTR_SYMB sym = entry->symbol() ;
  PTR_TYPE struct_type = 0 ;
  
  if (!sym)
  {
    struct_type = make_type (T_STRUCT) ;
    return (make_sym_entry (entry, TYPE_NAME, struct_type,
			    cur_scope (), kind)) ;
  }

  if ( scope_is_global (sym->scope) ) {
    struct_type = make_type (T_STRUCT) ;
    return make_local_symbol (entry->identifier(), TYPE_NAME, 
			      struct_type, kind) ;
  }
  
  if ( scope_is_current (sym->scope) ) {
    symbol_tag sym_variant = sym->symb_variant () ;
    switch (sym_variant) 
    {
    case VARIABLE_NAME:

      if (!(sym->type))
	sym->type = type;
      
      switch ( sym->type->type_variant() )
      {
      case T_STRING:
      case T_ARRAY:
	if (type)
	  sym->type->node.ar_decl.base_type = type ;
	return sym ;
      case T_INT:
      case T_FLOAT:
      case T_DOUBLE:
      case T_COMPLEX:
      case T_BOOL:
      case T_POINTER:
      case T_STRUCT:
      {
	struct_type = make_type (T_STRUCT) ;
	sym->type = struct_type ;
	return sym ;
      }
      default:
	return sym ;
      }
//      break ;
    default:
      errstr ("inconsistent struct declaration for identifier %s", 
	      entry->identifier() ) ;
      return 0 ;
    }
  }

  struct_type = make_type (T_STRUCT) ;
  return make_local_symbol (entry->identifier(), TYPE_NAME, struct_type, kind) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::make_local_entity                   |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_local_entity (PTR_HASH entry, symbol_tag variant,
					    PTR_TYPE type, int kind)
{
  PTR_SYMB sym = entry->symbol() ;
  
#ifdef TRACE_ON
  printf ("make_local_entity: %s\n", entry->identifier() ) ;
#endif

  if (!sym)
    return (make_sym_entry (entry, variant, type,
			    cur_scope (), kind)) ;
  
  symbol_tag sym_variant = sym->symb_variant () ;
  
  if ( scope_is_current (sym->scope) )
    if (variant == sym_variant)
      return sym ;
    else 
    {
      errstr ("make_local_entity: inconsistent declaration of identifier %s", 
	      entry->identifier()) ;
      return 0 ;
    }
  else
    return make_local_symbol (entry->identifier(), variant, type, kind) ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::make_global_entity                  |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB FortranContext::make_global_entity (PTR_HASH entry, symbol_tag varnt,
					     PTR_TYPE type, int kind)
{
  PTR_SYMB sym = entry->symbol() ;

#ifdef TRACE_ON
  printf ("make_global_entity: %s\n", entry->identifier()) ;
#endif

  if (!sym)
    return make_sym_entry (entry, varnt, type,
			   global_scope (), kind) ;
  
  symbol_tag sym_variant = sym->symb_variant () ;
  
  if ( scope_is_global (sym->scope) && (varnt == sym_variant) )
    return sym ;
  else 
  {
    errstr ("make_global_entity: inconsistent declaration of identifier %s", 
	    entry->identifier()) ;
    return 0 ;
  }
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                FortranContext::deal_with_options                   |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_EXPR FortranContext::deal_with_options (PTR_HASH name, 
					    PTR_TYPE type, 
					    int attributes, 
					    PTR_EXPR dims, 
					    int nd, 
					    PTR_EXPR value,
					    int decl_flag)
{
  PTR_SYMB s = 0 ;
  PTR_EXPR l = 0 ;
  
  if (attributes & PARAMETER_BIT) 
  {
    s = make_constant (name, type) ;
    s->node.const_value = value ;
    s->attr = attributes ;
    l = make_expr (CONST_REF, 0, 0, s) ;
    return l ;
  }
  if (attributes & EXTERNAL_BIT) 
  {
    s = make_scalar (name, type, NO) ;
    s->attr = attributes ;
    s->symb_variant (ROUTINE_NAME) ;
    l = make_expr (VAR_REF, 0, 0, s) ;
    return l ;
  }
  if (attributes & INTRINSIC_BIT) 
  {
    s = make_function (name, type, NO) ;
    s->attr = attributes ;
    l = make_expr (VAR_REF, 0, 0, s) ;
    return l ;
  }

  if ( (s = name->symbol()) != 0 )
    {
      // The symbol is already declared so just update its fields
      if (type)
      {
        s->type=type;
      }
      s->attr = s->attr | attributes ;
      // Others ??
      l = make_expr (VAR_REF, 0, 0, s) ;
    }
  else
    {
      s = make_scalar (name, type, LOCAL) ;
      s->attr = attributes ;
    }

/*
  was before fgg changed on 2000/12/08
  for constructions like:
    dimension x(10,10)
    real x
    
  if ((attributes & DIMENSION_BIT) || (dims != 0))
*/
  if ((s->attr & DIMENSION_BIT) || (attributes & DIMENSION_BIT) || (dims != 0))
  {
/* was before fgg changed this on 2000/12/11
    s = make_array (name, 0, dims, nd, LOCAL) ;
*/
    if (type)
    {
      s = make_array (name, type, dims, nd, LOCAL) ;
    }
    else
    {
      s = make_array (name, 0, dims, nd, LOCAL) ;
    }

    s->type->node.ar_decl.ranges = dims ;

    if (decl_flag == ENTITY_DECL)
      l = make_expr (ARRAY_REF, dims, 0, s) ;
    else
      l = make_expr (VAR_REF, 0, 0, s) ;
  }
  if (attributes & POINTER_BIT) 
  {
    s = make_pointer (name, 0, LOCAL) ;
    l = make_expr (VAR_REF, 0, 0, s) ;
  }
  
  if (l == 0)
  {
    l = make_expr (VAR_REF, 0, 0, s) ;
  }
  
  if (value != 0)
    l = make_expr (ASSGN_OP, l, value, 0) ;       
  return l ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                  FortranContext::lookup_type                       |
 |                                                                    |
 | This routine is called by the Fortran parser when a type name is   |
 | encountered during the declaration of data of a derived type, thus |
 |                                                                    |
 | If we see:   TYPE (typename)                                       |
 |                                                                    |
 | then, the type with name typename needs to be found in the current |
 | scope, and checked to see if it is a valid type name.              |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_TYPE FortranContext::lookup_type (PTR_HASH tpname)
{
  PTR_SYMB sym ;
  PTR_TYPE newtype ;
  
  if (tpname && ( sym = tpname->symbol() ) 
      && (sym->symb_variant() == TYPE_NAME) 
      && ( scope_is_current (sym->scope) ) )
  {
    newtype = make_type (T_DERIVED_TYPE) ;
    newtype->name = sym ;
    return  newtype ;
  }
  errstr ("undefined type %s", tpname->identifier()) ;
  return 0 ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                     FortranParser::convci                          |
 |                                                                    |
 | Converts a character representation of an integer into the integer |
 | itself. The length of the string is specified.                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

long FortranParser::convci (int n, const char *s)
{
  long sum = 0 ;

  while (n-- > 0)
    sum = 10 * sum + (*s++ - '0') ;
  return sum ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                     FortranParser::convic                          |
 |                                                                    |
 | Converts a long integer into a character representation and yields |
 | a pointer to that character string.                                |
 |                                                                    |
 +--------------------------------------------------------------------*/

char* FortranParser::convic (long n)
{
  char tmp [32] ;
  char *p ;

  sprintf (tmp, "%ld", n) ;
  char *t = new char [ strlen(tmp) ] ;
  
  for (p = tmp; *p; p++, t++)
    *t = *p ;

  return t ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   FortranContext::initialise                       |
 |                                                                    |
 | Initialises the parser and also the static members of the global   |
 | parse_constant class if they are undefined.                        |
 |                                                                    |
 +--------------------------------------------------------------------*/

void FortranContext::initialise () {
  if (!global_int) {
    parse_constants::global_int = make_type (T_INT) ;
    parse_constants::global_float = make_type (T_FLOAT) ;
    parse_constants::global_double = make_type (T_DOUBLE) ;
    parse_constants::global_char = make_type (T_CHAR) ;
    parse_constants::global_string = make_type (T_STRING) ;
    parse_constants::global_bool = make_type (T_BOOL) ;
    parse_constants::global_complex = make_type (T_COMPLEX) ;
    parse_constants::global_gate = make_type (T_GATE) ;
    parse_constants::global_event = make_type (T_EVENT) ;
    parse_constants::global_sequence = make_type (T_SEQUENCE) ;
    parse_constants::global_default = make_type (T_DEFAULT) ;
    PTR_EXPR one = make_expr (INT_VAL, 0, 0, 0) ;
    one->node.ival = 1 ;
    one->type = global_int ;
    parse_constants::global_byte =  make_type_node (parse_constants::global_int, 
						    make_expr (LEN_OP, one, 0, 0)) ;
  }
  root_stmt = make_stmt (GLOBAL, 0, 0, 0, 0) ;
  push_scope_level (root_stmt, 0) ;
  root_scope = root_stmt->scoping () ;
  start_new_group1 (root_stmt) ;
  initIntrinsics (this) ;
}

static bool matchIntrinsic (const char* str, Intrinsic_func **fun)
{
  struct Intrinsic_func *pk, *pend ;
  int k = tolower(*str) - 'a' ;

  if (pk = funstart[k])
    for (pend = funend[k]; pk != pend; ++pk) {
      const char* i = pk->funcName;
      const char* j = str ;

      while (*++i == tolower(*++j) && *i != '\0' && *j != '\0') ;

      if ((*i == '\0') && (*j == '\0'))
	{
	  *fun = pk ;
	  return true ;
	}
    }
  *fun = 0 ;
  return false ;
}

void initIntrinsics (FortranContext *fc)
{
  Intrinsic_func *p ;
  register int i, j ;
  
  for (i = 0; i < 26; ++i)
    funstart[i] = 0 ;
  
  for (p = intrinsics; p->funcName; ++p)
    {
      // 1. Create global symbols for all intrinsic functions
      char *fn = p->funcName ;
      PTR_SYMB fsym = fc->make_global_symbol (fn, FUNCTION_NAME, fc->global_default, LOCAL) ;
      fsym->attr = fsym->attr | INTRINSIC_BIT | EXTERNAL_BIT ;

      // 2. Create the alphabetic index into the list of intrinsics

      j = p->funcName[0] - 'a';
      if (funstart[j] == 0)
	funstart[j] = p ;
      funend[j] = p + 1 ;
  }
}
