/*--------------------------------------------------------+
 |                                                        |
 |     Lexer for Convex Fortran 77 with F90 Extensions    |
 |                                                        |
 |     Copyright (C) Nigel Topham, 1996                   |
 |                                                        |
 +--------------------------------------------------------*/

%{
// $Id: lexer.l,v 1.5 1999/08/06 14:45:48 npt Exp $

#include "Octave/Fortran/FortranLexer.h"

// #define DEBUG_LEXER 1

// Set system limits

#define YY_READ_BUF_SIZE 13210

// Local scanner support macros and variables
  
#define BufferYYText(_to_)  \
  { \
    int i ; \
    const char *from = YYText() ; \
    int len = YYLeng() ; \
    for (i = 0; i < len; i++) \
      *(_to_)++ = *from++ ; \
      }

#define UnBufferYYText(_from_, _to_) \
{ \
  do { \
    unput (*(--(_from_))) ; \
  } while ((_from_) != (_to_)) ; \
}

#define UngetYYText() \
{ \
  int i ; \
  const char* theText = YYText() ; \
  for (i = YYLeng() - 1; i >= 0; --i) \
    unput (theText[i]) ; \
}

#define UngetString(_s_,_e_) \
{ \
  const char *sp = _e_ ; \
  for (; sp != _s_ ; --sp) \
    unput (*sp) ; \
}

#ifdef DEBUG_LEXER
#define GOTO_STATE(_s_)	{						      \
  switch (_s_)								      \
  {									      \
    case inprefix: printf ("Lexer entering state 'inprefix'\n") ; break ;	      \
    case incmnt: printf ("Lexer entering state 'incmnt'\n") ; break ;		      \
    case inexcl: printf ("Lexer entering state 'inexcl'\n") ; break ;		      \
    case needstat: printf ("Lexer entering state 'needstat'\n") ; break ;	      \
    case instat: printf ("Lexer entering state 'instat'\n") ; break ;		      \
    case needdir: printf ("Lexer entering state 'needdir'\n") ; break ;	      \
    case indir: printf ("Lexer entering state 'indir'\n") ; break ;		      \
    case ifs: printf ("Lexer entering state 'ifs'\n") ; break ;		      \
    case ifend: printf ("Lexer entering state 'ifend'\n") ; break ;		      \
    case inassign: printf ("Lexer entering state 'inassign'\n") ; break ;	      \
    case infmt: printf ("Lexer entering state 'infmt'\n") ; break ;		      \
    case indata: printf ("Lexer entering state 'indata'\n") ; break ;		      \
    case arrays: printf ("Lexer entering state 'arrays'\n") ; break ;		      \
    case arrend: printf ("Lexer entering state 'arrend'\n") ; break ;		      \
    case inio: printf ("Lexer entering state 'inio'\n") ; break ;		      \
    case innameeq: printf ("Lexer entering state 'innameeq'\n") ; break ;	      \
    case chkcomma: printf ("Lexer entering state 'chkcomma'\n") ; break ;	      \
    case INITIAL: printf ("Lexer entering state 'INITIAL'\n") ; break ;	      \
    default: printf ("Lexer entering UNDOCUMENTED state\n") ; break ;		      \
  }									      \
  BEGIN(_s_) ; }
#else
#define GOTO_STATE(_s_) BEGIN(_s_);
#endif // DEBUG_LEXER

static char  ifBuf[2000] ;   // backtracking buffer
static char* ifptr ;         // input pointer for backtracking buffer
static int   parlev ;        // parenthesis nesting level when scanning "if"
static int   in_logif = 0 ;  // 1 if we are processing a logical if statement
static int   if_len ;        // length of conditional keyword (2 or 6)

%}

%x inprefix incmnt inexcl needstat instat needdir indir ifs ifend
%x inassign infmt indata arrays arrend inio innameeq chkcomma
%x indowhile

alpha  [A-Za-z]
dig    [0-9]
oct    [0-7]
hex    [0-9A-Fa-f]
punct  [_\$]
pchar  [0-9A-Za-z!@#\$\%^&\*\(\)_\-\+=\[\{\]\}\;:\'\"\`~\/\??\.><\,]
hdelim \003

do     "do"
if     "if"
else   "else"
goto   "goto"
while  "while"
assign "assign"
to     "to"
func   "function"

other  {alpha}|{dig}|{punct}

dec    {dig}+
octs   {oct}+
hexs   {hex}+

ostr   "'"{octs}"'"
hstr   "'"{hexs}"'"

holl   {hdelim}.+{hdelim}

lnum   {dec}":"

cmnt   ("C"|"c"|"D"|"d"|"!"|"*")

stat   " "

dir    "C""$""D""I""R"

label  " "*{dec}

ctxt   "!".*$

cstr   "'"(\"|[^"'"]|("'""'"))+"'"

qstr   \"("'"|[^\"]|(\"\"))+\"

symb   {alpha}{other}*

%%

<INITIAL>{lnum} {
  if (lptr)
    *lptr = atoi (YYText()) ;
  GOTO_STATE(inprefix)
}

<INITIAL>"\n" {
  return NEW_STATEMENT ;
}

<inprefix>{dir}  {
  GOTO_STATE(needdir)
  return DIRECTIVE ;
}

<inprefix>{cmnt} { 
  GOTO_STATE(incmnt)
}

<inprefix>{label}  { 
  // cout << "LABEL: " << YYText() ; 
  GOTO_STATE(needstat)
  // yystno = atoi (YYText()) ;  // needed by parser
  return LABEL ;
}

<inprefix>{stat} {
  GOTO_STATE(needstat)
  // yystno = 0 ;
  return LABEL ;
}

<inprefix>"\n" {
  GOTO_STATE(INITIAL)
  return BLANK_STATEMENT ;
}

<incmnt>.*/"\n" { 
  GOTO_STATE(INITIAL)
  return COMMENT ;
}


<needstat,instat,indir>"!" {
  GOTO_STATE(inexcl)
}

<needdir,indir>" "|"\t"    /* chew up white space within directives */

<inexcl>.*/"\n"          {
  // YYText() holds the !-delimited comment for this line
  // For the moment we simply discard it.
  GOTO_STATE(INITIAL)
  return SHORT_COMMENT ;
}

<needstat>{do}({dec}","?)?{symb}"=" { 
  /* When we identify a possible "DO" statement we enter
   * a mode which scans forward to see if there an exposed
   * "," further on in the current statement.
   * If there is no exposed "," then it is a LET token, and
   * we accept up to the "=" sign as the identifier,
   * otherwise the "DO" string is a PLAINDO token.
   */

  // start buffering lookahead text into ifBuf

  ifptr = ifBuf ;
  BufferYYText (ifptr) ;
  parlev = 0 ;
  GOTO_STATE(chkcomma)
}

<chkcomma>"(" {
  BufferYYText (ifptr) ;
  ++parlev ;
}

<chkcomma>")" {
  BufferYYText (ifptr) ;
  --parlev ;
  if (parlev < 0) parlev = 0 ;
}

<chkcomma>[^"("")"",""\n"]+ {
  BufferYYText (ifptr) ;
}

<chkcomma>"," {
  if (parlev == 0)
    {
      UngetYYText () ;
      UnBufferYYText (ifptr, ifBuf + 2) ; // consume the "DO"
      GOTO_STATE(instat)
      return PLAINDO ;
    }
  else
    BufferYYText (ifptr) ;
}

<chkcomma>"\n" {
  UngetYYText () ;
  UnBufferYYText (ifptr, ifBuf) ;
  GOTO_STATE(instat)
  return LET ;
}

<needstat>{do}/({dec}","?)?{while}"(".+")"[^=]*"\n" {
  GOTO_STATE(indowhile)
  return DOWHILE ;
}

<indowhile>{while} {
  GOTO_STATE(instat)
  return WHILE ;
}

<needstat>{assign}/{dec}{to}{symb}[!\n] {
  GOTO_STATE(inassign)
  return ASSIGN ;
}

<inassign>{symb} {
  const char *keywd = YYText() ;
  const char *textEnd = &(keywd[YYLeng() - 1]) ;
  int keytok ;
  const char *rest = matchkeyword (keywd, &keytok) ;

  if (keytok == UNKNOWN_TOK)
  {
    return IDENTIFIER ;
  }
  else
    {
      UngetString(rest-1,textEnd) ;
      return keytok ;
    }
}

<needstat>{if}"("       {
  // start copying current YYText into ifBuf
  
  ifptr = ifBuf ;
  BufferYYText (ifptr) ;
  parlev = 1 ;
  if_len = 2 ;
  GOTO_STATE(ifs)
  // cout << "=> state ifs (if)\n" ;
}


<needstat>{else}{if}"(" {
  // start copying current YYText into ifBuf
  
  ifptr = ifBuf ;
  BufferYYText (ifptr) ;
  parlev = 1 ;
  if_len = 6 ;
  GOTO_STATE(ifs)
  // cout << "=> state ifs (elseif)\n" ;
}

<ifs>{cstr}             { /* <ifs>{cstr} */
  BufferYYText (ifptr) ;
}

<ifs>{qstr}             { /* <ifs>{qstr> */
  BufferYYText (ifptr) ;
}

<ifs>"("                { 
  *ifptr++ = '(' ;
  ++parlev ;
  // cout << "after '(' parlev = " << parlev << "\n" ;
}

<ifs>")"                { 
  *ifptr++ = ')' ;
  --parlev ;
  // cout << "after ')' parlev = " << parlev << "\n" ;
  if (parlev == 0)
    {
      GOTO_STATE(ifend)
      // cout << "=> state ifend\n" ;
    }
}

<ifs>"\n"               {
  // this is a lexical error - just stay quiet for now
  parlev = 0 ;
  GOTO_STATE(ifend)
  unput ('\n') ;
}

<ifs>[^"("")""\n"]      { 
  // cout << "buffering '" << YYText() << "'\n" ;
  *ifptr++ = *(YYText()) ;
}

<ifend>"="              { /* return a LET token */ 
  UngetYYText() ;
  UnBufferYYText (ifptr, ifBuf) ;
  GOTO_STATE(instat)
  return LET ;
}

<ifend>{dig}            { 
  if (if_len == 2)
    {
      /* return an ARITHIF token */
      UngetYYText() ;
      UnBufferYYText (ifptr, ifBuf + 2) ;
      GOTO_STATE(instat)
      return ARITHIF ;
    }
  else
    {
      /* return a LET token */ 
      UngetYYText() ;
      UnBufferYYText (ifptr, ifBuf) ;
      GOTO_STATE(instat)
      return LET ;
    }
}

<ifend>[^=0-9\n]          { /* either a logical IF or a block IF/ELSEIF */
  UngetYYText() ;
  UnBufferYYText (ifptr, ifBuf + if_len) ;
  in_logif = 1 ;
  GOTO_STATE(instat)
  if (if_len == 2)
    return LOGICALIF ;
  else
    return ELSEIF ;
}

<ifend>"\n"              { /* an error will occur on this line */
  UngetYYText() ;
  UnBufferYYText (ifptr, ifBuf) ;
  GOTO_STATE(instat)
  return LET ;
}

<needstat>{goto}/{dec}   { 
  // cout << "GOTO: " ;
  GOTO_STATE(instat)
  return PLAINGOTO ;
}

<needstat>{goto}/{symb}"(" {
  // cout << "ASSIGNED_GOTO: " ;
  GOTO_STATE(instat)
  return ASSIGNGOTO ;
}

<needstat>{goto}/"("    { 
  // cout << "COMP_GOTO: " ;
  GOTO_STATE(instat)
  return COMPGOTO ;
}

<instat>{ostr}"O"       {
  // cout << "OCTAL_CONSTANT: " << YYText() ;
  return OCTAL_CONSTANT ;
}

<instat>"O"{ostr}       {
  // cout << "OCTAL_CONSTANT: " << YYText() ;
  return OCTAL_CONSTANT ;
}

<instat>{ostr}"B"       {
  // cout << "OCTAL_CONSTANT: " << YYText() ;
  return OCTAL_CONSTANT ;
}

<instat>"B"{ostr}       {
  // cout << "OCTAL_CONSTANT: " << YYText() ;
  return OCTAL_CONSTANT ;
}

<instat>{hstr}"X"       {
  // cout << "HEX_CONSTANT: " << YYText() ;
  return HEX_CONSTANT ;
}

<instat>"X"{hstr}       {
  // cout << "HEX_CONSTANT: " << YYText() ;
  return HEX_CONSTANT ;
}

<instat>"Z"{hstr}       {
  // cout << "HEX_CONSTANT: " << YYText() ;
  return HEX_CONSTANT ;
}

<instat>{cstr}          {
  // cout << "CSTR: " << YYText() ;
  return CHAR_CONSTANT ;
}

<instat>{qstr}          {
  // cout << "QSTR: " << YYText() ;
  return CHAR_CONSTANT ;
}

<instat>{holl}          {
  return HOLLERITH ;
}

<instat>{dec}/"."{alpha}{alpha}          { return INT_CONSTANT ;  }

<instat>{dec}"."{dec}(E[+-]?{dec})?      { return REAL_CONSTANT ; }
<instat>{dec}"."{dec}([DQ][+-]?{dec})?   { return DP_CONSTANT ; }

<instat>{dec}"."(E[+-]?{dec})            { return REAL_CONSTANT ; }
<instat>{dec}"."([DQ][+-]?{dec})         { return DP_CONSTANT ; }

<instat>{dec}"."/[^a-z0-9]               { return REAL_CONSTANT ; }

<instat>{dec}?"."{dec}(E[+-]?{dec})?     { return REAL_CONSTANT ; }
<instat>{dec}?"."{dec}([DQ][+-]?{dec})?  { return DP_CONSTANT ; }

<instat>{dec}E[+-]?{dec}                 { return REAL_CONSTANT ; }
<instat>{dec}[DQ][+-]?{dec}              { return DP_CONSTANT ; }

<instat,indir,inassign,indowhile>{dec}   { return INT_CONSTANT ;  }

<instat>".eqv."  { return EQV ; }
<instat>".neqv." { return NEQV ; }
<instat>".xor."  { return XOR ; }
<instat>".or."   { return OR ; }
<instat>".and."  { return AND ; }
<instat>".not."  { return NOT ; }
<instat>".lt."   { return LT ; }
<instat>".le."   { return LE ; }
<instat>".eq."   { return EQ ; }
<instat>".ne."   { return NE ; }
<instat>".gt."   { return GT; }
<instat>".ge."   { return GE ; }
<instat>".t."    { return TRUE_VAL ; }
<instat>".true." { return TRUE_VAL ; }
<instat>".f."    { return FALSE_VAL ; }
<instat>".false." { return FALSE_VAL ; }

<needstat>{symb}/"="    { 
  // cout << "SLET: " ; 
  UngetYYText() ;
  GOTO_STATE(instat)
  return LET ;
}

<needstat>{symb}"("    {
  // this could be a lhs array ref (LET) or an I/O statement
  // we must scan ahead to see if there is an "=" character
  // at parlev == 0.
  // Note: "if (" will match with the earlier rule <needstat>{if}"("

  ifptr = ifBuf ;         // Start the buffering of this line
  BufferYYText (ifptr) ;  // so we can re-scan the symbol when
  parlev = 1 ;            // the ambiguity is resolved by 
  GOTO_STATE(arrays)      // lookahead
}

<arrays>{cstr}             { BufferYYText (ifptr) ; }

<arrays>{qstr}             { BufferYYText (ifptr) ; }

<arrays>"("                {
  *ifptr++ = '(' ;
  ++parlev ;
}

<arrays>")"                {
  *ifptr++ = ')' ;
  --parlev ;
  if (parlev == 0)
    GOTO_STATE(arrend)
}

<arrays>"\n"               {
  parlev = 0 ;        // This should only happen if there is syntactic
  GOTO_STATE(arrend)  // error, so this is for recovery purposes only.
  unput ('\n') ;
}

<arrays>[^"("")""\n"]      {
  *ifptr++ = *(YYText()) ;
}

<arrend>"="                { /* return a LET token */
  UngetYYText() ;
  UnBufferYYText (ifptr, ifBuf) ;
  GOTO_STATE(instat)
  return LET ;
}

<arrend>[^=]               {
  // seen a possible io statement

  UngetYYText() ;
  UnBufferYYText (ifptr, ifBuf) ;
  GOTO_STATE(inio)
  // now re-scan the original symbol and match it
  // against the known set of keywords
}

<inio>{symb} { 

  /* This is the keyword of a non-assignment
   * statement, either  because it is followed by a 
   * '(' and the <arrays> lookahead state machine
   * determined that this is not an array assignment.
   */

  const char *keywd = YYText() ;
  const char *textEnd = &(keywd[YYLeng() - 1]) ;
  int keytok ;
  const char *rest = matchkeyword (keywd, &keytok) ;

  // fgg added 'TO' on 2000/12/05
  if ((keytok != UNKNOWN_TOK) && (keytok != TO))
    {
      // cout << "KEYWD [" << keytok << "] (leaves '" << rest << "')\n" ;
      UngetString(rest-1,textEnd) ;
    }
  else
    {
      // cout << "UNKNOWN '" YYText() << "' " ;
      UngetYYText() ;
      GOTO_STATE(instat)
      keytok = LET ;
    }

  // currently we deal with format statements specially

  if (keytok == FORMAT_TOKEN)
    GOTO_STATE(infmt)
  else if (keytok == DATA)
    GOTO_STATE(indata)
  else
    GOTO_STATE(instat)

  return keytok ;
}

<needstat>{symb}/[^(= ]   { 
  /* This is the keyword of a non-assignment
   * statement  because a symbol is followed by a
   * character other than a '(', ' ', or '='.
   */

  const char *keywd = YYText() ;
  const char *textEnd = &(keywd[YYLeng() - 1]) ;
  int keytok ;
  const char *rest = matchkeyword (keywd, &keytok) ;

  // fgg added 'TO' on 2000/12/05
  if ((keytok != UNKNOWN_TOK) && (keytok != TO))
    {
      UngetString(rest-1,textEnd) ;
    }
  else
    {
      // fprintf (stderr, "faulty keyword '%s'\n", keywd) ;
    }

  // currently we deal with format statements specially

  if (keytok == FORMAT_TOKEN)
    GOTO_STATE(infmt)
  else if (keytok == DATA)
    GOTO_STATE(indata)
  else
    GOTO_STATE(instat)

  return keytok ;
}


<needdir>{symb}           {
  const char *keywd = YYText() ;
  const char *textEnd = &(keywd[YYLeng() - 1]) ;
  int keytok ;
  const char *rest = matchdirective (keywd, &keytok) ;

  // fgg added 'TO' on 2000/12/05
  if ((keytok != UNKNOWN_TOK) && (keytok != TO))
    {
      // cout << "DIRNAME [" << keytok << "] (leaves '" << rest << "')\n" ;
      UngetString(rest-1,textEnd) ;
      GOTO_STATE(indir)
    }
  else
    {
      // cout << "UNKNOWN DIRECTIVE '" YYText() << "' " ;
      // for error recovery we slip silently into a comment
      GOTO_STATE(incmnt)
    }

  return keytok ;
}

<instat>{func}/{symb}("()"|"("{alpha}|"*"|"\n") {
  return FUNCTION ;
}

<instat>{symb}/"=" {
  if (inioctl)
    {
      GOTO_STATE(innameeq)
      return NAMEEQ ;
    }
  else
    return IDENTIFIER ;
}

<innameeq>"=" { GOTO_STATE(instat) }

<instat,indir>{symb}          {
  if (*needkwd)
    {
      const char *keywd = YYText() ;
      const char *textEnd = &(keywd[YYLeng() - 1]) ;
      int keytok ;
      const char *rest = matchkeyword (keywd, &keytok) ;
      
      if ((keytok == UNKNOWN_TOK) || (keytok == TO))
	return IDENTIFIER ;
      else
	{
	  UngetString(rest-1,textEnd) ;
	  *needkwd = 0 ;

	  return keytok ;
	}
    }
  else
    return IDENTIFIER ;
}

<instat,indir>"("       { 
  // cout << "(" ;
  ++parlev ;
  return LEFTPAR ;
}

<instat,indir>")"       {
  // cout << ")" ;
  --parlev ;
  if (parlev == 0)
    {
     if (in_logif)
       {
         in_logif = 0 ;       // logical IF must be followed by
         GOTO_STATE(needstat) // a statement
       }
   }
  return RIGHTPAR ;
}

<instat>"*""*"          {
  // cout << "DASTER " ;
  return DASTER ;
}

<instat,indir>"*"       {
  // cout << "ASTER " ;
  return ASTER ;
}

<instat>"/""/"          {
  // cout << "DSLASH " ;
  return DSLASH ;
}

<instat>"/"             {
  // cout << "SLASH " ;
  return SLASH ;
}

<instat>"+"             {
  // cout << "PLUS " ;
  return PLUS ;
}

<instat>"-"             {
  // cout << "MINUS " ;
  return MINUS ;
}

<instat>":"             {
  // cout << "COLON " ;
  return COLON ;
}

<instat>"%"             { // or prefix {symb} with this to get special-symb ?
  // cout << "PERCENT " ;
  return PERCENT ;
}

<instat,indowhile>","   {
  // cout << "COMMA " ;
  return COMMA ;
}

<indir>","              {
  // cout << "COMMA " ;
  if (parlev == 0)
    GOTO_STATE(needdir)
  return COMMA ;
}

<instat>"="            {
  // cout << "EQUAL " ;
  return EQUAL ;
}

<infmt,indata>[^\n]*  {
  return CHAR_CONSTANT  ;  // THIS SHOULD BE TEMPORARY
}

<instat,incmnt,indir,infmt,indata,inassign,needdir>"\n" {
  parlev = 0 ;
  GOTO_STATE(INITIAL)
  return NEW_STATEMENT ;
}

<instat>[ \t] { /* consume white space where expected */ }

\0  { return 0 ; }

<*>.|"\n"                 {
/*
  int lno = lptr ? *lptr : 0 ;
  fprintf (stderr, "line %d, start = %d, character = '%s'\n",
	   lno, YY_START, YYText()) ;
*/
}

%%

int yywrap () { return 1 ; }


