//
// Working with AST
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

#include "Octave/Fortran/FortranParser.h"
#include "Octave/Base/ParseTree.h"
#include "Octave/Base/Unparse.h"

#include "misc.h"

#include <stdlib.h>

static FortranParser* parser=NULL;
static ofstream fo;
static Statement* currentFunc=NULL;
static Statement* currentCP=NULL;	// pointer to current control parent

static int currentLine=0;		// the current source line number

static int frmt=0;
static int fnd=0;
static int lp=0;

static int loop_corr=0;

static char eos_var_name_tmp[]="eos_tmp";
static long eos_var_seq=0;

char array1[128];

void        MiscTreeProcessBlockOfStatements(StmtList*, int);
void        MiscTreeExtractReferences (Expression**, int);
void        print_frmt(int);
int         MiscPadProcessBlockOfStatements(StmtList*, int, char*, long, int);
int         MiscPadExtractReferences (Expression**, char*, long, int);
int         MiscPadExtractReferences1 (Expression**, char*, long, int);
int         MiscGirProcessBlockOfStatements(StmtList*, int, char*, long, long);
int         MiscGirExtractReferences (Expression**, char*, long, long);
void        PerfectLoopNest(StmtList*, int*);
int         MiscLoopProcessBlockOfStatements(StmtList*, int, long, int, char*, char*);
Statement*  NextPerfectLoopNest(Statement*, int*);
Statement*  DuplicateStmt(Statement*);
Expression* DuplicateExpr(Expression*);
Symbol*     NewTemporaryVar(Datatype*, char*);
bool        isConst(Expression*);
double      getValue(Expression*);
void        addZero(char*);
void        ChangeExprInStmt(Statement*, Expression*, Expression*);
void        ChangeExpr(Expression*, Expression*, Expression*);
void        MiscCountLoops(StmtList*, int*);

Expression* make_int_const(ParseContext* pc, int val);
Expression* make_float_const(ParseContext* pc, string* val);

#define X_ARRAY_DECL 0
#define X_ARRAY_COMM 1

int MiscParse(char* str_file_in)
{
  parser=NULL;
  currentFunc=NULL;
  currentCP=NULL;

  currentLine=0;

  frmt=0;
  fnd=0;
  lp=0;

  eos_var_seq=0;

  //***

  string* fname=new string(str_file_in);

  eos_var_seq=0;
  loop_corr=0;

  parser=new FortranParser(*fname);

  return parser->parse();
}

int MiscErrors(char* str)
{
  string* msg;

  slist_iterator<char*, string>* next_error;

  next_error=new slist_iterator<char*, string> (parser->errors());

  ofstream fout(str);
  if (!fout) return -1;

  while (msg=(*next_error)())
  {
    fout << (*msg)();
  }

  fout.close();

  return 0;
}

int MiscUnparse(char* str)
{
  StmtListIterator* next_routine;
  PTR_STMT proc;

  next_routine=new StmtListIterator(parser->fileRoot()->then_part());

  ofstream fout(str);
  if (!fout) return -1;

  while (proc=(*next_routine)())
  {
    SourceListing list;
    list.unparse (proc, ANSII_F77);
    
    SourceListingIterator next_line(list);
    SourceLine* line;
  
    while (line=next_line())
    {
      fout << (*line)();
    }
  }

  fout.close();

  return 0;
}

void MiscClean(void)
{
  delete parser;
}

int MiscCorrect(void)
{
  int loop_corr1=loop_corr;
  loop_corr=0;
  return loop_corr1;
}

int MiscTree(char* str)
{
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;

  fo.open(str);
  if (!fo) return -1;

  while ((func = next_routine()))
  {
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        parser->push_scope(func->scope);
  
        char* procName = func->symbol()->ident();

        fo << "procedure '" << procName << "'\n";
        frmt=1;

        MiscTreeProcessBlockOfStatements(func->statements(), 0);

        fo << "end\n";
    
        int firstLine = func->g_line;
        int noLines = func->statements()->tail()->g_line - firstLine + 1;
  
        break;
      }
      default:
      {
        break;
      }
    }
  }

  fo.close();

  return 0;
}

void MiscTreeProcessBlockOfStatements(StmtList* block, 
                                      int nestLevel)
{
  if (!block)
    return;

  StmtListIterator next_stmt(block);
  Statement* stmt;
  while (stmt=next_stmt.head()) 
  {
    statement_tag stag = stmt->stmt_variant();
    currentLine = stmt->g_line;

    switch (stag)
    {
      case FOR_NODE:
      case PARDO_NODE:
      case PDO_NODE:
      {
        print_frmt(frmt);
        frmt++;

        int nest=0;

        int endLine = stmt->loop_body()->tail()->g_line;

        PerfectLoopNest(stmt->loop_body(), &nest); //collected into nest
        fo << "loop '" << stmt->symbol()->ident() << "', '" 
           << nest << "', '" << currentLine << "', '"
           << endLine << "'\n";

	MiscTreeProcessBlockOfStatements(stmt->loop_body(), nestLevel+1);

        frmt--;
        print_frmt(frmt);
        fo << "end\n";

	next_stmt();
	break;
      }
      case LOGIF_NODE:
      {
        // Have to turn this into an IF_NODE otherwise we cannot transform the
	// the conditional statement. Then we drop through to the IF_NODE
	// case and complete the transformation.
	  
	stmt->stmt_variant (IF_NODE);
	Statement* condStat = stmt->node.if_node.true_branch;
	StmtList* then_pt = new StmtList();
	then_pt->append (condStat);
	then_pt->append (parser->make_stmt (CONTROL_END, 0, 0, 0, 0));
	stmt->node.fields.bl_ptr1 = then_pt;
	stmt->node.fields.bl_ptr2 = 0;
      }
      case IF_NODE:
      {
	MiscTreeProcessBlockOfStatements (stmt->then_part(), 0);
	MiscTreeProcessBlockOfStatements (stmt->else_part(), 0);
	break;
      }
      case ELSEIF_NODE:
      {
	MiscTreeProcessBlockOfStatements (stmt->then_part(), 0);
	MiscTreeProcessBlockOfStatements (stmt->else_part(), 0);
	break;
      }
      case COMM_STAT:
      {
        if (stmt->node.fields.e1)
          MiscTreeExtractReferences(&(stmt->node.fields.e1), X_ARRAY_COMM);

        break;
      }
      case VAR_DECL:
      case DIMENSION_OP:
      case DIM_STAT:
      {
        if (stmt->node.fields.e1)
          MiscTreeExtractReferences(&(stmt->node.fields.e1), X_ARRAY_DECL);

        break;
      }
    }
    if (stag != FOR_NODE) next_stmt();
    // nestLevel = 0 ; // this numbers only for perfectly-nested loops
  }
}

void print_frmt(int frmt)
{
  int i=0;

  for (i=0; i<frmt; i++)
  {
    fo << ' ';
  }
}

void MiscTreeExtractReferences(Expression** exptr, int type)
{
  if (!exptr || !(*exptr)) return;

  Expression* expr = *exptr;

  switch (expr->expr_variant())
  {
    case ARRAY_REF:
    {
      Symbol* varSym = (*exptr)->referenced_symbol();
      char* varName = varSym->ident();

      print_frmt(frmt);
      if (type==X_ARRAY_COMM) fo << "array_comm";
      else fo << "array_decl";
      fo << " '" << varName << "'\n";

      break;
    }

    case COMM_LIST:
    {
      // standard recursive list processing...
      if (expr->node.fields.e1)
        MiscTreeExtractReferences (&(expr->node.fields.e1), type);
      if (expr->node.fields.e2)
        MiscTreeExtractReferences (&(expr->node.fields.e2), type);
      break;
    }

    // Traverse all expressions in an expression list
    case EXPR_LIST:
    {
      // standard recursive list processing...
      if (expr->node.fields.e1)
        MiscTreeExtractReferences (&(expr->node.fields.e1), type);
      if (expr->node.fields.e2)
        MiscTreeExtractReferences (&(expr->node.fields.e2), type);
      break;
    }
  }
}

int MiscArraypad(int sockfd, char* proc, char* array, char* p1)
{
  int ret=0;
  int fake=0; // if array=fgg then do not pad, but increase parameter fggx

  long padFactor=atol(p1);

  if (padFactor<0)
  {
    SendMes(sockfd, "Error: Pad factor should be >= 0!", 0, 1024); 
    return -1;
  }

  if (strlen(array)>3) 
  {
    if ((array[0]=='f') && (array[1]=='g') && (array[2]=='g'))
    {
      fake=1;
      strcpy(array1, array);
      strcat(array1,"x");
    }
  }

  //Searching procedure
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;
  int found=0;

  while (func = next_routine())
  {
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        char* procName = func->symbol()->ident();
        if (strcmp(procName, proc)==0) found=1;
        break;
      }
      default:
      {
        break;
      }
    }

    if (found==1) break;
  }

  if (found==0)
  {
    SendMes(sockfd, "Error: Procedure was not found!", 0, 1024); 
    return -1;
  }

  ret=MiscPadProcessBlockOfStatements(func->statements(), 0, array, padFactor, fake);

  if (ret==0)
  {
    SendMes(sockfd, "Error: Array was not found!", 0, 1024); 
    return -1;
  }
  else if (ret==-1)
  {
    SendMes(sockfd, "Error: Can't deal with assumed size dimension!", 0, 1024); 
    return -1;
  }

  return 0;
}

int MiscPadProcessBlockOfStatements(StmtList* block, 
                                    int nestLevel,
                                    char* array,
                                    long padFactor,
                                    int fake)
{  
  int ret=0;

  if (!block) return 0;

  StmtListIterator next_stmt(block);
  Statement* stmt;
  while (stmt=next_stmt.head()) 
  {
    statement_tag stag = stmt->stmt_variant();
    currentLine = stmt->g_line;

    if ((fake==1) && (stag==PARAM_DECL))
    {
      ret=MiscPadExtractReferences1(&(stmt->node.fields.e1), array, padFactor, fake);
      if (ret!=0) return ret;
    }

    switch (stag)
    {
      case FOR_NODE:
      case PARDO_NODE:
      case PDO_NODE:
      case WHILE_NODE:
      {
	MiscPadProcessBlockOfStatements(stmt->loop_body(), 
                                        nestLevel+1, array, 
                                        padFactor, fake);

	next_stmt();
	break;
      }
      case LOGIF_NODE:
      {
        // Have to turn this into an IF_NODE otherwise we cannot transform the
	// the conditional statement. Then we drop through to the IF_NODE
	// case and complete the transformation.
	  
	stmt->stmt_variant (IF_NODE);
	Statement* condStat = stmt->node.if_node.true_branch;
	StmtList* then_pt = new StmtList();
	then_pt->append (condStat);
	then_pt->append (parser->make_stmt (CONTROL_END, 0, 0, 0, 0));
	stmt->node.fields.bl_ptr1 = then_pt;
	stmt->node.fields.bl_ptr2 = 0;
        break;
      }
      case IF_NODE:
      {
	MiscPadProcessBlockOfStatements (stmt->then_part(), 0, array, padFactor, fake);
	MiscPadProcessBlockOfStatements (stmt->else_part(), 0, array, padFactor, fake);
	break;
      }
      case ELSEIF_NODE:
      {
	MiscPadProcessBlockOfStatements (stmt->then_part(), 0, array, padFactor, fake);
	MiscPadProcessBlockOfStatements (stmt->else_part(), 0, array, padFactor, fake);
	break;
      }
      case COMM_STAT:
      {
        if ((fake==0) && stmt->node.fields.e1)
        {
          ret=MiscPadExtractReferences(&(stmt->node.fields.e1), array, padFactor, fake);
          if (ret!=0) return ret;
        }

        break;
      }
      case VAR_DECL:
      case DIMENSION_OP:
      case DIM_STAT:
      {
        if ((fake==0) && stmt->node.fields.e1)
        {
          ret=MiscPadExtractReferences(&(stmt->node.fields.e1), array, padFactor, fake);
          if (ret!=0) return ret;
        }

        break;
      }
    }
    if (stag != FOR_NODE) next_stmt();
  }

  return 0;
}

int MiscPadExtractReferences(Expression** exptr, char* array, long padFactor, int fake)
{
  int ret=0;

  if (!exptr || !(*exptr)) return 0;

  Expression* expr = *exptr;

  switch (expr->expr_variant())
  {
    case ARRAY_REF:
    {
      Symbol* varSym = (*exptr)->referenced_symbol();
      if (varSym)
      {
        Datatype* t=varSym->type;
        char* varName = varSym->ident();

        if ((strcmp(array, varName)==0) && t && t->type_variant() == T_ARRAY)
        {
          Expression* expr0 = expr->node.array_ref.index;

          if (expr0->node.fields.e1->expr_variant()==STAR_RANGE)
          {
            return -1;
          }
  
          ParseContext* pc = varSym->pc;
          Expression* expr1 = make_int_const(pc, padFactor);

          if (expr0->node.fields.e1->expr_variant()==DDOT)
          {
            expr0=expr0->node.fields.e1;
            if (!expr0->node.fields.e2) 
            {
              return -1;
            }

            Expression* expr9 = pc->make_expr(ADD_OP, expr0->node.fields.e2, expr1, 0);
            expr0->node.fields.e2 = expr9;
          }
          else
          {
            Expression* expr9 = pc->make_expr(ADD_OP, expr0->node.fields.e1, expr1, 0);
            expr0->node.fields.e1 = expr9;
          }

          return 1;
        }

        break;
      }
    }
      
    case COMM_LIST:
    {
      // standard recursive list processing...
      if (expr->node.fields.e1)
      {
        ret=MiscPadExtractReferences (&(expr->node.fields.e1), array, padFactor, fake);
        if (ret!=0) return ret;
      }
      if (expr->node.fields.e2)
      {
        ret=MiscPadExtractReferences (&(expr->node.fields.e2), array, padFactor, fake);
        if (ret!=0) return ret;
      }
      break;
    }

    // Traverse all expressions in an expression list
    case EXPR_LIST:
    {
      // standard recursive list processing...
      if (expr->node.fields.e1)
      {
        ret=MiscPadExtractReferences (&(expr->node.fields.e1), array, padFactor, fake);
        if (ret!=0) return ret;
      }
      if (expr->node.fields.e2)
      {
        ret=MiscPadExtractReferences (&(expr->node.fields.e2), array, padFactor, fake);
        if (ret!=0) return ret;
      }
      break;
    }
  }

  return 0;
}

int MiscPadExtractReferences1(Expression** exptr, char* array, long padFactor, int fake)
{
  int ret=0;

  if (!exptr || !(*exptr)) return 0;

  Expression* expr = *exptr;

  switch (expr->expr_variant())
  {
    case CONST_REF:
    {
      Symbol* varSym = (*exptr)->node.fields.symbol;
      if (varSym)
      {
        Datatype* t=varSym->type;
        char* varName = varSym->ident();
        if ((strcmp(array1, varName)==0))
        {
          Expression* expr0 = varSym->node.const_value;

          ParseContext* pc = varSym->pc;
          Expression* expr1 = make_int_const(pc, padFactor);
  
          varSym->node.const_value = expr1;

          return 1;
        }
      }
      cout << "\n";
    }

    // Traverse all expressions in an expression list
    case EXPR_LIST:
    {
      // standard recursive list processing...
      if (expr->node.fields.e1)
      {
        ret=MiscPadExtractReferences1 (&(expr->node.fields.e1), array, padFactor, fake);
        if (ret!=0) return ret;
      }
      if (expr->node.fields.e2)
      {
        ret=MiscPadExtractReferences1 (&(expr->node.fields.e2), array, padFactor, fake);
        if (ret!=0) return ret;
      }
      break;
    }
  }

  return 0;
}

int MiscArraygir(int sockfd, char* proc, char* array, char* p1, char* p2)
{
  int ret=0;
  fnd=0;
  long girFactor=atol(p1);

  long girAdd=1;
  if (strcmp(p2, "")!=0) girAdd=atol(p2);

  if (girFactor<=0)
  {
    SendMes(sockfd, "Error: GIR factor should be > 0!", 0, 1024); 
    return -1;
  }
  if (girAdd<=0)
  {
    SendMes(sockfd, "Error: GIR parameter should be > 0!", 0, 1024); 
    return -1;
  }

  //Searching procedure
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;
  int found=0;

  while (func = next_routine())
  {
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        char* procName = func->symbol()->ident();
        if (strcmp(procName, proc)==0) found=1;
        break;
      }
      default:
      {
        break;
      }
    }

    if (found==1) break;
  }

  if (found==0)
  {
    SendMes(sockfd, "Error: Procedure was not found!", 0, 1024); 
    return -1;
  }

  ret=MiscGirProcessBlockOfStatements(func->statements(), 0, array, girFactor, girAdd);

  if (ret==-1)
  {
    SendMes(sockfd, "Error: GIR Factor+1 is outside dimensions of the array!", 0, 1024); 
    return -1;
  }

  if (fnd==0)
  {
    SendMes(sockfd, "Error: Array was not found!", 0, 1024); 
    return -1;
  }

  return 0;
}

int MiscGirProcessBlockOfStatements(StmtList* block, 
                                    int nestLevel,
                                    char* array,
                                    long girFactor,
                                    long girAdd)
{  
  int ret=0;

  if (!block) return 0;

  StmtListIterator next_stmt(block);
  Statement* stmt;
  while (stmt=next_stmt.head()) 
  {
    statement_tag stag = stmt->stmt_variant();
    currentLine = stmt->g_line;
    switch (stag)
    {
      case FOR_NODE:
      case PARDO_NODE:
      case PDO_NODE:
      {
        if ((stmt->stmt_variant () == PARDO_NODE) || 
            (stmt->stmt_variant () == PDO_NODE))
        {
          if (stmt->node.for_node.where_cond)
          {
            ret=MiscGirExtractReferences(&(stmt->node.for_node.where_cond), array, girFactor, girAdd);
            if (ret!=0) return ret;
          }
        }

        if (stmt->node.for_node.range->node.binary_op.l_operand)
        {
          ret=MiscGirExtractReferences(&(stmt->node.for_node.range->node.binary_op.l_operand), array, girFactor, girAdd);
          if (ret!=0) return ret;
        }
        if (stmt->node.for_node.range->node.binary_op.r_operand)
        {
          ret=MiscGirExtractReferences(&(stmt->node.for_node.range->node.binary_op.r_operand), array, girFactor, girAdd);
          if (ret!=0) return ret;
        }
        if (stmt->node.for_node.increment) 
        {
          ret=MiscGirExtractReferences(&(stmt->node.for_node.increment), array, girFactor, girAdd);
          if (ret!=0) return ret;
        }

	MiscGirProcessBlockOfStatements(stmt->loop_body(), 
                                        nestLevel+1, array, 
                                        girFactor, girAdd);

	next_stmt();
	break;
      }
      case WHILE_NODE:
      {

        if (stmt->node.while_node.condition)
        {
          ret=MiscGirExtractReferences(&(stmt->node.while_node.condition), array, girFactor, girAdd);
          if (ret!=0) return ret; 
        }

	MiscGirProcessBlockOfStatements(stmt->loop_body(), 
                                        nestLevel+1, array, 
                                        girFactor, girAdd);

	next_stmt();
	break;
      }
      case LOGIF_NODE:
      {
        // Have to turn this into an IF_NODE otherwise we cannot transform the
	// the conditional statement. Then we drop through to the IF_NODE
	// case and complete the transformation.
	  
	stmt->stmt_variant (IF_NODE);
	Statement* condStat = stmt->node.if_node.true_branch;
	StmtList* then_pt = new StmtList();
	then_pt->append (condStat);
	then_pt->append (parser->make_stmt (CONTROL_END, 0, 0, 0, 0));
	stmt->node.fields.bl_ptr1 = then_pt;
	stmt->node.fields.bl_ptr2 = 0;
      }
      case IF_NODE:
      {
        if (stmt->node.if_node.condition)
        {
          ret=MiscGirExtractReferences(&(stmt->node.if_node.condition), array, girFactor, girAdd);
          if (ret!=0) return ret;
        }

	MiscGirProcessBlockOfStatements (stmt->then_part(), 0, array, girFactor, girAdd);
	MiscGirProcessBlockOfStatements (stmt->else_part(), 0, array, girFactor, girAdd);
	break;
      }
      case ELSEIF_NODE:
      {
	MiscGirProcessBlockOfStatements (stmt->then_part(), 0, array, girFactor, girAdd);
	MiscGirProcessBlockOfStatements (stmt->else_part(), 0, array, girFactor, girAdd);
	break;
      }
      case COMM_STAT:
      {
        if (stmt->node.fields.e1)
        {
          ret=MiscGirExtractReferences(&(stmt->node.fields.e1), array, girFactor, girAdd);
          if (ret!=0) return ret;
        }

        break;
      }
      case VAR_DECL:
      case DIMENSION_OP:
      case DIM_STAT:
      {
        if (stmt->node.fields.e1)
        {
          ret=MiscGirExtractReferences(&(stmt->node.fields.e1), array, girFactor, girAdd);
          if (ret!=0) return ret;
        }

        break;
      }
      default:
      {
        if (stmt->node.fields.e1)
        {
          ret=MiscGirExtractReferences(&(stmt->node.fields.e1), array, girFactor, girAdd);
          if (ret!=0) return ret;
        }
        if (stmt->node.fields.e2)
        {
          ret=MiscGirExtractReferences(&(stmt->node.fields.e2), array, girFactor, girAdd);
          if (ret!=0) return ret;
        }

        break;
      }
    }
    if (stag != FOR_NODE) next_stmt();
  }

  return 0;
}

int MiscGirExtractReferences(Expression** exptr, char* array, long girFactor, long girAdd)
{
  int ret=0;

  if (!exptr || !(*exptr)) return 0;

  Expression* expr = *exptr;

  switch (expr->expr_variant())
  {
    case ARRAY_REF:
    {
      Symbol* varSym = (*exptr)->referenced_symbol();
      Datatype* t=varSym->type;
      char* varName = varSym->ident();

      if ((strcmp(array, varName)==0) && t && t->type_variant() == T_ARRAY)
      {
        int ndims = t->node.ar_decl.num_dimensions ;
        if (girFactor+girAdd>ndims) return -1;

        //iterate through dimension expressions  
        Expression* d1=expr->node.fields.e1;
        int i;

        for (i=1; i<girFactor; i++)
        {
          d1 = d1->node.fields.e2;
        }

        Expression* d=d1;

        for (i=1; i<=girAdd; i++)
        {
          d1 = d1->node.fields.e2;
        }

        //swap index & index+1
        Expression* temp=d->node.fields.e1;

        d->node.fields.e1=d1->node.fields.e1;
        d1->node.fields.e1=temp;

        fnd=1;
      }

      break;
    }

    case COMM_LIST:
    {
      // standard recursive list processing...
      if (expr->node.fields.e1)
      {
        ret=MiscGirExtractReferences (&(expr->node.fields.e1), array, girFactor, girAdd);
        if (ret!=0) return ret;
      }
      if (expr->node.fields.e2)
      {
        ret=MiscGirExtractReferences (&(expr->node.fields.e2), array, girFactor, girAdd);
        if (ret!=0) return ret;
      }
      break;
    }

    // Traverse all expressions in an expression list
    default:
    {
      // standard recursive list processing...
      if (expr->node.fields.e1)
      {
        ret=MiscGirExtractReferences (&(expr->node.fields.e1), array, girFactor, girAdd);
        if (ret!=0) return ret;
      }
      if (expr->node.fields.e2)
      {
        ret=MiscGirExtractReferences (&(expr->node.fields.e2), array, girFactor, girAdd);
        if (ret!=0) return ret;
      }
      break;
    }
  }

  return 0;
}

void PerfectLoopNest(StmtList* block, int* nest)
{
  if (!block) return;

  StmtListIterator next_stmt(block);
  Statement* stmt;

  stmt=next_stmt.head();
  statement_tag stag = stmt->stmt_variant();

  if ((stag==FOR_NODE) || (stag==PARDO_NODE) || (stag==PDO_NODE))
  {
    next_stmt();
    Statement* stmt1=next_stmt.head();

    if (!stmt1 || (stmt1 && ((stmt1->stmt_variant()==CONT_STAT) ||
                (stmt1 && stmt1->stmt_variant()==CONTROL_END))))
    {
      (*nest)++;
      PerfectLoopNest(stmt->loop_body(), nest);
    }
  }
}

int MiscLoopInterchange(int sockfd, char* proc, char* sloop, char* p1)
{
  int ret=0;
  fnd=0;
  long loop=atol(sloop);
  long index=atol(p1);

  if (loop<=0)
  {
    SendMes(sockfd, "Error: Loop should be > 0!", 0, 1024); 
    return -1;
  }
  if (index<=0)
  {
    SendMes(sockfd, "Error: Loop interchange parameter should be > 0!", 0, 1024); 
    return -1;
  }

  //Searching procedure
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;
  int found=0;

  while (func = next_routine())
  {
    currentCP=func;
    currentFunc=func;
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        char* procName = func->symbol()->ident();
        if (strcmp(procName, proc)==0) found=1;
        break;
      }
      default:
      {
        break;
      }
    }

    if (found==1) break;
  }

  if (found==0)
  {
    SendMes(sockfd, "Error: Procedure was not found!", 0, 1024); 
    return -1;
  }

  lp=0;
  ret=MiscLoopProcessBlockOfStatements(func->statements(), 0, loop, 2, p1, NULL);

  if (ret==0)
  {
    SendMes(sockfd, "Error: Loop was not found!", 0, 1024); 
    return -1;
  }
  else if (ret==-1)
  {
    SendMes(sockfd, "Error: Loop interchange parameter is outside loop nest!", 0, 1024); 
    return -1;
  }

  return 0;
}

int MiscLoopProcessBlockOfStatements(StmtList* block, 
                                     int nestLevel,
                                     long loop,
                                     int tr,
                                     char* sp1,
                                     char* sp2)
{  
  Statement* oldCP;

  int ret=0;

  if (!block) return 0;

  StmtListIterator next_stmt(block);
  Statement* stmt;
  while (stmt=next_stmt.head()) 
  {
    statement_tag stag = stmt->stmt_variant();
    currentLine = stmt->g_line;
    switch (stag)
    {
      case FOR_NODE:
      case PARDO_NODE:
      case PDO_NODE:
      {
        lp++;
        if (lp==loop)
        {
          int nest=0;
          PerfectLoopNest(stmt->loop_body(), &nest); //collected into nest

          loop_corr=0;

          if (tr==1)
          {
            //************************************************************
            //************************************************************
            //*********************** loop tiling ************************
            //************************************************************
            //************************************************************

            for (int n=0; n<=nest; n++)
            {
              lp=0;  
              int ret1=MiscLoopProcessBlockOfStatements(currentFunc->statements(), 0, loop+n, 4, sp1, NULL);
              if (ret1<=0) return ret1;

              if (nest>0)
              {
                for (int q=0; q<nest; q++)
                {
                  lp=0;  
                  int ret1=MiscLoopProcessBlockOfStatements(currentFunc->statements(), 0, loop+n+q+1, 2, "1", NULL);
                  if (ret1<=0) return ret1;
                }
              }
            }

            loop_corr=nest+1;
          }
          else if (tr==2)
          {
            //************************************************************
            //************************************************************
            //********************* loop interchange *********************
            //************************************************************
            //************************************************************

            int nest1=atol(sp1);

            if (nest1>nest) return -1;

            Statement* stmt1=NextPerfectLoopNest(stmt, &nest1);

            PTR_STMT       tempa;
            PTR_SYMB       tempb;
            PTR_EXPR       tempc;
            PTR_EXPR       tempd;
            PTR_EXPR       tempe;
            PTR_FTN_LABEL  tempf;

            tempa=stmt->node.for_node.for_end;
            tempb=stmt->node.for_node.control_var;
            tempc=stmt->node.for_node.range;
            tempd=stmt->node.for_node.increment;
            tempe=stmt->node.for_node.where_cond;
            tempf=stmt->node.for_node.doend;

            stmt->node.for_node.for_end    =stmt1->node.for_node.for_end;
            stmt->node.for_node.control_var=stmt1->node.for_node.control_var;
            stmt->node.for_node.range      =stmt1->node.for_node.range;
            stmt->node.for_node.increment  =stmt1->node.for_node.increment;
            stmt->node.for_node.where_cond =stmt1->node.for_node.where_cond;
            stmt->node.for_node.doend      =stmt1->node.for_node.doend;

            stmt1->node.for_node.for_end    =tempa;
            stmt1->node.for_node.control_var=tempb;
            stmt1->node.for_node.range      =tempc;
            stmt1->node.for_node.increment  =tempd;
            stmt1->node.for_node.where_cond =tempe;
            stmt1->node.for_node.doend      =tempf;
          }
          else if (tr==3)
          {
            //************************************************************
            //************************************************************
            //********************** loop unrolling **********************
            //************************************************************
            //************************************************************

            int param=atol(sp1);
            char buf[1024];

            Expression* e1=NULL;
            Expression* e2=NULL;

            strcpy(buf, stmt->symbol()->ident());
            strcat(buf, "_x");
            Symbol* tsym = NewTemporaryVar (stmt->symbol()->type, buf);
            tsym->dovar=1;

            Expression* e4=parser->make_expr(VAR_REF, 0, 0, stmt->symbol());

            Expression* ex3=stmt->node.for_node.increment;

            if (!ex3)
            {
              //Increment=1
              string* x1=new string(sp1);

              e2=make_float_const(parser, x1);

              sprintf(buf, "%i", param-1);
              string* x2=new string(buf);

              Expression* e1x=make_float_const(parser, x2);
              Expression* e1y=DuplicateExpr(stmt->node.for_node.range->node.binary_op.r_operand);

              e1=parser->intrinsic_op_node("-", SUBT_OP, e1y, e1x);
            }       	
            else if (isConst(ex3))
            {
              //Increment=const
              int ex3val=(int) getValue(ex3);

              sprintf(buf, "%i", param*ex3val);
              string* x1=new string(buf);

              e2=make_float_const(parser, x1);

              sprintf(buf, "%i", (param-1)*ex3val);
              string* x2=new string(buf);

              Expression* e1x=make_float_const(parser, x2);
              Expression* e1y=DuplicateExpr(stmt->node.for_node.range->node.binary_op.r_operand);

              e1=parser->intrinsic_op_node("-", SUBT_OP, e1y, e1x);
            }
            else
            {
              //Increment=var
              string* x1=new string(sp1);

              Expression* e2x=make_float_const(parser, x1);
              Expression* e2y=DuplicateExpr(stmt->node.for_node.increment);

              e2=parser->intrinsic_op_node("*", MULT_OP, e2x, e2y);

              sprintf(buf, "%i", param-1);
              string* x2=new string(buf);

              Expression* e1x=make_float_const(parser, x2);
              Expression* e1y=DuplicateExpr(ex3);

              Expression* e1z=parser->intrinsic_op_node("*", MULT_OP, e1x, e1y);

              Expression* e1q=DuplicateExpr(stmt->node.for_node.range->node.binary_op.r_operand);

              e1=parser->intrinsic_op_node("-", SUBT_OP, e1q, e1z);
            }

            //check number of loops inside stmt
            int loops=0;
            MiscCountLoops(stmt->loop_body(), &loops);
            loop_corr=1+loops*param;

            //duplicate loop statement
            Statement* nstmt1=DuplicateStmt(stmt);
            nstmt1->enforce_enddo();
            next_stmt.insert (nstmt1, currentCP);

            // (including 'end do' or 'continue')
            int numStmt=0;

            StmtListIterator next_stmt1(nstmt1->loop_body());

            while (next_stmt1.head())
            {
              numStmt++;
              next_stmt1();
            }

            StmtListIterator next_stmt0(nstmt1->loop_body());

            for (int i=1; i<numStmt; i++)
            {
              next_stmt0.head();
              next_stmt0();
            }

            //duplicating statements inside loop
            Expression* eb1=parser->make_expr(VAR_REF, 0, 0, stmt->node.fields.symbol);
            Expression* eb2=parser->make_expr(VAR_REF, 0, 0, tsym);

            for (int j=1; j<param; j++)
            {
              StmtListIterator next_stmt2(nstmt1->loop_body());
              next_stmt2.head();

              for (int i=1; i<numStmt; i++)
              {
                Statement* st1=DuplicateStmt(next_stmt2.head());

                Expression* ey1=NULL;

                if (!ex3)
                {
                  sprintf(buf, "%i", j);
                  string* x3=new string(buf);

                  ey1=make_float_const(parser, x3);
                }       	
                else if (isConst(ex3))
                {
                  int ex3val=(int) getValue(ex3);

                  sprintf(buf, "%i", j*ex3val);
                  string* x3=new string(buf);

                  ey1=make_float_const(parser, x3);
                }
                else
                {
                  sprintf(buf, "%i", j);
                  string* x3=new string(buf);

                  Expression* ey1a=make_float_const(parser, x3);
                  Expression* ey1b=DuplicateExpr(stmt->node.for_node.increment);

                  ey1=parser->intrinsic_op_node("*", MULT_OP, ey1a, ey1b);
                }
            
                Expression* ey3=parser->intrinsic_op_node("+", ADD_OP, eb1, ey1);

                ChangeExprInStmt(st1, eb1, ey3);

                next_stmt0.insert (st1, currentCP);
  
                next_stmt2();
              }
            }

            //Changing iterator for second loop
            StmtListIterator next_stmt3(stmt->loop_body());

            Statement* st2;
            while (st2=next_stmt3.head())
            {
              ChangeExprInStmt(st2, eb1, eb2);
              next_stmt3();
            }

            nstmt1->node.for_node.range->node.binary_op.r_operand=e1;
            nstmt1->node.for_node.increment=e2;

            stmt->node.fields.symbol=tsym;
            stmt->node.for_node.range->node.binary_op.l_operand=e4;
          }
          else if (tr==4)
          {
            //************************************************************
            //************************************************************
            //********************* loop strip_mining ********************
            //************************************************************
            //************************************************************

            loop_corr=1;

            double param=atof(sp1);
            char buf[1024];

            strcpy(buf, stmt->symbol()->ident());
            strcat(buf, "_x");
            Symbol* tsym = NewTemporaryVar (stmt->symbol()->type, buf);
            tsym->dovar=1;

            Expression* ex1=stmt->node.for_node.range->node.binary_op.l_operand;
            Expression* ex2=stmt->node.for_node.range->node.binary_op.r_operand;
            Expression* ex3=stmt->node.for_node.increment;

            if (!ex3)
            {
              string* x1=new string("1");

	      ex3=parser->make_expr (FLOAT_VAL, 0, 0, 0);
	      ex3->node.string_val = x1 ;
	      ex3->type = parser->global_float;
            }

            bool ex1const=isConst(ex1);
            bool ex2const=isConst(ex2);
            bool ex3const=isConst(ex3);

            if (!ex3const) return -1;

            double ex1val=getValue(ex1);
            double ex2val=getValue(ex2);
            double ex3val=getValue(ex3);

            Expression* ea1=DuplicateExpr(ex1);
            Expression* ea2=DuplicateExpr(ex2);
            Expression* ea3=NULL;

            Expression* eb1=parser->make_expr(VAR_REF, 0, 0, tsym);
            Expression* eb2=NULL;
            Expression* eb3=NULL;

            eb3=DuplicateExpr(ex3);

            sprintf(buf, "%g", param*ex3val);
            string* x1=new string(buf);

            ea3=make_float_const(parser, x1);

            sprintf(buf, "%g", param*ex3val);
            string* x2=new string(buf);

            Expression* e1=make_float_const(parser, x2);

            Expression* e2=DuplicateExpr(ex3);

            Expression* e3=parser->make_expr(VAR_REF, 0, 0, tsym);;

            Expression* e4=parser->intrinsic_op_node("+", ADD_OP, e3, e1);

            Expression* e5=NULL;

            e5=parser->intrinsic_op_node("-", SUBT_OP, e4, e2);

            Expression* e7=NULL;
            if (ex2const && (stmt->symbol()->type!=parser->global_int))
            {
              sprintf(buf, "%g", ex2val);
              addZero(buf);
              string* x1=new string(buf);
              e7=make_float_const(parser, x1);
            }
            else e7=DuplicateExpr(ex2);

            Expression* e8=parser->set_ll_list(e5, 0, EXPR_LIST);
            Expression* e9=parser->set_ll_list(e8, e7, EXPR_LIST);

            if (ex3val>0)
            {
              Symbol* sym1=(parser->global_scope()->look_up_ftn_sym("min"))->symbol();
              eb2=parser->make_expr(FUNC_CALL, e9, 0, sym1);
            }
            else
            {
              Symbol* sym1=(parser->global_scope()->look_up_ftn_sym("max"))->symbol();
              eb2=parser->make_expr(FUNC_CALL, e9, 0, sym1);
            }

            if (ex3val==1) eb3=NULL;

            //Creating new loop
            Expression* ddot=parser->make_expr(DDOT, ea1, ea2, 0);
            Statement* s1=parser->make_stmt (FOR_NODE, tsym, ddot, ea3, 0);

            //Creating list of statements inside new loop
            StmtList* sl1=new StmtList();
            StmtListIterator next_newstmt1(sl1);

            //insert original loop statement
            //(the reference to it will be removed later from the list
            // not to have duplicating references)
            next_newstmt1.insert (stmt, s1);

            Statement* s3=parser->make_stmt(CONTROL_END, 0, 0, 0, 0);
            next_newstmt1.insert (s3, s1);

            s1->node.fields.bl_ptr1=sl1;

            //Inserting new loop (which already contains original loop)
            next_stmt.insert(s1, currentCP);

            //Removing reference to original loop
            next_stmt.remove();

            stmt->node.for_node.range->node.binary_op.l_operand=eb1;
            stmt->node.for_node.range->node.binary_op.r_operand=eb2;
            stmt->node.for_node.increment=eb3;
          }
          else if (tr==5)
          {
            //************************************************************
            //************************************************************
            //********************** loop splitting **********************
            //************************************************************
            //************************************************************

            double param=atof(sp1);
            char buf[1024];

            Expression* ex1=stmt->node.for_node.range->node.binary_op.l_operand;
            Expression* ex2=stmt->node.for_node.range->node.binary_op.r_operand;
            Expression* ex3=stmt->node.for_node.increment;

            if (!ex3)
            {
              string* x1=new string("1");

	      ex3=parser->make_expr (FLOAT_VAL, 0, 0, 0);
	      ex3->node.string_val = x1 ;
	      ex3->type = parser->global_float;
            }

            bool ex1const=isConst(ex1);
            bool ex2const=isConst(ex2);
            bool ex3const=isConst(ex3);

            if (!ex3const) return -1;

            double ex1val=getValue(ex1);
            double ex2val=getValue(ex2);
            double ex3val=getValue(ex3);

            Expression* nexpr1=NULL;
            Expression* nexpr2=NULL;

            if (ex2const)
            {
              double var;
              var=param;
              if (ex3val>0) 
              {
                if (ex2val<var) var=ex2val; //min
              }
              else 
              {
                if (ex2val>var) var=ex2val; //max
              }
 
              sprintf(buf, "%g", var);
              string* x2=new string(buf);

              nexpr2=make_float_const(parser, x2);
            }
            else
            {
              sprintf(buf, "%g", param);
              if (stmt->symbol()->type!=parser->global_int) addZero(buf);
              string* x2=new string(buf);

              Expression* ex01=make_float_const(parser, x2);
              Expression* ex02=DuplicateExpr(ex2);

              Expression* ex03=parser->set_ll_list (ex02, 0, EXPR_LIST);

              Expression* ex04=parser->set_ll_list (ex03, ex01, EXPR_LIST);
              Symbol* sm01=NULL;
              if (ex3val>0) 
              {
                sm01=(parser->global_scope()->look_up_ftn_sym("min"))->symbol();
              }
              else
              {
                sm01=(parser->global_scope()->look_up_ftn_sym("max"))->symbol();
              }
              Expression* ex05=parser->make_expr (FUNC_CALL, ex04, 0, sm01);
 
              nexpr2=ex05;
            }

            if (ex1const)
            {
              double var;
              var=int(((param-ex1val)/ex3val)+1)*ex3val+ex1val;
              if (ex3val>0) 
              {
                if (ex1val>var) var=ex1val; //max
              }
              else 
              {
                if (ex1val<var) var=ex1val; //min
              }

              sprintf(buf, "%g", var);
              string* x1=new string(buf);

              nexpr1=make_float_const(parser, x1);
            }
            else
            {
              sprintf(buf, "%g", param);
              string* x1=new string(buf);

              Expression* ex01=make_float_const(parser, x1);
              Expression* ex02=DuplicateExpr(ex1);
              Expression* ex03=parser->intrinsic_op_node("-", SUBT_OP, ex01, ex02);

              Expression* ex03x1=DuplicateExpr(ex3);
              Expression* ex03x2=parser->intrinsic_op_node("/", DIV_OP, ex03, ex03x1);

              if (ex3val==1) ex03x2=ex03;

              string* x3=new string("1");
              Expression* ex03x3=make_float_const(parser, x3);
              Expression* ex03x4=parser->intrinsic_op_node("+", ADD_OP, ex03x2, ex03x3);

              Expression* ex03a=parser->set_ll_list (ex03x4, 0, EXPR_LIST);

              Symbol* sm01=(parser->global_scope()->look_up_ftn_sym("int"))->symbol();
              Expression* ex04=parser->make_expr (FUNC_CALL, ex03a, 0, sm01);

              Expression* ex03x5=DuplicateExpr(ex3);
              Expression* ex03x6=parser->intrinsic_op_node("*", MULT_OP, ex04, ex03x5);

              if (ex3val==1) ex03x6=ex04;

              Expression* ex03x7=DuplicateExpr(ex1);
              Expression* ex03x8=parser->intrinsic_op_node("+", ADD_OP, ex03x6, ex03x7);

              Expression* ex03x9=parser->set_ll_list (ex03x8, 0, EXPR_LIST);

              Expression* ex08=DuplicateExpr(ex1);
              Expression* ex09=parser->set_ll_list (ex08, ex03x9, EXPR_LIST);

              Symbol* sm02=NULL;
              if (ex3val>0) 
              {
                sm02=(parser->global_scope()->look_up_ftn_sym("max"))->symbol();
              }
              else
              {
                sm02=(parser->global_scope()->look_up_ftn_sym("min"))->symbol();
              }
              Expression* ex10=parser->make_expr (FUNC_CALL, ex09, 0, sm02);
  
              nexpr1=ex10;
            }

            //check number of loops inside stmt
            int loops=0;
            MiscCountLoops(stmt->loop_body(), &loops);
            loop_corr=1+loops;

            Statement* nstmt1=DuplicateStmt(stmt);
            nstmt1->enforce_enddo();
            nstmt1->node.for_node.range->node.binary_op.r_operand=nexpr2;
            next_stmt.insert (nstmt1, currentCP);
            
            stmt->node.for_node.range->node.binary_op.l_operand=nexpr1;
          }
          else if (tr==6)
          {
            //*************************************************************
            //*************************************************************
            //********************* loop distribution *********************
            //*************************************************************
            //*************************************************************

            // (including 'end do' or 'continue')
            int numStmt=0;

            StmtListIterator next_stmt1(stmt->loop_body());

            while (next_stmt1.head())
            {
              numStmt++;
              next_stmt1();
            }

            numStmt--;
            if (numStmt<1) return -2;

            loop_corr=numStmt-1;

            //duplicating loop
            int i=0;
            for (i=1; i<numStmt; i++)
            {
              Statement* nstmt1=DuplicateStmt(stmt);
              nstmt1->enforce_enddo();

              //removing elements
              StmtListIterator next_stmt2(nstmt1->loop_body());
              next_stmt2.head();

              for (int j=1; j<numStmt; j++) //xyz
              {
                if (i!=j) next_stmt2.remove();
                else next_stmt2();
              }
              next_stmt2.remove();

              next_stmt.insert (nstmt1, currentCP) ;
            }

            //removing elements from original loop
            //(don't do it in a previous step
            // since this loop may have 'continue'
            // so that we will not use 'enforce_enddo')

            StmtListIterator next_stmt3(stmt->loop_body());
            next_stmt3.head();

            for (int j=1; j<numStmt; j++)
            {
              next_stmt3.remove();
            }
          }

          return 1;
        }

	oldCP = currentCP ;
	currentCP = stmt ;

        ret=MiscLoopProcessBlockOfStatements(stmt->loop_body(),
                                             nestLevel+1,
                                             loop, tr, sp1, sp2);
        if (ret!=0) return ret;

	currentCP = oldCP ;
                  
        next_stmt();
        break;
      }
      case WHILE_NODE:
      {
	oldCP = currentCP ;
	currentCP = stmt ;

	ret=MiscLoopProcessBlockOfStatements(stmt->loop_body(), 
                                             nestLevel+1, 
                                             loop, tr, sp1, sp2);
        if (ret!=0) return ret;

	currentCP = oldCP ;

	next_stmt();
	break;
      }
      case LOGIF_NODE:
      {
        // Have to turn this into an IF_NODE otherwise we cannot transform the
	// the conditional statement. Then we drop through to the IF_NODE
	// case and complete the transformation.
	  
	stmt->stmt_variant (IF_NODE);
	Statement* condStat = stmt->node.if_node.true_branch;
	StmtList* then_pt = new StmtList();
	then_pt->append (condStat);
	then_pt->append (parser->make_stmt (CONTROL_END, 0, 0, 0, 0));
	stmt->node.fields.bl_ptr1 = then_pt;
	stmt->node.fields.bl_ptr2 = 0;
      }
      case IF_NODE:
      {
	oldCP = currentCP ;
	currentCP = stmt ;

	ret=MiscLoopProcessBlockOfStatements (stmt->then_part(), 0, loop, tr, sp1, sp2);
        if (ret!=0) return ret;
	ret=MiscLoopProcessBlockOfStatements (stmt->else_part(), 0, loop, tr, sp1, sp2);
        if (ret!=0) return ret;

	currentCP = oldCP ;

	break;
      }
      case ELSEIF_NODE:
      {
	oldCP = currentCP ;
	currentCP = stmt ;

	ret=MiscLoopProcessBlockOfStatements (stmt->then_part(), 0, loop, tr, sp1, sp2);
        if (ret!=0) return ret;
	ret=MiscLoopProcessBlockOfStatements (stmt->else_part(), 0, loop, tr, sp1, sp2);
        if (ret!=0) return ret;

	currentCP = oldCP ;

	break;
      }
    }
    if (stag != FOR_NODE) next_stmt();
  }

  return 0;
}

Statement* NextPerfectLoopNest(Statement* stmtx, int* nest)
{
  StmtList* block=stmtx->loop_body();

  if (!block) return NULL;

  StmtListIterator next_stmt(block);
  Statement* stmt;

  stmt=next_stmt.head();
  statement_tag stag = stmt->stmt_variant();

  if ((stag==FOR_NODE) || (stag==PARDO_NODE) || (stag==PDO_NODE))
  {
    (*nest)--;
    if ((*nest)==0) return stmt;
    return NextPerfectLoopNest(stmt, nest);
  }

  return NULL;
}

int MiscLoopDistribution(int sockfd, char* proc, char* sloop)
{
  int ret=0;
  fnd=0;
  long loop=atol(sloop);

  if (loop<=0)
  {
    SendMes(sockfd, "Error: Loop should be > 0!", 0, 1024); 
    return -1;
  }

  //Searching procedure
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;
  int found=0;

  while (func = next_routine())
  {
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        char* procName = func->symbol()->ident();
        if (strcmp(procName, proc)==0) found=1;
        break;
      }
      default:
      {
        break;
      }
    }

    if (found==1) break;
  }

  if (found==0)
  {
    SendMes(sockfd, "Error: Procedure was not found!", 0, 1024); 
    return -1;
  }

  lp=0;
  ret=MiscLoopProcessBlockOfStatements(func->statements(), 0, loop, 6, NULL, NULL);

  if (ret==0)
  {
    SendMes(sockfd, "Error: Loop was not found!", 0, 1024); 
    return -1;
  }
  else if (ret==-2)
  {
    SendMes(sockfd, "Error: Not enough statements inside loop to duplicate!", 0, 1024); 
    return -1;
  }

  return 0;
}

Statement* DuplicateStmt(Statement* stmt)
{
  if (!stmt) return NULL;

  Expression* e1=DuplicateExpr(stmt->node.fields.e1);
  Expression* e2=DuplicateExpr(stmt->node.fields.e2);
  Expression* e3=DuplicateExpr(stmt->node.fields.e3);  

  Statement* newstmt=parser->make_stmt (stmt->stmt_variant(), stmt->symbol(), e1, e2, e3);

  StmtList* block1=stmt->node.fields.bl_ptr1;
  if (block1)
  {
    StmtList* sl=new StmtList();

    StmtListIterator next_stmt1(block1);
    StmtListIterator next_newstmt1(sl);

    Statement* stmt1;
    while (stmt1=next_stmt1.head()) 
    {
      Statement* newstmt1=DuplicateStmt(stmt1);

      next_newstmt1.insert (newstmt1, stmt) ;

      next_stmt1();
    }

    newstmt->node.fields.bl_ptr1=sl;
  }

  return newstmt;
}

Expression* DuplicateExpr(Expression* expr)
{
  if (!expr) return NULL;

  Expression* e1=DuplicateExpr(expr->node.fields.e1);
  Expression* e2=DuplicateExpr(expr->node.fields.e2);

  Expression* newexpr=parser->make_expr (expr->expr_variant(), e1, e2, expr->node.fields.symbol);

  switch (expr->expr_variant())
  {
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case STRING_VAL:
    case STMT_STR:
    case KEYWORD_VAL:
    {
      newexpr->node.string_val=new string (expr->node.string_val);
    }
  }

  return newexpr;
}

int MiscLoopSplitting(int sockfd, char* proc, char* sloop, char* p1)
{
  int ret=0;
  fnd=0;
  long loop=atol(sloop);

  if (loop<=0)
  {
    SendMes(sockfd, "Error: Loop should be > 0!", 0, 1024); 
    return -1;
  }

  //Searching procedure
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;
  int found=0;

  while (func = next_routine())
  {
    currentCP=func;
    currentFunc=func;
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        char* procName = func->symbol()->ident();
        if (strcmp(procName, proc)==0) found=1;
        break;
      }
      default:
      {
        break;
      }
    }

    if (found==1) break;
  }

  if (found==0)
  {
    SendMes(sockfd, "Error: Procedure was not found!", 0, 1024); 
    return -1;
  }

  lp=0;
  ret=MiscLoopProcessBlockOfStatements(func->statements(), 0, loop, 5, p1, NULL);

  if (ret==0)
  {
    SendMes(sockfd, "Error: Loop was not found!", 0, 1024); 
    return -1;
  }
  else if (ret==-1)
  {
    SendMes(sockfd, "Error: Can't split loop with non-constant iterator!", 0, 1024); 
    return -1;
  }

  return 0;
}

Symbol* NewTemporaryVar(Datatype *newSymType, char* name)
{
  if (!newSymType) return NULL;

  StmtListIterator code_insert_point (currentFunc->statements()) ;

  while (code_insert_point.head()->stmt_variant()==IMPL_DECL) 
  {
    code_insert_point();
  }

  // 1. Create a new symbolic name for the temporary variable
  char nameBuff [128];
  char nameBuff1[128];
  if ((!name) || (strcmp(name, "")==0))
  {
    sprintf(nameBuff, "%s%d", eos_var_name_tmp, eos_var_seq++);
  }
  else
  {
    strcpy(nameBuff, name);
  }
           
  //checking if the symbol exists
  PTR_HASH h;
  long l=0;

  strcpy(nameBuff1, nameBuff);
  do
  {
    h=currentFunc->scoping()->just_look_up_ftn_sym_in_scope (nameBuff);
    if (h)
    {
      sprintf(nameBuff, "%s%d", nameBuff1, l);
      l++;  
    }
  } while (h);

  // 2. Create a new Symbol object and give it the correct type
//  Symbol* newSym=parser->make_local_symbol(nameBuff, VARIABLE_NAME, newSymType, HARD);

  PTR_HASH entry = currentFunc->scoping()->look_up_ftn (nameBuff, HARD);
  Symbol* newSym=parser->make_sym_entry(entry, VARIABLE_NAME, newSymType,  currentFunc->scoping(), HARD);

  // 3. Create a declaration statement for this variable
  Expression* vref = parser->make_expr(VAR_REF, 0, 0, newSym);
  Expression* vlst = parser->make_expr(EXPR_LIST, vref, 0, 0);
  Expression* tyop = parser->make_expr(TYPE_OP, 0, 0, 0);
  tyop->type=newSymType;
  Statement* newDecl=parser->make_stmt(VAR_DECL, 0, vlst, tyop, 0) ;

  // 4. Insert the new declaration at the decl_point
  code_insert_point.insert(newDecl, currentFunc);

  return newSym;
}

int MiscLoopStripmining(int sockfd, char* proc, char* sloop, char* p1)
{
  int ret=0;
  fnd=0;
  long loop=atol(sloop);
  double index=atof(p1);

  if (loop<=0)
  {
    SendMes(sockfd, "Error: Loop should be > 0!", 0, 1024); 
    return -1;
  }
  if (index<2)
  {
    SendMes(sockfd, "Error: Loop stripmining parameter should be > 1!", 0, 1024); 
    return -1;
  }

  //Searching procedure
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;
  int found=0;

  while (func = next_routine())
  {
    currentCP=func;
    currentFunc=func;
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        char* procName = func->symbol()->ident();
        if (strcmp(procName, proc)==0) found=1;
        break;
      }
      default:
      {
        break;
      }
    }

    if (found==1) break;
  }

  if (found==0)
  {
    SendMes(sockfd, "Error: Procedure was not found!", 0, 1024); 
    return -1;
  }

  lp=0;
  ret=MiscLoopProcessBlockOfStatements(func->statements(), 0, loop, 4, p1, NULL);

  if (ret==0)
  {
    SendMes(sockfd, "Error: Loop was not found!", 0, 1024); 
    return -1;
  }
  else if (ret==-1)
  {
    SendMes(sockfd, "Error: Can't stripmine loop with non-constant iterator!", 0, 1024); 
    return -1;
  }

  return 0;
}

bool isConst(Expression* expr)
{
  if (expr->expr_variant()==MINUS_OP)
  {
    if ((expr->node.fields.e1->expr_variant()==INT_VAL) ||
        (expr->node.fields.e1->expr_variant()==FLOAT_VAL) ||
        (expr->node.fields.e1->expr_variant()==DOUBLE_VAL)) return true;
  }
  else
  {
    if ((expr->expr_variant()==INT_VAL) ||
        (expr->expr_variant()==FLOAT_VAL) ||
        (expr->expr_variant()==DOUBLE_VAL)) return true;
  }

  return false;
}

double getValue(Expression* expr)
{
  if (!isConst(expr)) return 0;

  Expression *e1=expr;
  if (expr->expr_variant()==MINUS_OP) e1=expr->node.fields.e1;

  double ret;

  if      (e1->expr_variant()==INT_VAL)
          ret=e1->node.ival;
  else if (e1->expr_variant()==FLOAT_VAL)
          ret=atof((*(e1->node.string_val))());
  else if (e1->expr_variant()==DOUBLE_VAL)
          ret=atof((*(e1->node.string_val))());
  else ret=0;

  if (expr->expr_variant()==MINUS_OP) ret=-ret;  

  return ret;
}

void addZero(char* str)
{
  for (int i=0; i<strlen(str); i++)
  {
    if (str[i]=='.') return;
  }

  strcat(str, ".0");
}

int MiscLoopTiling(int sockfd, char* proc, char* sloop, char* p1)
{
  int ret=0;
  fnd=0;
  long loop=atol(sloop);
  double index=atof(p1);

  if (loop<=0)
  {
    SendMes(sockfd, "Error: Loop should be > 0!", 0, 1024); 
    return -1;
  }
  if (index<1)
  {
    SendMes(sockfd, "Error: Loop tiling parameter should be > 0!", 0, 1024); 
    return -1;
  }
  else if (index==1) return 0; //if tiling paramter == 1 - do nothing


  //Searching procedure
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;
  int found=0;

  while (func = next_routine())
  {
    currentCP=func;
    currentFunc=func;
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        char* procName = func->symbol()->ident();
        if (strcmp(procName, proc)==0) found=1;
        break;
      }
      default:
      {
        break;
      }
    }

    if (found==1) break;
  }

  if (found==0)
  {
    SendMes(sockfd, "Error: Procedure was not found!", 0, 1024); 
    return -1;
  }

  lp=0;
  ret=MiscLoopProcessBlockOfStatements(func->statements(), 0, loop, 1, p1, NULL);

  if (ret==0)
  {
    SendMes(sockfd, "Error: Loop was not found!", 0, 1024); 
    return -1;
  }
  else if (ret==-1)
  {
    SendMes(sockfd, "Error: Can't tile loop with non-constant iterator!", 0, 1024); 
    return -1;
  }

  return 0;
}

int MiscLoopUnrolling(int sockfd, char* proc, char* sloop, char* p1)
{
  int ret=0;
  fnd=0;
  long loop=atol(sloop);
  int index=atoi(p1);

  if (loop<=0)
  {
    SendMes(sockfd, "Error: Loop should be > 0!", 0, 1024); 
    return -1;
  }
  if (index<1)
  {
    SendMes(sockfd, "Error: Loop unrolling parameter should be > 0!", 0, 1024); 
    return -1;
  }
  else if (index==1) return 0; //if unrolling paramter == 1 - do nothing

  //Searching procedure
  StmtListIterator next_routine (parser->fileRoot()->headers());
  PTR_STMT func;
  int found=0;

  while (func = next_routine())
  {
    currentCP=func;
    currentFunc=func;
    switch (func->stmt_variant())
    {
      case PROG_HEDR:
      case PROC_HEDR:
      case FUNC_HEDR:
      {
        char* procName = func->symbol()->ident();
        if (strcmp(procName, proc)==0) found=1;
        break;
      }
      default:
      {
        break;
      }
    }

    if (found==1) break;
  }

  if (found==0)
  {
    SendMes(sockfd, "Error: Procedure was not found!", 0, 1024); 
    return -1;
  }

  lp=0;
  ret=MiscLoopProcessBlockOfStatements(func->statements(), 0, loop, 3, p1, NULL);

  if (ret==0)
  {
    SendMes(sockfd, "Error: Loop was not found!", 0, 1024); 
    return -1;
  }
  else if (ret==-1)
  {
    SendMes(sockfd, "Error: Unrolling factor should be > 1 !", 0, 1024); 
    return -1;
  }

  return 0;
}

void ChangeExprInStmt(Statement* st, Expression* e1, Expression* e2)
{
  if (!st) return;

  ChangeExpr(st->node.fields.e1, e1, e2);
  ChangeExpr(st->node.fields.e2, e1, e2);
  ChangeExpr(st->node.fields.e3, e1, e2);

  StmtList* block1=st->node.fields.bl_ptr1;
  if (block1)
  {
    StmtListIterator next_stmt1(block1);

    Statement* stmt1;
    while (stmt1=next_stmt1.head()) 
    {
      ChangeExprInStmt(stmt1, e1, e2);

      next_stmt1();
    }
  }
}

void ChangeExpr(Expression* expr, Expression* e1, Expression* e2)
{
  if (!expr) return;

  Expression* ex1=expr->node.fields.e1;
  Expression* ex2=expr->node.fields.e2;

  if (ex1)
  {
    if (ex1->expr_variant()==e1->expr_variant())
    {
      if (ex1->node.fields.symbol==e1->node.fields.symbol)
      {
        expr->node.fields.e1=e2;
      }
    }
  }

  if (ex2)
  {
    if (ex2->expr_variant()==e1->expr_variant())
    {
      if (ex2->node.fields.symbol==e1->node.fields.symbol)
      {
        expr->node.fields.e2=e2;
      }
    }
  }

  ChangeExpr(ex1, e1, e2);
  ChangeExpr(ex2, e1, e2);
}

void MiscCountLoops(StmtList* block, 
                   int* loops)
{  
  Statement* oldCP;

  if (!block) return;

  StmtListIterator next_stmt(block);
  Statement* stmt;
  while (stmt=next_stmt.head()) 
  {
    statement_tag stag = stmt->stmt_variant();
    currentLine = stmt->g_line;
    switch (stag)
    {
      case FOR_NODE:
      case PARDO_NODE:
      case PDO_NODE:
      {
        (*loops)++;

	oldCP = currentCP ;
	currentCP = stmt ;

        MiscCountLoops(stmt->loop_body(), loops);

	currentCP = oldCP ;
                  
        next_stmt();
        break;
      }
      case WHILE_NODE:
      {
	oldCP = currentCP ;
	currentCP = stmt ;

        MiscCountLoops(stmt->loop_body(), loops);

	currentCP = oldCP ;

	next_stmt();
	break;
      }
      case LOGIF_NODE:
      {
        // Have to turn this into an IF_NODE otherwise we cannot transform the
	// the conditional statement. Then we drop through to the IF_NODE
	// case and complete the transformation.
	  
	stmt->stmt_variant (IF_NODE);
	Statement* condStat = stmt->node.if_node.true_branch;
	StmtList* then_pt = new StmtList();
	then_pt->append (condStat);
	then_pt->append (parser->make_stmt (CONTROL_END, 0, 0, 0, 0));
	stmt->node.fields.bl_ptr1 = then_pt;
	stmt->node.fields.bl_ptr2 = 0;
      }
      case IF_NODE:
      {
	oldCP = currentCP ;
	currentCP = stmt ;

        MiscCountLoops(stmt->then_part(), loops);
        MiscCountLoops(stmt->else_part(), loops);

	currentCP = oldCP ;

	break;
      }
      case ELSEIF_NODE:
      {
	oldCP = currentCP ;
	currentCP = stmt ;

        MiscCountLoops(stmt->then_part(), loops);
        MiscCountLoops(stmt->else_part(), loops);

	currentCP = oldCP ;

	break;
      }
    }
    if (stag != FOR_NODE) next_stmt();
  }

  return;
}

Expression* make_int_const(ParseContext* pc, int val)
{
  Expression* e=NULL;

  e=pc->make_expr (INT_VAL, 0, 0, 0);
  e->node.ival = val;
  e->type = pc->global_int;

  return e;
}

Expression* make_float_const(ParseContext* pc, string* val)
{
  Expression* e=NULL;

  e=pc->make_expr (FLOAT_VAL, 0, 0, 0);
  e->node.string_val = val;
  e->type = pc->global_float;

  return e;
}

